/*

 APUskip.S - Contains APU skipper

*/

#include "misc.i"

EXPORT_C APUskip_start

.bss
ALIGN
.globl APUI00a,APUI00b,APUI00c
APUI00a:    skipb   // This is the APU value when APUI00b=0
APUI00b:    skipb   // This is a count type of variable for flipping APUI00a
APUI00c:    skipb   // Binary counter used in conjuction with APUI01c
APUI01a:    skipb   // This is the APU value when APUI01b=0
APUI01b:    skipb   // This is a count type of variable for flipping APUI01a
APUI01c:    skipb   // Binary counter upper byte of APUI00c
APUI02a:    skipb   // This is the APU value when APUI02b=0
APUI02b:    skipb   // This is a count type of variable for flipping APUI02a
APUI02c:    skipb   // Binary counter used in conjuction with APUI03c
APUI03a:    skipb   // This is the APU value when APUI03b=0
APUI03b:    skipb   // This is a count type of variable for flipping APUI03a
APUI03c:    skipb   // Binary counter upper byte of APUI02c

.text
ALIGN
.globl _Reset_APU_Skipper
_Reset_APU_Skipper:
  pusha
  // Set eax to 0, as we're setting most everything to 0...
  xorl %eax,%eax

  movb %al,APUI00a
  movb %al,APUI00b
  movb %al,APUI00c
  movb %al,APUI01a
  movb %al,APUI01b
  movb %al,APUI01c
  movb %al,APUI02a
  movb %al,APUI02b
  movb %al,APUI02c
  movb %al,APUI03a
  movb %al,APUI03b
  movb %al,APUI03c
  popa

  ret

.globl _Make_APU_Skipper
ALIGN
_Make_APU_Skipper:
    pusha
    movl $SNES_R2140_SKIP,%eax
    movl %eax,Read_Map_21+0x40*4
    movl $SNES_W2140_SKIP,%eax
    movl %eax,Write_Map_21+0x40*4
    movl $SNES_R2141_SKIP,%eax
    movl %eax,Read_Map_21+0x41*4
    movl $SNES_W2141_SKIP,%eax
    movl %eax,Write_Map_21+0x41*4
    movl $SNES_R2142_SKIP,%eax
    movl %eax,Read_Map_21+0x42*4
    movl $SNES_W2142_SKIP,%eax
    movl %eax,Write_Map_21+0x42*4
    movl $SNES_R2143_SKIP,%eax
    movl %eax,Read_Map_21+0x43*4
    movl $SNES_W2143_SKIP,%eax
    movl %eax,Write_Map_21+0x43*4
    popa
    ret

ALIGN
SNES_R2140_SKIP:    // APUI00
    cmpb $0,APUI00b
    jnz 0f

    movb __65816_A,%al
    incb APUI00b
    movb $0,APUI01b
    ret

0:  cmpb $1,APUI00b
    jne  0f

    movb __65816_X,%al
    incb APUI00b
    movb $1,APUI01b
    ret

0:  cmpb $2,APUI00b
    jne 0f

    movb __65816_Y,%al
    incb APUI00b
    movb $2,APUI01b
    ret

0:  cmpb $3,APUI00b
    jne 0f

    movb $0,%al
    incb APUI00b
    ret

0:  cmpb $4,APUI00b
    jne 0f

    movb $0xff,%al
    incb APUI00b
    ret

0:  cmpb $5,APUI00b
    jne 0f

    movb $0x55,%al
    incb APUI00b
    ret

0:  cmpb $6,APUI00b
    jne 0f

    movb $0x01,%al
    incb APUI00b
    ret

0:  cmpb $7,APUI00b
    jne 0f

    movb $0xaa,%al
    movb $6,APUI01b
    incb APUI00b
    ret

0:  cmpb $8,APUI00b
    jne 0f

    movb APUI00a,%al
    incb APUI00b
    ret

0:  movb APUI00c,%al            // New extra skipper, if all else fails this should work for 2140!
    incb APUI00c
    movb $0xb,APUI01b           // This keeps high word at 0 during cycle (for now at least)
    cmpb $0,APUI00c
    jz 0f
    ret

0:  movb $0,APUI00b             // Ensures the skipper switches off
    incb APUI01c
    movb $0,APUI01b
    ret

ALIGN
SNES_R2141_SKIP:    // APUI01
    cmpb $0,APUI01b
    jnz 0f

    movb __65816_A+1,%al
    incb APUI01b
    ret

0:  cmpb $1,APUI01b
    jnz 0f

    movb __65816_X+1,%al
    incb APUI01b
    ret

0:  cmpb $2,APUI01b
    jnz 0f

    movb __65816_Y+1,%al
    incb APUI01b
    ret

0:  cmpb $3,APUI01b
    jnz 0f

    movb __65816_A,%al
    incb APUI01b
    ret

0:  cmpb $4,APUI01b
    jnz 0f

    movb __65816_X,%al
    incb APUI01b
    ret

0:  cmpb $5,APUI01b
    jnz 0f

    movb __65816_Y,%al
    incb APUI01b
    ret

0:  cmpb $6,APUI01b
    jnz 0f

    movb $0xbb,%al
    incb APUI01b
    ret

0:  cmpb $7,APUI01b
    jnz 0f

    movb $0,%al
    incb APUI01b
    ret

0:  cmpb $8,APUI01b
    jnz 0f

    movb $0xff,%al
    incb APUI01b
    ret

0:  cmpb $9,APUI01b
    jnz 0f

    movb $0x55,%al
    incb APUI01b
    ret

0:  cmpb $0xa,APUI01b
    jnz 0f

    movb APUI01a,%al
    movb $0,APUI01b
    ret

0:  movb APUI01c,%al        // This can only be reached in special cases
    ret

ALIGN
SNES_R2142_SKIP:    // APUI02
    cmpb $0,APUI02b
    jnz 0f

    movb __65816_A,%al
    incb APUI02b
    movb $0,APUI03b
    ret

0:  cmpb $1,APUI02b
    jnz 0f

    movb __65816_X,%al
    incb APUI02b
    movb $1,APUI03b
    ret

0:  cmpb $2,APUI02b
    jnz 0f

    movb __65816_Y,%al
    incb APUI02b
    movb $2,APUI03b
    ret

0:  cmpb $3,APUI02b
    jnz 0f

    movb $0,%al
    incb APUI02b
    ret

0:  cmpb $4,APUI02b
    jnz 0f

    movb $0xff,%al
    incb APUI02b
    ret

0:  cmpb $5,APUI02b
    jnz 0f

    movb $0x55,%al
    incb APUI02b
    ret

0:  cmpb $6,APUI02b
    jnz 0f

    movb $0xaa,%al
    movb $6,APUI03b
    incb APUI02b
    ret

0:  movb APUI02a,%al
    movb $0,APUI02b
    ret

ALIGN
SNES_R2143_SKIP:    // APUI03
    cmpb $0,APUI03b
    jnz 0f

    movb __65816_A+1,%al
    incb APUI03b
    ret

0:  cmpb $1,APUI03b
    jnz 0f

    movb __65816_X+1,%al
    incb APUI03b
    ret

0:  cmpb $2,APUI03b
    jnz 0f

    movb __65816_Y+1,%al
    incb APUI03b
    ret

0:  cmpb $3,APUI03b
    jnz 0f

    movb __65816_A,%al
    incb APUI03b
    ret

0:  cmpb $4,APUI03b
    jnz 0f

    movb __65816_X,%al
    incb APUI03b
    ret

0:  cmpb $5,APUI03b
    jnz 0f

    movb __65816_Y,%al
    incb APUI03b
    ret

0:  cmpb $6,APUI03b
    jnz 0f

    movb $0xbb,%al
    incb APUI03b
    ret

0:  cmpb $7,APUI03b
    jnz 0f

    movb $0,%al
    incb APUI03b
    ret

0:  cmpb $8,APUI03b
    jnz 0f

    movb $0xff,%al
    incb APUI03b
    ret

0:  cmpb $9,APUI03b
    jnz 0f

    movb $0x55,%al
    incb APUI03b
    ret

0:  movb APUI03a,%al
    movb $0,APUI03b
    ret

ALIGN
SNES_W2140_SKIP:    // APUI00
    movb %al,APUI00a
    cmpb $0xff,%al
    je 0f
//  movb $0,APUI00b
    ret
0://movb $7,APUI00b
    ret

ALIGN
SNES_W2141_SKIP:    // APUI01
    movb %al,APUI01a
    movb $0,APUI01b
    ret

ALIGN
SNES_W2142_SKIP:    // APUI02
    movb %al,APUI02a
    movb $0,APUI02b
    ret

ALIGN
SNES_W2143_SKIP:    // APUI03
    movb %al,APUI03a
    movb $0,APUI03b
    ret

.text; ALIGN
.data; ALIGN
.bss; ALIGN
