#define BANK_B(variable) (variable+2)
#define _P_B    %cl // Flags Register - See information in CPU.S!
#define _P_W    %ecx
#define _P  __65816_P
#define _A  __65816_A
#define _X  __65816_X
#define _Y  __65816_Y
#define _S  __65816_S
#define _D  __65816_D
#define _PC __65816_PC
#define _DB_Shifted __65816_DB_Shifted
#define _PB_Shifted __65816_PB_Shifted
#define _DB (_DB_Shifted+2)
#define _PB (_PB_Shifted+2)

#define NMITrip _NMITrip
#define IRQTrip _IRQTrip
#define LineCycles _LineCycles

//  True 65816 layout, Native Mode    = |E|N|V|M|X|D|I|Z|C|
//  True 65816 layout, Emulation Mode = |E|N|V|1|B|D|I|Z|C|

// These are the bits for flag set/clr operations
// |E|N|Z|I|D|X|M|V|C|
//           |B|1|

SNES_FLAG_C = 1     // Carry
SNES_FLAG_V = 2     // Overflow
SNES_FLAG_M = 4     // When E=0: Memory/accumulator operations 8-bit
SNES_FLAG_1 = 4     // When E=1: This bit is always set
SNES_FLAG_X = 8     // When E=0: Index registers 8-bit
SNES_FLAG_B = 8     // When E=1: Break (clear only on stack after IRQ/NMI)
SNES_FLAG_D = 0x10  // Decimal ADC/SBC mode
SNES_FLAG_I = 0x20  // Interrupt Disable
SNES_FLAG_Z = 0x40  // Zero result
SNES_FLAG_N = 0x80  // Negative result
SNES_FLAG_E = 0x100 // Emulation mode

SNES_FLAG_B1 = (SNES_FLAG_B | SNES_FLAG_1)
SNES_FLAG_MX = (SNES_FLAG_M | SNES_FLAG_X)
SNES_FLAG_NZ = (SNES_FLAG_N | SNES_FLAG_Z)
SNES_FLAG_NZC = (SNES_FLAG_NZ | SNES_FLAG_C)

REAL_SNES_FLAG_C = 1    // Carry
REAL_SNES_FLAG_Z = 2    // Zero result
REAL_SNES_FLAG_I = 4    // Interrupt Disable
REAL_SNES_FLAG_D = 8    // Decimal ADC/SBC mode
REAL_SNES_FLAG_X = 0x10 // When E=0: Index registers 8-bit
REAL_SNES_FLAG_B = 0x10 // When E=1: Break (clear only on stack after IRQ/NMI)
REAL_SNES_FLAG_M = 0x20 // When E=0: Memory/accumulator operations 8-bit
REAL_SNES_FLAG_1 = 0x20 // When E=1: This bit is always set
REAL_SNES_FLAG_V = 0x40 // Overflow
REAL_SNES_FLAG_N = 0x80 // Negative result
REAL_SNES_FLAG_E = 0x100    // Emulation mode

.macro  SET_FLAG flag
.if \flag >> 8
 orl $\flag,_P_W
.else
 orb $\flag,_P_B
.endif
.endm

.macro  CLR_FLAG flag
.if \flag >> 8
 andl $~\flag,_P_W
.else
 andb $~\flag,_P_B
.endif
.endm

.macro  CPL_FLAG flag
.if \flag >> 8
 xorl $\flag,_P_W
.else
 xorb $\flag,_P_B
.endif
.endm

.macro TEST_FLAG flag
.if \flag >> 8
 testl $\flag,_P_W  // ZF=!flag
.else
 testb $\flag,_P_B  // ZF=!flag
.endif
.endm

.macro JUMP_FLAG flag,wheretogo
 TEST_FLAG \flag
 jnz \wheretogo
.endm

.macro JUMP_NOT_FLAG flag,wheretogo
 TEST_FLAG \flag
 jz  \wheretogo
.endm

// Set 65816 flags to PC flags
// Can only set flags NZC - corrupts ah

.macro STORE_FLAGS flags
    lahf
    STORE_FLAGS_FROM \flags,%ah
.endm

// Set 65816 flags to PC flags
// Can only set flags NZC - corrupts loc
// Assumes loc has PC flags (as from lahf)
.macro STORE_FLAGS_FROM flags,loc
    andb $~(\flags),_P_B
    andb $\flags,\loc
    orb \loc,_P_B
.endm

// Set data to P
.macro FLAGS_TO data
.if \data >> 8
 movl $\data,_P
.else
 movb $\data,_P
.endif
.endm

// Load P to register _P_W
.macro LOAD_P
 movl _P,_P_W
.endm

// Save register _P_W to P
.macro SAVE_P
 movl _P_W,_P
.endm

// PC carry flag = 65816 carry flag (all others undefined)
.macro MAKE_CARRY
    rolb $8,_P_B
.endm

// PC carry flag = NOT 65816 carry flag (all others undefined)
.macro MAKE_NOT_CARRY
    rolb $8,_P_B
    cmc
.endm

// Set up the flags from PC flag format to 65816 flag format stored in %al
// Corrupts eax
.macro  E0_SETUPFLAGS from=_P_B,to=%al
//.macro  Flags_Native_to_65c816_E0 from=_P_B,to=%al
    xorl %eax,%eax
    movb \from,%al
    movb E0_Setup_Restore_Flags_Table(%eax),\to
.endm

.macro  E1_SETUPFLAGS from=_P_B,to=%al
//.macro  Flags_Native_to_65c816_E1 from=_P_B,to=%al
    xorl %eax,%eax
    movb \from,%al
    movb E1_Setup_Flags_Table(%eax),\to
.endm

// Restore flags from SNES format in %al to PC format
// Corrupts eax
.macro  E0_RESTOREFLAGS to=_P_B
//.macro  Flags_65c816_to_Native_E0 to=_P_B
    andl $0xFF,%eax
    movb E0_Setup_Restore_Flags_Table(%eax),%al
    movb %al,\to
.endm

.macro  E1_RESTOREFLAGS to=_P_B
//.macro  Flags_65c816_to_Native_E1 to=_P_B
    andl $0xFF,%eax
    movb E1_Restore_Flags_Table(%eax),%al
    movb %al,\to
.endm

/*
 Set the current opcode execution and timing table pointers based
 on the M and X bits of flag register. This returns control to the
 execution loop, so it must be the last instruction
 Corrupts eax,ecx
*/

.macro  SET_TABLE_MX from=_P_B
    movb \from,%al
    andl $SNES_FLAG_MX,%eax // Mask out M & X flags
    JUMP_NOT_FLAG SNES_FLAG_X,7f
    xorl %edx,%edx          // Clear XH/YH if X flag set
    movb %dl,_X+1
    movb %dl,_Y+1
7:  movl CPU_OpTables(%eax),%eax
    movl %eax,OpTable
    jmp CPU_RETURN
.endm

/* Push / Pull Macros assume eax contains value - corrupt ebx,edi */

.macro  E0_PUSH_B       // Non Emulation Push Byte (S_W--)
    movl _S,%ebx        // S only - bank always 0!
#ifdef OLD_65c816_STACK
    call *_Write_BankMapping    // Write to bank 0
#else
    cmpb $0x20,%bh
    jnb 0f
    movb %al,_WRAM(%ebx)
0:
#endif
    decl %ebx           // Decrement S
    movw %bx,_S         // Set stack pointer
.endm

.macro  E1_PUSH_B       // Emulation Push Byte (S_B--)
    movl _S,%ebx        // S only - bank always 0!
    decb _S             // Decrement S (Byte)
#ifdef OLD_65c816_STACK
    call *_Write_BankMapping    // Write to bank 0
#else
    movb %al,_WRAM(%ebx)
#endif
.endm

.macro  E0_PULL_B       // Non Emulation Pull Byte (S_W++)
    movl _S,%ebx        // S only - bank always 0!
    incw %bx            // Increment S
    movl %ebx,_S        // Set stack pointer
#ifdef OLD_65c816_STACK
    call *_Read_BankMapping // Read from bank 0
#else
    cmpb $0x20,%bh
    jnb 0f
    movb _WRAM(%ebx),%al
0:
#endif
.endm

.macro  E1_PULL_B       // Emulation Pull Byte (S_B--)
    incb _S             // Increment S (Byte)
    movl _S,%ebx        // S only - bank always 0!
#ifdef OLD_65c816_STACK
    call *_Read_BankMapping // Read from bank 0
#else
    movb _WRAM(%ebx),%al
#endif
.endm

.macro  E0_PUSH_W       // Non Emulation Push Word (S_W--)
    movl _S,%ebx        // S only - bank always 0!
#ifdef OLD_65c816_STACK
    xchgb %al,%ah
    call *_Write_BankMapping    // Write to bank 0
#else
    cmpb $0x20,%bh
    jnb 0f
    movb %ah,_WRAM(%ebx)
0:
#endif
    decw %bx            // Decrement S
#ifdef OLD_65c816_STACK
    xchgb %al,%ah
    call *_Write_BankMapping    // Write to bank 0
#else
    cmpb $0x20,%bh
    jnb 0f
    movb %al,_WRAM(%ebx)
0:
#endif
    decw %bx            // Decrement S
    movl %ebx,_S        // Set stack pointer
.endm

.macro  E1_PUSH_W       // Emulation Push Word (S_B--)
    movl _S,%ebx        // S only - bank always 0!
#ifdef OLD_65c816_STACK
    xchgb %al,%ah       // Get high byte in al
    call *_Write_BankMapping    // Write to bank 0
#else
    movb %ah,_WRAM(%ebx)
#endif
    decb _S             // Decrement S (Byte)
    movl _S,%ebx        // S only - bank always 0!
#ifdef OLD_65c816_STACK
    xchgb %al,%ah       // Now lo byte in al
    call *_Write_BankMapping    // Write to bank 0
#else
    movb %al,_WRAM(%ebx)
#endif
    decb _S             // Decrement S (Byte)
.endm

.macro  E0_PULL_W       // Non Emulation Pull Word (S_W--)
    movl _S,%ebx        // S only - bank always 0!
    incw %bx            // Increment S
#ifdef OLD_65c816_STACK
    call *_Read_BankMapping // Read from bank 0
    movb %al,%ah        // Put lo byte into high
#else
    cmpb $0x20,%bh
    jnb 0f
    movb _WRAM(%ebx),%al
0:
#endif
    incw %bx            // Increment S
#ifdef OLD_65c816_STACK
    call *_Read_BankMapping // Read from bank 0
    xchgb %al,%ah       // swap hi,lo bytes
#else
    cmpb $0x20,%bh
    jnb 0f
    movb _WRAM(%ebx),%ah
0:
#endif
    movl %ebx,_S        // Set stack pointer
.endm

.macro  E1_PULL_W       // Emulation Pull Word (S_B--)
    incb _S             // Increment S
    movl _S,%ebx        // S only - bank always 0!
#ifdef OLD_65c816_STACK
    call *_Read_BankMapping // Read from bank 0
    movb %al,%ah        // Put lo byte into high
#else
    movb _WRAM(%ebx),%al
#endif
    incb _S             // Increment S
    movl _S,%ebx        // S only - bank always 0!
#ifdef OLD_65c816_STACK
    call *_Read_BankMapping // Read from bank 0
    xchgb %al,%ah       // swap hi,lo bytes
#else
    movb _WRAM(%ebx),%ah
#endif
.endm

// Requires current scanline to be in %eax
.macro CheckIRQ
    movb NMITIMEN,%dl

    testb $0x10,%dl         // Horizontal IRQ enabled?
    jz  2f
    movb $0x40,%dh          // H-IRQ
    testb $0x20,%dl         // Vertical IRQ enabled?
    jz  1f
    movb $0xC0,%dh          // H+V-IRQ
    cmpl VTIMEL,%eax        // IRQ set for this scanline?
    jne 0f
    jmp 1f
2:
    movb $0x80,%dh          // V-IRQ
    cmpl VTIMEL,%eax        // IRQ set for this scanline?
    jne 0f
    testb $0x20,%dl         // Vertical IRQ enabled?
    jz  0f
1:  movb $0x80,TIMEUP       // Set IRQ recv'd flag
//  orb  $0x80,IRQ_pin
    orb  %dh,IRQ_pin
    movl In_WAI,%eax
    testl %eax,%eax
    jz  1f
    incw _PC
    movl $0,In_WAI
1:  LOAD_P
    JUMP_FLAG SNES_FLAG_I,0f    // Interrupts disabled?
    pushl $1f
    JUMP_NOT_FLAG SNES_FLAG_E,E0_IRQ
    jmp E1_IRQ
ALIGN
1:  SAVE_P
0:
.endm
