/*

 65c816 opcodes 0x00-0x0F:
    00: BRK s       2,8
    01: ORA (d,x)   2,6
    02: COP s       2,8
    03: ORA d,s     2,4
    04: TSB d       2,5
    05: ORA d       2,3
    06: ASL d       2,5
    07: ORA [d]     2,6
    08: PHP s       1,3
    09: ORA #       2,2
    0A: ASL a       1,2
    0B: PHD s       1,4
    0C: TSB a       3,6
    0D: ORA a       3,4
    0E: ASL a       3,6
    0F: ORA al      4,5
*/

.macro NativeInterrupt vector,offset
    incl %ebx
    movb _PB,%al
    pushl %ebx
    E0_PUSH_B
    popl %eax
    E0_PUSH_W
    E0_SETUPFLAGS               // put flags into SNES packed flag format
    E0_PUSH_B

    movl \vector,%eax           // Get Native mode vector
    movl %eax,_PC               // Setup PC vector
    movl \offset,%eax
    movb $0,_PB                 // Setup bank
    movl %eax,_PBOffset // Offset of bank 0 ROM
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    movl %eax,_PB_Read  // Bank 0 read through memory mapper

    SET_FLAG SNES_FLAG_I    // Disable interrupts
    CLR_FLAG SNES_FLAG_D
    jmp CPU_RETURN
.endm

.macro Emu6502Interrupt vector,offset
    incl %ebx
    movl %ebx,%eax
    E1_PUSH_W
    E1_SETUPFLAGS               // Put flags into SNES packed flag format
    E1_PUSH_B

    movl \vector,%eax           // Get Emulation mode vector
    movl %eax,_PC               // Setup PC vector
    movl \offset,%eax
    movb $0,_PB                 // Setup bank
    movl %eax,_PBOffset // Offset of bank 0 ROM
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    movl %eax,_PB_Read  // Bank 0 read through memory mapper

    SET_FLAG SNES_FLAG_I    // Disable interrupts
    CLR_FLAG SNES_FLAG_D
    jmp CPU_RETURN
.endm

/* 00 */

ALIGN
E0_BRK:
    NativeInterrupt _BRK_Nvector,_BRK_Noffset

ALIGN
E1_BRK:
    Emu6502Interrupt _IRQ_Evector,_IRQ_Eoffset

/* 01 */

ALIGN
EM_ORA_Od_xO:
    EM_ORA Direct_Indexed_Indirect
    jmp CPU_RETURN
ALIGN
E0_ORA_Od_xO:
    E0_ORA Direct_Indexed_Indirect
    jmp CPU_RETURN

/* 02 */

ALIGN
E0_COP:
    NativeInterrupt _COP_Nvector,_COP_Noffset

ALIGN
E1_COP:
    Emu6502Interrupt _COP_Evector,_COP_Eoffset

/* 03 */

ALIGN
EM_ORA_d_s:
    EM_ORA Stack_Relative
    jmp CPU_RETURN
ALIGN
E0_ORA_d_s:
    E0_ORA Stack_Relative
    jmp CPU_RETURN

/* 04 */

ALIGN
EM_TSB_d:
    EM_TSB Direct
    jmp CPU_RETURN
ALIGN
E0_TSB_d:
    E0_TSB Direct
    jmp CPU_RETURN

/* 05 */

ALIGN
EM_ORA_d:
    EM_ORA Direct
    jmp CPU_RETURN
ALIGN
E0_ORA_d:
    E0_ORA Direct
    jmp CPU_RETURN

/* 06 */

ALIGN
EM_ASL_d:
    EM_ASL Direct
    jmp CPU_RETURN
ALIGN
E0_ASL_d:
    E0_ASL Direct
    jmp CPU_RETURN

/* 07 */

ALIGN
EM_ORA_IdI:
    EM_ORA Direct_Indirect_Long
    jmp CPU_RETURN
ALIGN
E0_ORA_IdI:
    E0_ORA Direct_Indirect_Long
    jmp CPU_RETURN

/* 08 */

ALIGN
E0_PHP:
    movw %bx,_PC
    E0_SETUPFLAGS
    E0_PUSH_B
    jmp CPU_RETURN
ALIGN
E1_PHP:
    movw %bx,_PC
    E1_SETUPFLAGS
    E1_PUSH_B
    jmp CPU_RETURN

/* 09 */

ALIGN
EM_ORA_i:
    EM_ORA Immediate
    jmp CPU_RETURN
ALIGN
E0_ORA_i:
    E0_ORA Immediate
    jmp CPU_RETURN

/* 0A */

ALIGN
EM_SLA:
    movw %bx,_PC
    shlb $1,_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
ALIGN
E0_SLA:
    movw %bx,_PC
    shlw $1,_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN

/* 0B */
ALIGN
E0_PHD:
    movw %bx,_PC
    movl _D,%eax
    E0_PUSH_W
    jmp CPU_RETURN
ALIGN
E1_PHD:
    movw %bx,_PC
    movl _D,%eax
    E1_PUSH_W
    jmp CPU_RETURN

/* 0C */
ALIGN
EM_TSB_a:
    EM_TSB Absolute
    jmp CPU_RETURN
ALIGN
E0_TSB_a:
    E0_TSB Absolute
    jmp CPU_RETURN

/* 0D */
ALIGN
EM_ORA_a:
    EM_ORA Absolute
    jmp CPU_RETURN
ALIGN
E0_ORA_a:
    E0_ORA Absolute
    jmp CPU_RETURN

/* 0E */
ALIGN
EM_ASL_a:
    EM_ASL Absolute
    jmp CPU_RETURN
ALIGN
E0_ASL_a:
    E0_ASL Absolute
    jmp CPU_RETURN

/* 0F */

ALIGN
EM_ORA_al:
    EM_ORA Absolute_Long
    jmp CPU_RETURN
ALIGN
E0_ORA_al:
    E0_ORA Absolute_Long
    jmp CPU_RETURN
