/*

 65c816 opcodes 0x20-0x2F:
    20: JSR a       3,6
    21: ORA (d,x)   2,6
    22: JSL al      4,8
    23: AND d,s     2,4
    24: BIT d       2,3
    25: AND d       2,3
    26: ROL d       2,5
    27: AND [d]     2,6
    28: PLP s       1,4
    29: AND #       2,2
    2A: ROL a       1,2
    2B: PLD s       1,5
    2C: BIT a       3,4
    2D: AND a       3,4
    2E: ROL a       3,6
    2F: AND al      4,5
*/

/* 20 */

ALIGN
E0_JSR_a:
    ADDR_Absolute_JMP
    movl %ebx,%eax      // last instruction byte PC in bx, new PC already set
    E0_PUSH_W           // Address of last byte not next instruction (huh!)
    movb _PC+1,%al
    xorl %ebx,%ebx
    addb %al,%al
    movb _PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

ALIGN
E1_JSR_a:
    ADDR_Absolute_JMP
    movl %ebx,%eax      // last instruction byte PC in bx, new PC already set
    E1_PUSH_W           // Address of last byte not next instruction (huh!)
    movb _PC+1,%al
    xorl %ebx,%ebx
    addb %al,%al
    movb _PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

/* 21 */

ALIGN
EM_AND_Od_xO:
    EM_AND Direct_Indexed_Indirect
    jmp CPU_RETURN
ALIGN
E0_AND_Od_xO:
    E0_AND Direct_Indexed_Indirect
    jmp CPU_RETURN

/* 22 */

ALIGN
E0_JSL_al:              // JSR Absolute long address
    movl %ebx,%edx
    movb _PB,%al
    addl $2,%edx        // Address of last byte of instruction
    E0_PUSH_B           // Push PB
    movl %edx,%eax
    E0_PUSH_W           // Push PC

    movb 1(%esi),%bl
    movb 2(%esi),%ah
    movb %bl,_PC
    xorl %ebx,%ebx
    movb 3(%esi),%bl
    movb %ah,_PC+1
    movb %bl,_PB
    testb %bl,MEMSEL    // Check bus speed against PB
    movb $5,%dl
    jnz 0f
    movb $4,%dl
0:  addb %ah,%ah
    movb %dl,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

ALIGN
E1_JSL_al:
    movl %ebx,%edx
    movb _PB,%al
    addl $2,%edx        // Address of last byte of instruction
    E1_PUSH_B           // Push PB
    movl %edx,%eax
    E1_PUSH_W           // Push PC

    movb 1(%esi),%bl
    movb 2(%esi),%ah
    movb %bl,_PC
    xorl %ebx,%ebx
    movb 3(%esi),%bl
    movb %ah,_PC+1
    movb %bl,_PB
    testb %bl,MEMSEL    // Check bus speed against PB
    movb $5,%dl
    jnz 0f
    movb $4,%dl
0:  addb %ah,%ah
    movb %dl,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

/* 23 */

ALIGN
EM_AND_d_s:
    EM_AND Stack_Relative
    jmp CPU_RETURN
ALIGN
E0_AND_d_s:
    E0_AND Stack_Relative
    jmp CPU_RETURN

/* 24 */

ALIGN
EM_BIT_d:
    EM_BIT Direct
    jmp CPU_RETURN
ALIGN
E0_BIT_d:
    E0_BIT Direct
    jmp CPU_RETURN

/* 25 */

ALIGN
EM_AND_d:
    EM_AND Direct
    jmp CPU_RETURN
ALIGN
E0_AND_d:
    E0_AND Direct
    jmp CPU_RETURN

/* 26 */

ALIGN
EM_ROL_d:
    EM_ROL Direct
    jmp CPU_RETURN
ALIGN
E0_ROL_d:
    E0_ROL Direct
    jmp CPU_RETURN

/* 27 */

ALIGN
EM_AND_IdI:
    EM_AND Direct_Indirect_Long
    jmp CPU_RETURN
ALIGN
E0_AND_IdI:
    E0_AND Direct_Indirect_Long
    jmp CPU_RETURN

/* 28 */

ALIGN
E0_PLP:
    movw %bx,_PC
    E0_PULL_B
    E0_RESTOREFLAGS

    JUMP_FLAG SNES_FLAG_I,0f
    movb IRQ_pin,%al
    testb %al,%al
    jz  0f
    call E0_IRQ
0:

    SET_TABLE_MX
ALIGN
E1_PLP:
    movw %bx,_PC
    E1_PULL_B
    E1_RESTOREFLAGS

    JUMP_FLAG SNES_FLAG_I,CPU_RETURN
    movb IRQ_pin,%al
    testb %al,%al
    jz  CPU_RETURN
    pushl $CPU_RETURN
    jmp E1_IRQ

/* 29 */

ALIGN
EM_AND_i:
    EM_AND Immediate
    jmp CPU_RETURN
ALIGN
E0_AND_i:
    E0_AND Immediate
    jmp CPU_RETURN

/* 2A */

ALIGN
EM_RLA:
    movw %bx,_PC
    movb _A,%al
    MAKE_CARRY
    adcb %al,%al
    movb %al,_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN

ALIGN
E0_RLA:
    movw %bx,_PC
    MAKE_CARRY
    movl _A,%eax
    adcw %ax,%ax
    movl %eax,_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN

/* 2B */

ALIGN
E0_PLD:
    movw %bx,_PC
    E0_PULL_W
    movw %ax,_D
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
ALIGN
E1_PLD:
    movw %bx,_PC
    E1_PULL_W
    movw %ax,_D
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN

/* 2C */

ALIGN
EM_BIT_a:
    EM_BIT Absolute
    jmp CPU_RETURN
ALIGN
E0_BIT_a:
    E0_BIT Absolute
    jmp CPU_RETURN

/* 2D */

ALIGN
EM_AND_a:
    EM_AND Absolute
    jmp CPU_RETURN
ALIGN
E0_AND_a:
    E0_AND Absolute
    jmp CPU_RETURN

/* 2E */

ALIGN
EM_ROL_a:
    EM_ROL Absolute
    jmp CPU_RETURN
ALIGN
E0_ROL_a:
    E0_ROL Absolute
    jmp CPU_RETURN

/* 2F */

ALIGN
EM_AND_al:
    EM_AND Absolute_Long
    jmp CPU_RETURN
ALIGN
E0_AND_al:
    E0_AND Absolute_Long
    jmp CPU_RETURN
