/*

 65c816 opcodes 0x40-0x4F:
    40: RTI s       1,7
    41: EOR (d,x)   2,6
    42: WDM         2,2
    43: EOR d,s     2,4
    44: MVP xya     3,7
    45: EOR d       2,3
    46: LSR d       2,5
    47: EOR [d]     2,6
    48: PHA s       1,3
    49: EOR #       2,2
    4A: LSR a       1,2
    4B: PHK s       1,3
    4C: JMP a       3,3
    4D: EOR a       3,4
    4E: LSR a       3,6
    4F: EOR al      4,5
*/

/* 40 */

ALIGN
E0_RTI:
    E0_PULL_B
    E0_RESTOREFLAGS
    E0_PULL_B
    movb %al,_PC
    E0_PULL_B
    movb %al,%ah
    movb %al,_PC+1
    E0_PULL_B
    xorl %ebx,%ebx
    movb %al,_PB
    movb %al,%bl
    testb %al,MEMSEL    // Check bus speed against PB
    movb $5,%dl
    jnz 0f
    movb $4,%dl
0:  addb %ah,%ah
    movb %dl,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset

    JUMP_FLAG SNES_FLAG_I,0f
    movb IRQ_pin,%al
    testb %al,%al
    jz  0f
    call E0_IRQ
0:

    SET_TABLE_MX

ALIGN
E1_RTI:
    E1_PULL_B
    E1_RESTOREFLAGS
    SET_FLAG(SNES_FLAG_B)   // Set break flag
    E1_PULL_B
    movb %al,_PC
    E1_PULL_B
    movb %al,_PC+1
    xorl %ebx,%ebx
    addb %al,%al
    movb _PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset

    JUMP_FLAG SNES_FLAG_I,CPU_RETURN
    movb IRQ_pin,%al
    testb %al,%al
    jz  CPU_RETURN
    pushl $CPU_RETURN
    jmp E1_IRQ

/* 41 */

ALIGN
EM_EOR_Od_xO:
    EM_EOR Direct_Indexed_Indirect
    jmp CPU_RETURN
ALIGN
E0_EOR_Od_xO:
    E0_EOR Direct_Indexed_Indirect
    jmp CPU_RETURN

/* 42 */
ALIGN
ALL_WDM:
    incl %ebx
    movw %bx,_PC
    jmp CPU_RETURN

/* 43 */

ALIGN
EM_EOR_d_s:
    EM_EOR Stack_Relative
    jmp CPU_RETURN
ALIGN
E0_EOR_d_s:
    E0_EOR Stack_Relative
    jmp CPU_RETURN

/* 44 */

ALIGN
E0_MVP:
    movb 1(%esi),%al    // Dest bank
    addl $2,%ebx
    movb %al,_DB
    movb 2(%esi),%al    // Src bank
    movw %bx,_PC
    movl _DB_Shifted,%edx
    shll $16,%eax
    movw _Y,%dx         // Dest Address
    movw _X,%ax         // Src Address
    movl %eax,%ebx
    decl %eax
    movw %ax,_X
    GET_BYTE
    movl %edx,%ebx
    decl %edx
    movw %dx,_Y
    SET_BYTE
    movl _A,%eax
    testl %eax,%eax
    jz 0f
    subw $3,_PC
0:  decl %eax
    movw %ax,_A
    jmp CPU_RETURN
	
ALIGN
EX_MVP:
    movb 1(%esi),%al    // Dest bank
    addl $2,%ebx
    movb %al,_DB
    movb 2(%esi),%al    // Src bank
    movw %bx,_PC
    movl _DB_Shifted,%edx
    shll $16,%eax
    movb _Y,%dl         // Dest Address
    movb _X,%al         // Src Address
    movl %eax,%ebx
    decl %eax
    movb %al,_X
    GET_BYTE
    movl %edx,%ebx
    decl %edx
    movb %dl,_Y
    SET_BYTE
    movl _A,%eax
    testl %eax,%eax
    jz 0f
    subw $3,_PC
0:  decl %eax
    movw %ax,_A
    jmp CPU_RETURN

/* 45 */

ALIGN
EM_EOR_d:
    EM_EOR Direct
    jmp CPU_RETURN
ALIGN
E0_EOR_d:
    E0_EOR Direct
    jmp CPU_RETURN

/* 46 */

ALIGN
EM_LSR_d:
    EM_LSR Direct
    jmp CPU_RETURN
ALIGN
E0_LSR_d:
    E0_LSR Direct
    jmp CPU_RETURN

/* 47 */

ALIGN
EM_EOR_IdI:
    EM_EOR Direct_Indirect_Long
    jmp CPU_RETURN
ALIGN
E0_EOR_IdI:
    E0_EOR Direct_Indirect_Long
    jmp CPU_RETURN

/* 48 */

ALIGN
E0_PHA:
    movw %bx,_PC
    movw _A,%ax
    E0_PUSH_W
    jmp CPU_RETURN

ALIGN
EM_PHA:
    movw %bx,_PC
    movb _A,%al
    E0_PUSH_B
    jmp CPU_RETURN

ALIGN
E1_PHA:
    movw %bx,_PC
    movb _A,%al
    E1_PUSH_B
    jmp CPU_RETURN

/* 49 */

ALIGN
EM_EOR_i:
    EM_EOR Immediate
    jmp CPU_RETURN
ALIGN
E0_EOR_i:
    E0_EOR Immediate
    jmp CPU_RETURN

/* 4A */

ALIGN
EM_SRA:
    movw %bx,_PC
    shrb $1,_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN
ALIGN
E0_SRA:
    movw %bx,_PC
    shrw $1,_A
    STORE_FLAGS (SNES_FLAG_NZC)
    jmp CPU_RETURN

/* 4B */

ALIGN
E0_PHK:				/* Push PB(K) onto stack */
    movw %bx,_PC
    movb _PB,%al
    E0_PUSH_B
    jmp CPU_RETURN

ALIGN
E1_PHK:
    movw %bx,_PC
    movb _PB,%al
    E1_PUSH_B
    jmp CPU_RETURN

/* 4C */

ALIGN
ALL_JMP_a:
    ADDR_Absolute_JMP
    movb _PC+1,%al
    xorl %ebx,%ebx
    addb %al,%al
    movb _PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

/* 4D */

ALIGN
EM_EOR_a:
    EM_EOR Absolute
    jmp CPU_RETURN
ALIGN
E0_EOR_a:
    E0_EOR Absolute
    jmp CPU_RETURN

/* 4E */

ALIGN
EM_LSR_a:
    EM_LSR Absolute
    jmp CPU_RETURN
ALIGN
E0_LSR_a:
    E0_LSR Absolute
    jmp CPU_RETURN

/* 4F */

ALIGN
EM_EOR_al:
    EM_EOR Absolute_Long
    jmp CPU_RETURN
ALIGN
E0_EOR_al:
    E0_EOR Absolute_Long
    jmp CPU_RETURN
