/*

 65c816 opcodes 0x60-0x6F:
    60: RTS s       1,6
    61: ADC (d,x)   2,6
    62: PER s       3,6
    63: ADC d,s     2,4
    64: STZ d       2,3
    65: ADC d       2,3
    66: ROR d       2,5
    67: ADC [d]     2,6
    68: PLA s       1,4
    69: ADC #       2,2
    6A: ROR a       1,2
    6B: RTL s       1,6
    6C: JMP (a)     3,5
    6D: ADC a       3,4
    6E: ROR a       3,6
    6F: ADC al      4,5
*/

/* 60 */

ALIGN
E0_RTS:
    E0_PULL_W
    incl %eax               // This is a MAD processor!
    movw %ax,_PC
    xorl %ebx,%ebx
    addb %ah,%ah
    movb _PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN
ALIGN
E1_RTS:
    E1_PULL_W
    incl %eax               // This is a MAD processor!
    movw %ax,_PC
    xorl %ebx,%ebx
    addb %ah,%ah
    movb _PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

/* 61 */

ALIGN
EM_ADC_Od_xO:
    EM_ADC Direct_Indexed_Indirect
    jmp CPU_RETURN
ALIGN
E0_ADC_Od_xO:
    E0_ADC Direct_Indexed_Indirect
    jmp CPU_RETURN

/* 62 */

ALIGN
E0_PER:
    addl $2,%ebx
    movb 1(%esi),%al
    movw %bx,_PC
    movb 2(%esi),%ah
    addl %ebx,%eax
	E0_PUSH_W
    jmp CPU_RETURN

ALIGN
E1_PER:
    addl $2,%ebx
    movb 1(%esi),%al
    movw %bx,_PC
    movb 2(%esi),%ah
    addl %ebx,%eax
	E1_PUSH_W
    jmp CPU_RETURN

/* 63 */

ALIGN
EM_ADC_d_s:
    EM_ADC Stack_Relative
    jmp CPU_RETURN
ALIGN
E0_ADC_d_s:
    E0_ADC Stack_Relative
    jmp CPU_RETURN

/* 64 */

ALIGN
EM_STZ_d:
    EM_STZ Direct
    jmp CPU_RETURN
ALIGN
E0_STZ_d:
    E0_STZ Direct
    jmp CPU_RETURN

/* 65 */

ALIGN
EM_ADC_d:
    EM_ADC Direct
    jmp CPU_RETURN
ALIGN
E0_ADC_d:
    E0_ADC Direct
    jmp CPU_RETURN

/* 66 */

ALIGN
EM_ROR_d:
    EM_ROR Direct
    jmp CPU_RETURN
ALIGN
E0_ROR_d:
    E0_ROR Direct
    jmp CPU_RETURN

/* 67 */

ALIGN
EM_ADC_IdI:
    EM_ADC Direct_Indirect_Long
    jmp CPU_RETURN
ALIGN
E0_ADC_IdI:
    E0_ADC Direct_Indirect_Long
    jmp CPU_RETURN

/* 68 */

ALIGN
E0_PLA:
    movw %bx,_PC
    E0_PULL_W
    movw %ax,_A
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
ALIGN
EM_PLA:
    movw %bx,_PC
    E0_PULL_B
    movb %al,_A
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN
ALIGN
E1_PLA:
    movw %bx,_PC
    E1_PULL_B
    movb %al,_A
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    jmp CPU_RETURN

/* 69 */

ALIGN
EM_ADC_i:
    EM_ADC Immediate
    jmp CPU_RETURN
ALIGN
E0_ADC_i:
    E0_ADC Immediate
    jmp CPU_RETURN

/* 6A */    // Fixed 0.25: RRA8 didn't set NZ flags correctly

ALIGN
EM_RRA:
    shrb $1,_P_B    //MAKE_CARRY
    movb _A,%al
    movw %bx,_PC
    rcrb $1,%al
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    movb %al,_A
    jmp CPU_RETURN

ALIGN
E0_RRA:
    shrb $1,_P_B    //MAKE_CARRY
    movl _A,%edx
    movw %bx,_PC
    rcrw $1,%dx
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testw %dx,%dx
    STORE_FLAGS (SNES_FLAG_NZ)
    movl %edx,_A
    jmp CPU_RETURN

/* 6B */

ALIGN
E0_RTL:
    E0_PULL_W
    incl %eax               // This is a MAD processor!
    movw %ax,_PC
    E0_PULL_B
    xorl %ebx,%ebx
    movb %al,_PB
    movb %al,%bl
    testb %al,MEMSEL    // Check bus speed against PB
    movb $5,%dl
    jnz 0f
    movb $4,%dl
0:  addb %ah,%ah
    movb %dl,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN
ALIGN
E1_RTL:
    E1_PULL_W
    incl %eax               // This is a MAD processor!
    movw %ax,_PC
    E1_PULL_B
    xorl %ebx,%ebx
    addb %ah,%ah
    movb %al,%bl
    movb %al,_PB
    testb %al,MEMSEL    // Check bus speed against PB
    movb $5,%dl
    jnz 0f
    movb $4,%dl
0:  movb %dl,CycleTable // Update bus speed
    movl _Read_BankMapping(,%ebx,4),%eax
    adcl %ebx,%ebx
    movl %eax,_PB_Read
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

/* 6C */

/*
 JMP Absolute Indirect
 Returns PCH in %al
  PC = (a16)
*/
ALIGN
ALL_JMP_OaO:
    movb 1(%esi),%al
    movb 2(%esi),%ah
    movl %eax,%ebx
    GET_BYTE
    incl %ebx
    movb %al,_PC
    GET_BYTE
    movb %al,_PC+1
    xorl %ebx,%ebx
    addb %al,%al
    movb _PB,%bl
    adcl %ebx,%ebx
    movl _Read_BankOffset(,%ebx,4),%eax
    movl %eax,_PBOffset
    jmp CPU_RETURN

/* 6D */

ALIGN
EM_ADC_a:
    EM_ADC Absolute
    jmp CPU_RETURN
ALIGN
E0_ADC_a:
    E0_ADC Absolute
    jmp CPU_RETURN

/* 6E */

ALIGN
EM_ROR_a:
    EM_ROR Absolute
    jmp CPU_RETURN
ALIGN
E0_ROR_a:
    E0_ROR Absolute
    jmp CPU_RETURN

/* 6F */

ALIGN
EM_ADC_al:
    EM_ADC Absolute_Long
    jmp CPU_RETURN
ALIGN
E0_ADC_al:
    E0_ADC Absolute_Long
    jmp CPU_RETURN
