/*
 SNEeSe 65c816 CPU emulation core
 Originally written by Savoury SnaX (Not quite sure if I like AT&T)
 Maintained/rewritten by Charles Bilyue'

 Compile under DJGPP (GNU-AS)
  NOTE - This must have .S (capital) or the # stuff wont be pre-processed!!!

 This file contains:
  CPU opcode macros

 Examples of usage:
  EM_LDA Direct = 8-bit LDA using "direct" address mode
  BFC SNES_FLAG_C = Branch if carry flag clear (BCC)

*/

.macro EM_ADC addr
    READ8_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_CARRY
    adcb %al,_A
    lahf
    jno 2f
    jmp 1f

ALIGN
0:  MAKE_CARRY
    adcb _A,%al
    daa
    movb %al,_A
    lahf
    jno 2f
1:  SET_FLAG(SNES_FLAG_V)
2:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
.endm

.macro E0_ADC addr
    READ16_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_CARRY
    adcw %ax,_A
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp 2f

ALIGN
0:  MAKE_CARRY
    movl %eax,%edx
    movl _A,%eax
    adcb %dl,%al
    daa
    movb %al,%dl
    movb %ah,%al
    adcb %dh,%al
    daa
    movb %al,%dh
    movw %dx,_A
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_C,%ah
    testw %dx,%dx
    STORE_FLAGS(SNES_FLAG_NZ)
2:
.endm

.macro	EM_AND addr
    READ8_\addr
    andb %al,_A
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro E0_AND addr
    READ16_\addr
    andw %ax,_A
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro	EM_ASL addr
    READ8_\addr
    addb %al,%al
    STORE_FLAGS (SNES_FLAG_NZC)
    SET_BYTE
.endm

.macro	E0_ASL addr
    READ16_\addr
    movl %eax,%edx
    addw %dx,%dx
    STORE_FLAGS (SNES_FLAG_NZC)
    movl %edx,%eax
    SET_WORD_HL
.endm

.macro BFC flag
    JUMP_FLAG \flag,0f          // flag set
    movsbl 1(%esi),%eax // sign extend for addition
    addl $6,%ebp                // branch taken
    addl %eax,%ebx
0:  incl %ebx
    movw %bx,_PC
.endm

.macro BFS flag
    JUMP_NOT_FLAG \flag,0f      // flag clear
    movsbl 1(%esi),%eax // sign extend for addition
    addl $6,%ebp                // branch taken
    addl %eax,%ebx
0:  incl %ebx
    movw %bx,_PC
.endm

.macro	EM_BIT addr
    READ8_\addr
    CLR_FLAG(SNES_FLAG_N+SNES_FLAG_V+SNES_FLAG_Z)
    testb %al,%al
    jns 1f
    SET_FLAG(SNES_FLAG_N)
1:  testb $0x40,%al
    jz  1f
    SET_FLAG(SNES_FLAG_V)
1:  testb _A,%al
    jnz 1f
    SET_FLAG(SNES_FLAG_Z)
1:
.endm

.macro	E0_BIT addr
    READ16_\addr
    CLR_FLAG(SNES_FLAG_N+SNES_FLAG_V+SNES_FLAG_Z)
    testb %ah,%ah
    jns 1f
    SET_FLAG(SNES_FLAG_N)
1:  testb $0x40,%ah
    jz  1f
    SET_FLAG(SNES_FLAG_V)
1:  testl _A,%eax
    jnz 1f
    SET_FLAG(SNES_FLAG_Z)
1:
.endm

.macro EM_CMP addr
    READ8_\addr
    cmpb %al,_A
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
.endm

.macro E0_CMP addr
    READ16_\addr
    cmpw %ax,_A
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
.endm

.macro EX_CPX addr
    READ8_\addr
    cmpb %al,_X
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
.endm

.macro E0_CPX addr
    READ16_\addr
    cmpw %ax,_X
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
.endm

.macro EX_CPY addr
    READ8_\addr
    cmpb %al,_Y
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
.endm

.macro E0_CPY addr
    READ16_\addr
    cmpw %ax,_Y
    cmc
    STORE_FLAGS (SNES_FLAG_NZC)
.endm

.macro EM_DEC addr
    READ8_\addr
    decl %eax
    SET_BYTE
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro E0_DEC addr
    READ16_\addr
    decl %eax
    SET_WORD_HL
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro EM_EOR addr
    READ8_\addr
    xorb %al,_A
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro E0_EOR addr
    READ16_\addr
    xorw %ax,_A
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro	EM_INC addr
    READ8_\addr
    incl %eax
    SET_BYTE
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro	E0_INC addr
    READ16_\addr
    incl %eax
    SET_WORD_HL
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro EM_LDA addr
    READ8_\addr
    movb %al,_A
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro E0_LDA addr
    READ16_\addr
    movw %ax,_A
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro EX_LDX addr
    READ8_\addr
    movb %al,_X
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro E0_LDX addr
    READ16_\addr
    movw %ax,_X
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro EX_LDY addr
    READ8_\addr
    movb %al,_Y
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro E0_LDY addr
    READ16_\addr
    movw %ax,_Y
    testw %ax,%ax
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro	EM_LSR addr
    READ8_\addr
    shrb $1,%al
    STORE_FLAGS (SNES_FLAG_NZC)
    SET_BYTE
.endm

.macro	E0_LSR addr
    READ16_\addr
    movl %eax,%edx
    shrw $1,%dx
    STORE_FLAGS (SNES_FLAG_NZC)
    movl %edx,%eax
    SET_WORD_HL
.endm

.macro	EM_ORA addr
    READ8_\addr
    orb %al,_A
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro	E0_ORA addr
    READ16_\addr
    orw %ax,_A
    STORE_FLAGS (SNES_FLAG_NZ)
.endm

.macro	EM_ROL addr
    READ8_\addr
    MAKE_CARRY
    adcb %al,%al
    STORE_FLAGS (SNES_FLAG_NZC)
    SET_BYTE
.endm

.macro	E0_ROL addr
    READ16_\addr
    movl %eax,%edx
    MAKE_CARRY
    adcw %dx,%dx
    STORE_FLAGS (SNES_FLAG_NZC)
    movl %edx,%eax
    SET_WORD_HL
.endm

.macro	EM_ROR addr
    READ8_\addr
    shrb $1,_P_B    //MAKE_CARRY
    rcrb $1,%al
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testb %al,%al
    STORE_FLAGS (SNES_FLAG_NZ)
    SET_BYTE
.endm

.macro	E0_ROR addr
    READ16_\addr
    movl %eax,%edx
    shrb $1,_P_B    //MAKE_CARRY
    rcrw $1,%dx
    adcb _P_B,_P_B  //STORE_FLAGS (SNES_FLAG_C)
    testw %dx,%dx
    STORE_FLAGS (SNES_FLAG_NZ)
    movl %edx,%eax
    SET_WORD_HL
.endm

.macro EM_SBC addr
    READ8_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_NOT_CARRY
    sbbb %al,_A
    cmc
    lahf
    jno 2f
    jmp 1f

ALIGN
0:  MAKE_NOT_CARRY
    movb _A,%ah
    sbbb %al,%ah
    movb %ah,%al
    das
    movb %al,_A
    cmc
    lahf
    jno 2f
1:  SET_FLAG(SNES_FLAG_V)
2:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
.endm

.macro E0_SBC addr
    READ16_\addr
    CLR_FLAG(SNES_FLAG_V)
    JUMP_FLAG SNES_FLAG_D,0f
    MAKE_NOT_CARRY
    sbbw %ax,_A
    cmc
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_NZC,%ah
    jmp 2f

ALIGN
0:  MAKE_NOT_CARRY
    movl %eax,%edx
    movl _A,%eax
    sbbb %dl,%al
    das
    movb %al,%dl
    movb %ah,%al
    sbbb %dh,%al
    das
    movb %al,%dh
    movw %dx,_A
    cmc
    lahf
    jno 1f
    SET_FLAG(SNES_FLAG_V)
1:  STORE_FLAGS_FROM SNES_FLAG_C,%ah
    testw %dx,%dx
    STORE_FLAGS(SNES_FLAG_NZ)
2:
.endm

.macro EM_STA addr
    ADDR_\addr
    movb _A,%al
    SET_BYTE
.endm

.macro E0_STA addr
    ADDR_\addr
    movl _A,%eax
    SET_WORD NoSave=1
.endm

.macro EX_STX addr
    ADDR_\addr
    movb _X,%al
    SET_BYTE
.endm

.macro E0_STX addr
    ADDR_\addr
    movl _X,%eax
    SET_WORD NoSave=1
.endm

.macro EX_STY addr
    ADDR_\addr
    movb _Y,%al
    SET_BYTE
.endm

.macro E0_STY addr
    ADDR_\addr
    movl _Y,%eax
    SET_WORD NoSave=1
.endm

.macro EM_STZ addr
    ADDR_\addr
    xorb %al,%al
    SET_BYTE
.endm

.macro E0_STZ addr
    ADDR_\addr
    xorb %al,%al
    SET_WORD Double=1
.endm

// TRB set flag Z on A & Memory Operand not A & !Memory Operand!

.macro EM_TRB addr      // Test and Reset bit
    READ8_\addr
    movb _A,%dl
    testb %al,%dl
    STORE_FLAGS (SNES_FLAG_Z)
    notb %dl
    andb %dl,%al
    SET_BYTE
.endm

.macro	E0_TRB addr
    READ16_\addr
    movl %eax,%edx
    testl _A,%edx
    STORE_FLAGS (SNES_FLAG_Z)
    movl _A,%eax
    notl %eax
    andl %edx,%eax
    SET_WORD_HL
.endm

// TSB set flag Z on A & Memory Operand not A | Memory Operand!

.macro EM_TSB addr      // Test and Set bit
    READ8_\addr
    movb _A,%dl
    testb %al,%dl
    STORE_FLAGS (SNES_FLAG_Z)
    orb %dl,%al
    SET_BYTE
.endm

.macro E0_TSB addr
    READ16_\addr
    movl %eax,%edx
    testl _A,%edx
    STORE_FLAGS (SNES_FLAG_Z)
    movl _A,%eax
    orl %edx,%eax
    SET_WORD_HL
.endm
