/*

 (H)DMA macros

 Thanks to zsKnight for all the information which helped get this
 working right! (I hope!)

*/

// Offsets from TableDMA#
#define DMAP 0
#define BBAD 1
#define DMA_Vid BBAD
#define NTRL 2
#define A1T 4
#define A1TL A1T
#define A1TH (A1T+1)
#define A1B (A1T+2)
#define DAS 8
#define DASL DAS
#define DASH (DAS+1)
#define DASB (DAS+2)
#define A2T 0x0C
#define A2L A2T
#define A2H (A2T+1)
#define A2B (A2T+2)
#define HDMA_Wr0 0x10
#define HDMA_Wr1 0x14
#define HDMA_Wr2 0x18
#define HDMA_Wr3 0x1C
#define HDMA_Siz 0x20
#define HDMA_Vid 0x24
#define DMATableSize 0x28

.macro MAP_READ_DMA_LIST num
  .long   MAP_READ_DMAP\num
  .long   MAP_READ_BBAD\num
  .long   MAP_READ_A1TL\num
  .long   MAP_READ_A1TH\num
  .long   MAP_READ_A1B\num
  .long   MAP_READ_DASL\num
  .long   MAP_READ_DASH\num
  .long   MAP_READ_DASB\num
  .long   MAP_READ_A2L\num
  .long   MAP_READ_A2H\num
  .long   MAP_READ_NTRL\num
  DUPLICATE long,5,UNSUPPORTED_READ
.endm

.macro MAP_WRITE_DMA_LIST num
  .long   MAP_WRITE_DMAP\num
  .long   MAP_WRITE_BBAD\num
  .long   MAP_WRITE_A1TL\num
  .long   MAP_WRITE_A1TH\num
  .long   MAP_WRITE_A1B\num
  .long   MAP_WRITE_DASL\num
  .long   MAP_WRITE_DASH\num
  .long   MAP_WRITE_DASB\num
  .long   MAP_WRITE_A2L\num
  .long   MAP_WRITE_A2H\num
  .long   MAP_WRITE_NTRL\num
  DUPLICATE long,5,UNSUPPORTED_WRITE
.endm

.macro LOAD_DMA_TABLE num
    movl $TableDMA\num,%edx
.endm

.macro HDMAOPERATION num
  movb HDMAON,%al
  testb $(1<<\num),%al
  jz  8f

  LOAD_DMA_TABLE \num

  call Do_HDMA_Channel  // CF clear if channel disabled

  jc  8f
  movb HDMAON,%al
  andb $~(1<<\num),%al
  movb %al,HDMAON       // Disable this channel
8:
.endm

.macro RELATCH_HDMA num
    testb $(1<<\num),HDMAEN
    jz  8f
    movb $0,NTRL_\num

#ifdef DEBUG
#ifdef TRACKERS
    movl _LastHDMA,%ebx     //
    addl _HDMAAddress,%ebx  //
#endif
#endif
    movl A1T_\num,%eax   // Src Address in ebx
    movl %eax,A2T_\num
#ifdef DEBUG
#ifdef TRACKERS
    bswapl %eax             //
    movl %eax,4(%ebx)       //
    movb DASB_\num,%al      //
    movb %al,2(%ebx)        //
#endif
#endif

    movb DMAP_\num,%al
#ifdef DEBUG
#ifdef TRACKERS
    movb %al,(%ebx)         //
#endif
#endif
    andl $7,%eax
    movb HDMA_Size(%eax),%dl
    movb %dl,HDMA_Siz_\num

    xorl %edx,%edx
    leal DMA_PPU_Order(,%eax,4),%eax
    movb DMA_Vid_\num,%dl   // PPU address in edx
    movl %edx,%ecx          // PPU address in ecx
#ifdef DEBUG
#ifdef TRACKERS
    movb %dl,1(%ebx)        //
#endif
#endif
    addb (%eax),%dl
    movl Write_Map_21(,%edx,4),%edx
    movl %edx,HDMA_Wr0_\num
    movl %ecx,%edx          // PPU address in edx
    addb 1(%eax),%cl
    movl Write_Map_21(,%ecx,4),%ecx
    movl %ecx,HDMA_Wr1_\num
    movl %edx,%ecx          // PPU address in ecx
    addb 2(%eax),%dl
    movl Write_Map_21(,%edx,4),%edx
    movl %edx,HDMA_Wr2_\num
    addb 3(%eax),%cl
    movl Write_Map_21(,%ecx,4),%ecx
    movl %ecx,HDMA_Wr3_\num
#ifdef DEBUG
#ifdef TRACKERS
    movl _LastHDMA,%ebx     //
    addl $8,%ebx            //
    andl $0x3FF,%ebx        //
    movl %ebx,_LastHDMA     //
#endif
#endif
8:
.endm

.macro RELATCH_HDMA_IN_FRAME num
    movb HDMAON,%al
    notb %al
    andb (%esp),%al
    testb $(1<<\num),%al
    jz  8f

    movb DMAP_\num,%al
    andl $7,%eax
    movb HDMA_Size(%eax),%dl
    movb %dl,HDMA_Siz_\num

    xorl %edx,%edx
    leal DMA_PPU_Order(,%eax,4),%eax
    movb DMA_Vid_\num,%dl   // PPU address in edx
    movl %edx,%ecx          // PPU address in ecx
    addb (%eax),%dl
    movl Write_Map_21(,%edx,4),%edx
    movl %edx,HDMA_Wr0_\num
    movl %ecx,%edx          // PPU address in edx
    addb 1(%eax),%cl
    movl Write_Map_21(,%ecx,4),%ecx
    movl %ecx,HDMA_Wr1_\num
    movl %edx,%ecx          // PPU address in ecx
    addb 2(%eax),%dl
    movl Write_Map_21(,%edx,4),%edx
    movl %edx,HDMA_Wr2_\num
    addb 3(%eax),%cl
    movl Write_Map_21(,%ecx,4),%ecx
    movl %ecx,HDMA_Wr3_\num
8:
.endm

.macro RESTORE_HDMA_ADDRESS
    movb HDMAEN,%al
    movb %al,HDMAON
    testb %al,%al
    jz  9f
    RELATCH_HDMA 0
    RELATCH_HDMA 1
    RELATCH_HDMA 2
    RELATCH_HDMA 3
    RELATCH_HDMA 4
    RELATCH_HDMA 5
    RELATCH_HDMA 6
    RELATCH_HDMA 7
9:
.endm
