/*

 DMA.S - (H)DMA emulation

*/

#include "misc.i"
#include "DMA.i"

EXPORT_C DMA_start

.data
ALIGN
.globl DMA_PPU_Order
DMA_PPU_Order:
.byte   0,0,0,0
.byte   0,1,0,1
.byte   0,0,0,0
.byte   0,0,1,1
.byte   0,1,2,3
.byte   0,1,0,1
.byte   0,0,0,0
.byte   0,0,1,1

.globl HDMA_Size
HDMA_Size:
.byte   1,2,2,4
.byte   4,2,2,4

.bss
ALIGN
.globl DMA_Access0,DMA_Access1,DMA_Access2,DMA_Access3,DMA_CPUBank
DMA_Access0:skipl   // Pointer for DMA PPU read/write
DMA_Access1:skipl   // Pointer for DMA PPU read/write
DMA_Access2:skipl   // Pointer for DMA PPU read/write
DMA_Access3:skipl   // Pointer for DMA PPU read/write
DMA_CPUBank:skipl   // Pointer for DMA CPU read/write

.macro DMA_DATA num
ALIGN
.globl TableDMA\num
TableDMA\num:
.globl DMAP_\num,BBAD_\num,DMA_Vid_\num,NTRL_\num,HDMA_On_\num
// dh0bcttt d=CPU->PPU,h=addr,b=inc/dec,c=inc/fixed,t=type
DMAP_\num:      skipb
BBAD_\num:      // Low byte of 0x21??
DMA_Vid_\num:   skipb
NTRL_\num:      skipb
                skipb
.globl A1T_\num,A1TL_\num,A1TH_\num,A1B_\num
A1T_\num:               // Source Address L/H/B
A1TL_\num:      skipb   // Source Address L
A1TH_\num:      skipb   // Source Address H
A1B_\num:       skipb   // Source Bank Address
                skipb
.globl DAS_\num,DASL_\num,DASH_\num,DASB_\num
DAS_\num:               // Data Size L/H
DASL_\num:      skipb   // Data Size L
DASH_\num:      skipb   // Data Size H
DASB_\num:      skipb   // Data address bank
                skipb
.globl A2T_\num,A2L_\num,A2H_\num,A2B_\num
A2T_\num:
A2L_\num:       skipb   // HDMA table address L
A2H_\num:       skipb   // HDMA table address H
A2B_\num:       skipb   // HDMA table bank address
                skipb

// HDMA_Wr#_x - These hold the write handlers for HDMA
// HDMA_Siz_x - These hold the register size for HDMA
// HDMA_Vid_x - These hold the PPU base address
.globl HDMA_Wr0_\num,HDMA_Wr1_\num,HDMA_Wr2_\num,HDMA_Wr3_\num
.globl HDMA_Siz_\num,HDMA_Vid_\num
ALIGN
HDMA_Wr0_\num:  skipl
HDMA_Wr1_\num:  skipl
HDMA_Wr2_\num:  skipl
HDMA_Wr3_\num:  skipl
HDMA_Siz_\num:  skipl
HDMA_Vid_\num:  skipl
.endm

DMA_DATA 0
DMA_DATA 1
DMA_DATA 2
DMA_DATA 3
DMA_DATA 4
DMA_DATA 5
DMA_DATA 6
DMA_DATA 7

.globl MDMAEN
MDMAEN: skipb   // DMA enable
.globl HDMAEN
HDMAEN: skipb   // HDMA enable
.globl HDMAON
HDMAON: skipb   // HDMA enabled this refresh

.text
ALIGN
Do_DMA_Channel:
  movl A1T(%edx),%ebx   // CPU address in ebx
  xorl %esi,%esi        // Set address increment to 0

  xorl %ebp,%ebp
  movw DAS(%edx),%bp
  decl %ebp
  andl $0xFFFF,%ebp
  incl %ebp

0:movb DMAP(%edx),%al
  testb $8,%al      // Does the operation require address adjustment?
  jnz 1f

  decl %esi         // Set address decrement
  testb $0x10,%al
  jnz 1f

  addl $2,%esi      // Set address increment

1:movl $0x2100,%ecx
  testb %al,%al     // Is the operation CPU->PPU?
  jns 0f

// PPU->CPU
  pushl %edx
  andl $7,%eax
  leal DMA_PPU_Order(,%eax,4),%eax
  movb DMA_Vid(%edx),%cl    // PPU address in edx
  movl %ecx,%edx            // PPU address in ecx
  addb (%eax),%cl
  movl Read_Map_20_5F-0x2000*4(,%ecx,4),%ecx
  movl %ecx,DMA_Access0
  movl %edx,%ecx            // PPU address in ecx
  addb 1(%eax),%dl
  movl Read_Map_20_5F-0x2000*4(,%edx,4),%edx
  movl %edx,DMA_Access1
  movl %ecx,%edx            // PPU address in edx
  addb 2(%eax),%cl
  movl Read_Map_20_5F-0x2000*4(,%ecx,4),%ecx
  movl %ecx,DMA_Access2
  addb 3(%eax),%dl
  movl Read_Map_20_5F-0x2000*4(,%edx,4),%edx
  movl %edx,DMA_Access3
  xorl %eax,%eax
  popl %edx
  movb A1B(%edx),%al
  movl _Write_BankMapping(,%eax,4),%eax
  movl %eax,DMA_CPUBank
1:call *DMA_Access0
  call *DMA_CPUBank
  addw %si,%bx
  decl %ebp
  jz  1f

  call *DMA_Access1
  call *DMA_CPUBank
  addw %si,%bx
  decl %ebp
  jz  1f

  call *DMA_Access2
  call *DMA_CPUBank
  addw %si,%bx
  decl %ebp
  jz  1f

  call *DMA_Access3
  call *DMA_CPUBank
  addw %si,%bx
  decl %ebp
  jnz 1b
1:movw %bx,A1T(%edx)     // v0.15 forgot to update DMA pointers!
  ret

ALIGN
// CPU->PPU
0:pushl %edx
  andl $7,%eax
  leal DMA_PPU_Order(,%eax,4),%eax
  movb DMA_Vid(%edx),%cl    // PPU address in edx
  movl %ecx,%edx            // PPU address in ecx
  addb (%eax),%cl
  movl Write_Map_20_5F-0x2000*4(,%ecx,4),%ecx
  movl %ecx,DMA_Access0
  movl %edx,%ecx            // PPU address in ecx
  addb 1(%eax),%dl
  movl Write_Map_20_5F-0x2000*4(,%edx,4),%edx
  movl %edx,DMA_Access1
  movl %ecx,%edx            // PPU address in edx
  addb 2(%eax),%cl
  movl Write_Map_20_5F-0x2000*4(,%ecx,4),%ecx
  movl %ecx,DMA_Access2
  addb 3(%eax),%dl
  movl Write_Map_20_5F-0x2000*4(,%edx,4),%edx
  movl %edx,DMA_Access3
  xorl %eax,%eax
  popl %edx
  movb A1B(%edx),%al
  movl _Read_BankMapping(,%eax,4),%eax
  movl %eax,DMA_CPUBank
0:call *DMA_CPUBank
  call *DMA_Access0
  addw %si,%bx
  decl %ebp
  jz  1f

  call *DMA_CPUBank
  call *DMA_Access1
  addw %si,%bx
  decl %ebp
  jz  1f

  call *DMA_CPUBank
  call *DMA_Access2
  addw %si,%bx
  decl %ebp
  jz  1f

  call *DMA_CPUBank
  call *DMA_Access3
  addw %si,%bx
  decl %ebp
  jnz 0b

1:movw %bx,A1T(%edx)     // v0.15 forgot to update DMA pointers!
  movw $0,DAS(%edx)
  ret

ALIGN
.globl Do_HDMA_Channel
Do_HDMA_Channel:
  movl A2T(%edx),%ebx     // Get table address
  movb DMAP(%edx),%al     // Get HDMA control byte
  testb $0x40,%al         // Check for indirect addressing
  movl HDMA_Siz(%edx),%ecx  // Get HDMA transfer size
  jnz Do_HDMA_Indirect

Do_HDMA_Absolute:
  xorl %eax,%eax
  movb A2B(%edx),%al
  movl _Read_BankMapping(,%eax,4),%eax
  movl %eax,DMA_CPUBank

  movb NTRL(%edx),%ah       // Get number of lines to transfer
  testb $0x7F,%ah           // Need new set?
  jz  HDMA_Absolute_Next_Set
  testb %ah,%ah
  js  2f
  jmp HDMA_Absolute_Continue

HDMA_Absolute_Next_Set:
  call *DMA_CPUBank
  incw %bx                // Adjust table address
  testb %al,%al           // Check for zero-length set
  movb %al,NTRL(%edx)     // Save length of set
  jz  0f
  movl %ebx,A2T(%edx)     // Save new table address
2:call *DMA_CPUBank       // Get HDMA data
  incw %bx                // Adjust temporary table pointer
  call *HDMA_Wr0(%edx)    // Write HDMA destination
  cmpb $2,%cl
  jb  1f

  call *DMA_CPUBank       // Get HDMA data
  incw %bx
  call *HDMA_Wr1(%edx)
  cmpb $4,%cl
  jb  1f

  call *DMA_CPUBank       // Get HDMA data
  incw %bx
  call *HDMA_Wr2(%edx)

  call *DMA_CPUBank       // Get HDMA data
  call *HDMA_Wr3(%edx)
1:addw %cx,A2T(%edx)
HDMA_Absolute_Continue:
  decb NTRL(%edx)
  stc
  ret

0:movl %ebx,A2T(%edx)
  clc
  ret

Do_HDMA_Indirect:
  movb NTRL(%edx),%ah     // Get number of lines to transfer
  testb $0x7F,%ah
  jz  HDMA_Indirect_Next_Set
  testb %ah,%ah
  js  2f
  jmp HDMA_Indirect_Continue

HDMA_Indirect_Next_Set:
  xorl %eax,%eax
  movb A2B(%edx),%al
  movl _Read_BankMapping(,%eax,4),%eax
  movl %eax,DMA_CPUBank
  call *DMA_CPUBank
  incw %bx
  movb %al,NTRL(%edx)
  testb %al,%al
  movb %al,%ah
  jz  0b
  call *DMA_CPUBank
  incw %bx
  movb %al,DASL(%edx)
  call *DMA_CPUBank
  incw %bx
  movb %al,DASH(%edx)
  movl %ebx,A2T(%edx)
2:xorl %eax,%eax
  movl DAS(%edx),%ebx
  movb DASB(%edx),%al
  movl _Read_BankMapping(,%eax,4),%eax
  movl %eax,DMA_CPUBank
  call *DMA_CPUBank       // Get HDMA data
  incw %bx
  call *HDMA_Wr0(%edx)
  cmpb $2,%cl
  jb  1f

  call *DMA_CPUBank       // Get HDMA data
  incw %bx
  call *HDMA_Wr1(%edx)
  cmpb $4,%cl
  jb  1f

  call *DMA_CPUBank       // Get HDMA data
  incw %bx
  call *HDMA_Wr2(%edx)

  call *DMA_CPUBank       // Get HDMA data
  call *HDMA_Wr3(%edx)
1:addw %cx,DAS(%edx)
HDMA_Indirect_Continue:
  decb NTRL(%edx)
  stc
  ret

.macro DMAOPERATION num
  movb MDMAEN,%al
  testb $(1<<\num),%al
  jz  7f

  LOAD_DMA_TABLE \num
  call Do_DMA_Channel
7:
.endm

ALIGN
.globl SNES_R420B
SNES_R420B: // MDMAEN
  movb $0,%al
  ret

ALIGN
.globl SNES_R420C
SNES_R420C: // HDMAEN
  movb HDMAON,%al
  ret

ALIGN
.globl SNES_W420B
SNES_W420B: // MDMAEN
  movb %al,MDMAEN
  pusha

  movb CPU_Executing,%al
  pushl %eax
  movb $0,CPU_Executing

  DMAOPERATION 0
  DMAOPERATION 1
  DMAOPERATION 2
  DMAOPERATION 3
  DMAOPERATION 4
  DMAOPERATION 5
  DMAOPERATION 6
  DMAOPERATION 7

  popl %eax
  movb %al,CPU_Executing

  popa
8:ret

ALIGN
.globl SNES_W420C
SNES_W420C: // HDMAEN   // Actually handled within screen core!
  movb %al,HDMAEN
//ret
  testb $0xC0,HVBJOY
  jz 0f
  pusha
  RELATCH_HDMA_IN_FRAME 0
  RELATCH_HDMA_IN_FRAME 1
  RELATCH_HDMA_IN_FRAME 2
  RELATCH_HDMA_IN_FRAME 3
  RELATCH_HDMA_IN_FRAME 4
  RELATCH_HDMA_IN_FRAME 5
  RELATCH_HDMA_IN_FRAME 6
  RELATCH_HDMA_IN_FRAME 7
  popa
  movb %al,HDMAON
0:ret

ALIGN
.globl do_HDMA
do_HDMA:
  movb HDMAON,%al
  testb %al,%al
  jz  9f
  pusha
  HDMAOPERATION 0
  HDMAOPERATION 1
  HDMAOPERATION 2
  HDMAOPERATION 3
  HDMAOPERATION 4
  HDMAOPERATION 5
  HDMAOPERATION 6
  HDMAOPERATION 7
  popa
9:ret

// Requires %eax to be 0!
.macro Reset_DMA_Channel num
  movb %al,DMAP_\num
  movb %al,BBAD_\num
  movb %al,NTRL_\num
  movl %eax,A1T_\num
  movl %eax,DAS_\num
  movl %eax,A2T_\num
.endm

.globl Reset_DMA
Reset_DMA:
  // Set eax to 0, as we're setting most everything to 0...
  xorl %eax,%eax

  movb %al,MDMAEN
  movb %al,HDMAEN
  movb %al,HDMAON

  Reset_DMA_Channel 0
  Reset_DMA_Channel 1
  Reset_DMA_Channel 2
  Reset_DMA_Channel 3
  Reset_DMA_Channel 4
  Reset_DMA_Channel 5
  Reset_DMA_Channel 6
  Reset_DMA_Channel 7

  movl %eax,DMA_Access0
  movl %eax,DMA_Access1
  movl %eax,DMA_Access2
  movl %eax,DMA_Access3
  movl %eax,DMA_CPUBank

  ret

// Read from 43xx handlers
.macro MAP_READ_DMA num
ALIGN
.globl MAP_READ_DMAP\num
MAP_READ_DMAP\num:
  movb DMAP_\num,%al
  ret

ALIGN
.globl MAP_READ_BBAD\num
MAP_READ_BBAD\num:
  movb BBAD_\num,%al
  ret

ALIGN
.globl MAP_READ_A1TL\num
MAP_READ_A1TL\num:
  movb A1TL_\num,%al
  ret

ALIGN
.globl MAP_READ_A1TH\num
MAP_READ_A1TH\num:
  movb A1TH_\num,%al
  ret

ALIGN
.globl MAP_READ_A1B\num
MAP_READ_A1B\num:
  movb A1B_\num,%al
  ret

ALIGN
.globl MAP_READ_DASL\num
MAP_READ_DASL\num:
  movb DASL_\num,%al
  ret

ALIGN
.globl MAP_READ_DASH\num
MAP_READ_DASH\num:
  movb DASH_\num,%al
  ret

ALIGN
.globl MAP_READ_DASB\num
MAP_READ_DASB\num:
  movb DASB_\num,%al
  ret

ALIGN
.globl MAP_READ_A2L\num
MAP_READ_A2L\num:
  movb A2L_\num,%al
  ret

ALIGN
.globl MAP_READ_A2H\num
MAP_READ_A2H\num:
  movb A2H_\num,%al
  ret

ALIGN
.globl MAP_READ_NTRL\num
MAP_READ_NTRL\num:
  movb NTRL_\num,%al
  ret
.endm

MAP_READ_DMA 0
MAP_READ_DMA 1
MAP_READ_DMA 2
MAP_READ_DMA 3
MAP_READ_DMA 4
MAP_READ_DMA 5
MAP_READ_DMA 6
MAP_READ_DMA 7

// Write to 43xx handlers
.macro MAP_WRITE_DMA num
ALIGN
.globl MAP_WRITE_DMAP\num
MAP_WRITE_DMAP\num:
  movb %al,DMAP_\num
  ret

ALIGN
.globl MAP_WRITE_BBAD\num
MAP_WRITE_BBAD\num:
  movb %al,BBAD_\num
  ret

ALIGN
.globl MAP_WRITE_A1TL\num
MAP_WRITE_A1TL\num:
  movb %al,A1TL_\num
  ret

ALIGN
.globl MAP_WRITE_A1TH\num
MAP_WRITE_A1TH\num:
  movb %al,A1TH_\num
  ret

ALIGN
.globl MAP_WRITE_A1B\num
MAP_WRITE_A1B\num:
  movb %al,A1B_\num
  ret

ALIGN
.globl MAP_WRITE_DASL\num
MAP_WRITE_DASL\num:
  movb %al,DASL_\num
  ret

ALIGN
.globl MAP_WRITE_DASH\num
MAP_WRITE_DASH\num:
  movb %al,DASH_\num
  ret

ALIGN
.globl MAP_WRITE_DASB\num
MAP_WRITE_DASB\num:
  movb %al,DASB_\num
  ret

ALIGN
.globl MAP_WRITE_A2L\num
MAP_WRITE_A2L\num:
  movb %al,A2L_\num
  ret

ALIGN
.globl MAP_WRITE_A2H\num
MAP_WRITE_A2H\num:
  movb %al,A2H_\num
  ret

ALIGN
.globl MAP_WRITE_NTRL\num
MAP_WRITE_NTRL\num:
  movb %al,NTRL_\num
  ret
.endm

MAP_WRITE_DMA 0
MAP_WRITE_DMA 1
MAP_WRITE_DMA 2
MAP_WRITE_DMA 3
MAP_WRITE_DMA 4
MAP_WRITE_DMA 5
MAP_WRITE_DMA 6
MAP_WRITE_DMA 7

.text; ALIGN
.data; ALIGN
.bss; ALIGN
