/*

 EmuGUI.cc
  Contains emulator specific GUI code
  Some stuff from helper.c will be moved here!

*/

//#include <fstream.h>
//#include <dpmi.h>
//#include <go32.h>
//#include <dos.h>
#include <stdio.h>
//#include <stdlib.h>
//#include <conio.h>
#include <unistd.h>

#include "wrapaleg.h"   // Wrapper header for Allegro
#include "guicore.h"

/* ------------------------- GUI STUFF ------------------------- */

#include "types.h"
#include "font.h"
#include "helper.h"
#include "cpu.h"
#include "spc.h"
#include "sound.h"
#include "romload.h"

BITMAP *joypad=0;
extern PALETTE sneesepal;

char *GUI_init(){
 char *errormsg=GUI_core_init();

 if(errormsg) return errormsg;

 #ifdef DEBUG
  joypad=load_pcx("sneese.dat#joypad",sneesepal);
 #else
  joypad=load_pcx("#joypad",sneesepal);
 #endif

 return 0;
}

WINDOW *GUI_window=0;

extern BITMAP *sneese;

WINDOW *Main_window=0;
#define NUM_OPTIONS 8
char Options[NUM_OPTIONS][17]={
 "Resume emulation",
 "Reset emulation ",
 "Load ROM        ",
 "Configure       ",
 "ROM information ",
 "HW status       ",
 "(H)DMA          ",
 "Exit SNEeSe :(  "};
WINDOW *Credits_window=0;

void UpdateGUI(int Selected){
 clear(GUI_Bitmap);
 Main_window->refresh();
 for(int a=0;a<NUM_OPTIONS;a++)
  if(a!=Selected) PlotMenuItem(Main_window,default_font,Options[a],0,default_font->get_heightspace()*a);
  else PlotSelectedMenuItem(Main_window,default_font,Options[a],0,default_font->get_heightspace()*a);
 Credits_window->refresh();
 PlotString(Credits_window,default_font," Code by  Lee Hammerton (SnaX) ",0,0);
 PlotString(Credits_window,default_font,"          Charles Bilyue       ",0,default_font->get_heightspace());
 PlotString(Credits_window,default_font,"          Santeri Saarimaa     ",0,default_font->get_heightspace()*2);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
}

WINDOW *Screen_window=0;
#define NUM_SCREEN_OPTIONS 8
char Screen_Options[NUM_SCREEN_OPTIONS][21]={
 "320x200x256 VGA     ",
 "320x200x256 VGA  FIT",
 "320x240x256 MODE-X  ",
 "256x239x256 VGA     ",
 "320x200x16b SVGA    ",
 "320x240x16b SVGA    ",
 "640x480x16b SVGA    ",
 "640x480x16b SVGA FIT"};

void UpdateScreenWindow(int Selected){
 Screen_window->refresh();
 for(int a=0;a<NUM_SCREEN_OPTIONS;a++)
  if(a!=Selected) PlotMenuItem(Screen_window,default_font,Screen_Options[a],0,default_font->get_heightspace()*a);
  else PlotSelectedMenuItem(Screen_window,default_font,Screen_Options[a],0,default_font->get_heightspace()*a);
 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();
}

int ScreenWindow(){
 int CursorAt = PC_SCREEN_MODE;

 UpdateScreenWindow(CursorAt);

 for(;;){
  if(key[KEY_UP]){
   if(CursorAt) CursorAt--; else CursorAt=NUM_SCREEN_OPTIONS-1;
   UpdateScreenWindow(CursorAt);
   while (key[KEY_UP]);
  }
  if (key[KEY_DOWN]){
   if(CursorAt < NUM_SCREEN_OPTIONS-1) CursorAt++; else CursorAt=0;
   UpdateScreenWindow(CursorAt);
   while(key[KEY_DOWN]);
  }
  if(key[KEY_ESC]){
   while(key[KEY_ESC]);
   return -1;
  }
  if(key[KEY_ENTER]){
   while(key[KEY_ENTER]);
   return CursorAt;
  }
 }
 return -1; // Signify normal exit
}

char FPS_ctr_str[] = "FPS counter: off";

WINDOW *Config_window=0;
#define NUM_CONFIG_OPTIONS 10
char *Config_Options[NUM_CONFIG_OPTIONS]={
 "320x200x256 VGA         ",
 "Min Frameskip: ?        ",
 "Max Frameskip: ?        ",
 "-------- on player 1    ",
 "-------- on player 2    ",
 "Define keys for player 1",
 "Define keys for player 2",
 "Emulate-SPC             ",
 "Sound-Enabled           ",
 FPS_ctr_str               };

char *on_off[2]={ "off", "on" };
char *en_dis[2]={ "enabled", "disabled" };

void UpdateConfigWindow(int Selected){
 Config_window->refresh();
 for(int a=0;a<NUM_CONFIG_OPTIONS;a++)
  if(a!=Selected) PlotMenuItem(Config_window,default_font,Config_Options[a],0,default_font->get_heightspace()*a);
  else PlotSelectedMenuItem(Config_window,default_font,Config_Options[a],0,default_font->get_heightspace()*a);
 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();
}

WINDOW *ControlSetup_window=0;

void UpdateController1Screen(){
 char tempch[5];

 ControlSetup_window->refresh();

 scantotext(KEY_UP_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,31,15,8);
 scantotext(KEY_DOWN_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,65,15,8);
 scantotext(KEY_LEFT_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,20,48,15,8);
 scantotext(KEY_RIGHT_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,55,48,15,8);
 scantotext(KEY_A_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,151,46,15,8);
 scantotext(KEY_B_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,130,69,15,8);
 scantotext(KEY_X_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,25,15,8);
 scantotext(KEY_Y_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,104,49,15,8);
 scantotext(KEY_L_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,40,1,15,8);
 scantotext(KEY_R_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,1,15,8);
 scantotext(KEY_SELECT_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,68,40,15,8);
 scantotext(KEY_START_1,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,91,40,15,8);
}

void UpdateController2Screen(){
 char tempch[5];

 ControlSetup_window->refresh();

 scantotext(KEY_UP_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,31,15,8);
 scantotext(KEY_DOWN_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,38,65,15,8);
 scantotext(KEY_LEFT_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,20,48,15,8);
 scantotext(KEY_RIGHT_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,55,48,15,8);
 scantotext(KEY_A_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,151,46,15,8);
 scantotext(KEY_B_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,130,69,15,8);
 scantotext(KEY_X_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,25,15,8);
 scantotext(KEY_Y_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,104,49,15,8);
 scantotext(KEY_L_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,40,1,15,8);
 scantotext(KEY_R_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,127,1,15,8);
 scantotext(KEY_SELECT_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,68,40,15,8);
 scantotext(KEY_START_2,tempch);
 PlotStringShadow(ControlSetup_window,default_font,tempch,91,40,15,8);
}

int AskKey(char *msg, unsigned *whatkey, int jp){
 unsigned tmp;

 PlotStringShadow(ControlSetup_window,default_font,msg,84,96,15,8);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 CopyGUIScreen();

 tmp=WaitForKeyPress();

 if(tmp==KEY_ESC) return FALSE;

 *whatkey=tmp;

 if(jp==1) UpdateController1Screen();
 else UpdateController2Screen();

 return TRUE;
}

extern "C" unsigned char FPS_ENABLED;

int ConfigWindow(){
 int temp;

 strcpy(Config_Options[0],Screen_Options[PC_SCREEN_MODE]);
 Config_Options[0][20]=' ';  // Stupid thing, but who cares
 Config_Options[0][21]=' ';
 Config_Options[0][22]=' ';
 Config_Options[0][23]=' ';
 Config_Options[0][24]=0;

 switch(JOYSTICK_ENABLED){
  case 0:
   strcpy(Config_Options[3],"Keyboard on player 1    ");
   break;
  case 1:
   JOYSTICK_ENABLED+=1;
// strcpy(Config_Options[3],"Joystick on player 1    ");
// break;
  default:
   strcpy(Config_Options[3],"Mouse on player 1       ");
 }
 switch(JOYSTICK_ENABLED2){
  case 0:
   strcpy(Config_Options[4],"Keyboard on player 2    ");
   break;
  case 1:
   JOYSTICK_ENABLED2+=1;
// strcpy(Config_Options[4],"Joystick on player 2    ");
// break;
  default:
   strcpy(Config_Options[4],"Mouse on player 2       ");
 }

 if(SPC_ENABLED) strcpy(Config_Options[7],"Emulate SPC            ");
 else strcpy(Config_Options[7],"Skip SPC               ");

 switch(SOUND_ENABLED){
  case 0:
   strcpy(Config_Options[8],"Sound Disabled          ");
   break;
  case 1:
   strcpy(Config_Options[8],"Sound Enabled (mono)    ");
   break;
  default:
   strcpy(Config_Options[8],"Sound Enabled (stereo)  ");
 }

 sprintf(Config_Options[1],"Min Frameskip: %-2ld       ",FRAME_SKIP_MIN);
 sprintf(Config_Options[2],"Max Frameskip: %-2ld       ",FRAME_SKIP_MAX);

 UpdateConfigWindow(0);
 int CursorAt=0;

 for(;;){
  if(key[KEY_UP]){
   CursorAt--;
   if(CursorAt == -1) // so it wraps
    CursorAt = NUM_CONFIG_OPTIONS-1;
   UpdateConfigWindow(CursorAt);
    while(key[KEY_UP]);
  }
  if(key[KEY_DOWN]){
   CursorAt++;
   if(CursorAt == NUM_CONFIG_OPTIONS) // so it wraps
   CursorAt=0;
   UpdateConfigWindow(CursorAt);
   while(key[KEY_DOWN]);
  }

  // For min frameskip!
  if(CursorAt==1){
   if(key[KEY_LEFT] && FRAME_SKIP_MIN>0){
    // 60,30,20,15,10-0
    if(FRAME_SKIP_MIN>30) FRAME_SKIP_MIN-=30;
    else if(FRAME_SKIP_MIN>20) FRAME_SKIP_MIN-=10;
    else if(FRAME_SKIP_MIN>10) FRAME_SKIP_MIN-=5;
    else FRAME_SKIP_MIN--;

    if(FRAME_SKIP_MIN>FRAME_SKIP_MAX){
     FRAME_SKIP_MAX=FRAME_SKIP_MIN;
     sprintf(Config_Options[2],"Max Frameskip: %-2ld       ",FRAME_SKIP_MAX);
    }
    sprintf(Config_Options[1],"Min Frameskip: %-2ld       ",FRAME_SKIP_MIN);
    UpdateConfigWindow(CursorAt);
    while(key[KEY_LEFT]);
   }
   if (key[KEY_RIGHT] && FRAME_SKIP_MIN<60){
    // 0-10,15,20,30,60
    if(FRAME_SKIP_MIN<10) FRAME_SKIP_MIN++;
    else if(FRAME_SKIP_MIN<20) FRAME_SKIP_MIN+=5;
    else if(FRAME_SKIP_MIN<30) FRAME_SKIP_MIN+=10;
    else FRAME_SKIP_MIN+=30;

    if(FRAME_SKIP_MIN>FRAME_SKIP_MAX){
     FRAME_SKIP_MAX=FRAME_SKIP_MIN;
     sprintf(Config_Options[2],"Max Frameskip: %-2ld       ",FRAME_SKIP_MAX);
    }
    sprintf(Config_Options[1],"Min Frameskip: %-2ld       ",FRAME_SKIP_MIN);
    UpdateConfigWindow(CursorAt);
    while (key[KEY_RIGHT]);
   }
  }
  // For max frameskip!
  if(CursorAt==2){
   if(key[KEY_LEFT] && FRAME_SKIP_MAX>1){
    // 60,30,20,15,10-0
    if(FRAME_SKIP_MAX>30) FRAME_SKIP_MAX-=30;
    else if(FRAME_SKIP_MAX>20) FRAME_SKIP_MAX-=10;
    else if(FRAME_SKIP_MAX>10) FRAME_SKIP_MAX-=5;
    else FRAME_SKIP_MAX--;

    if(FRAME_SKIP_MAX<FRAME_SKIP_MIN){
     FRAME_SKIP_MIN=FRAME_SKIP_MAX;
     sprintf(Config_Options[1],"Min Frameskip: %-2ld       ",FRAME_SKIP_MIN);
    }
    sprintf(Config_Options[2],"Max Frameskip: %-2ld       ",FRAME_SKIP_MAX);
    UpdateConfigWindow(CursorAt);
    while(key[KEY_LEFT]);
   }
   if (key[KEY_RIGHT] && FRAME_SKIP_MAX<60){
    // 0-10,15,20,30,60
    if(FRAME_SKIP_MAX<10) FRAME_SKIP_MAX++;
    else if(FRAME_SKIP_MAX<20) FRAME_SKIP_MAX+=5;
    else if(FRAME_SKIP_MAX<30) FRAME_SKIP_MAX+=10;
    else FRAME_SKIP_MAX+=30;

    if(FRAME_SKIP_MAX<FRAME_SKIP_MIN){
     FRAME_SKIP_MIN=FRAME_SKIP_MAX;
     sprintf(Config_Options[1],"Min Frameskip: %-2ld       ",FRAME_SKIP_MIN);
    }
    sprintf(Config_Options[2],"Max Frameskip: %-2ld       ",FRAME_SKIP_MAX);
    UpdateConfigWindow(CursorAt);
    while (key[KEY_RIGHT]);
   }
  }
  if (key[KEY_ESC]){
   while(key[KEY_ESC]);
   return -1;
  }
  if(key[KEY_ENTER]){
   while(key[KEY_ENTER]);
   if(CursorAt==0){
    while((temp=ScreenWindow())!=-1){
     SetGUIScreen(temp);    // Setup screen mode (PC_SCREEN_MODE is set here so following works)

     strcpy(Config_Options[0],Screen_Options[PC_SCREEN_MODE]);
     Config_Options[0][20]=' ';  // Stupid thing, but who cares
     Config_Options[0][21]=' ';
     Config_Options[0][22]=' ';
     Config_Options[0][23]=' ';
     Config_Options[0][24]=0;
     if(sneese){  // Prevent a crash if file not found!
      set_palette(sneesepal);
      stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
     }
     set_palette_range(&GUIPal[-240],240,255,1);    // Set the GUI palette up.
     UpdateConfigWindow(0);
    }
    UpdateGUI(3);
   }
   if(CursorAt==3){
    switch(JOYSTICK_ENABLED){
     case 0:
//    strcpy(Config_Options[3],"Joystick on player 1    ");
//    JOYSTICK_ENABLED=1;
//    break;
     case 1:
      if(MOUSE_PLUGGED_IN==1){
       strcpy(Config_Options[3],"Mouse on player 1       ");
       JOYSTICK_ENABLED=2;
       break;
      }
     default:
      strcpy(Config_Options[3],"Keyboard on player 1    ");
      JOYSTICK_ENABLED=0;
    }
   }
   if(CursorAt==4){
    switch(JOYSTICK_ENABLED2){
     case 0:
//    strcpy(Config_Options[4],"Joystick on player 2    ");
//    JOYSTICK_ENABLED2=1;
//    break;
     case 1:
      if(MOUSE_PLUGGED_IN==1){
       strcpy(Config_Options[4],"Mouse on player 2       ");
       JOYSTICK_ENABLED2=2;
       break;
      }
     default:
      strcpy(Config_Options[4],"Keyboard on player 2    ");
      JOYSTICK_ENABLED2=0;
    }
   }
   if(CursorAt==5){
    while(key[KEY_ENTER]);
    clear_keybuf();
    UpdateController1Screen();

    if(AskKey("Press key for UP",&KEY_UP_1,1)==TRUE)
    if(AskKey("Press key for DOWN",&KEY_DOWN_1,1)==TRUE)
    if(AskKey("Press key for LEFT",&KEY_LEFT_1,1)==TRUE) // this is what I call optimization :)
    if(AskKey("Press key for RIGHT",&KEY_RIGHT_1,1)==TRUE)
    if(AskKey("Press key for A",&KEY_A_1,1)==TRUE)
    if(AskKey("Press key for B",&KEY_B_1,1)==TRUE)
    if(AskKey("Press key for X",&KEY_X_1,1)==TRUE)
    if(AskKey("Press key for Y",&KEY_Y_1,1)==TRUE)
    if(AskKey("Press key for L",&KEY_L_1,1)==TRUE)
    if(AskKey("Press key for R",&KEY_R_1,1)==TRUE)
    if(AskKey("Press key for SELECT",&KEY_SELECT_1,1)==TRUE)
    if(AskKey("Press key for START",&KEY_START_1,1)==TRUE)
    {
     PlotStringShadow(ControlSetup_window,default_font,"Press ESC to exit",84,96,15,8);
     if(sneese)
      stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
     draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
     CopyGUIScreen();

     while (!key[KEY_ESC]);   // :)
    }
    while (key[KEY_ESC]);

    UpdateGUI(3);
    UpdateConfigWindow(CursorAt);
   }
   if(CursorAt==6){
    while(key[KEY_ENTER]);
    clear_keybuf();
    UpdateController2Screen();

    if(AskKey("Press key for UP",&KEY_UP_2,2)==TRUE)
    if(AskKey("Press key for DOWN",&KEY_DOWN_2,2)==TRUE)
    if(AskKey("Press key for LEFT",&KEY_LEFT_2,2)==TRUE) // this is what I call optimization :)
    if(AskKey("Press key for RIGHT",&KEY_RIGHT_2,2)==TRUE)
    if(AskKey("Press key for A",&KEY_A_2,2)==TRUE)
    if(AskKey("Press key for B",&KEY_B_2,2)==TRUE)
    if(AskKey("Press key for X",&KEY_X_2,2)==TRUE)
    if(AskKey("Press key for Y",&KEY_Y_2,2)==TRUE)
    if(AskKey("Press key for L",&KEY_L_2,2)==TRUE)
    if(AskKey("Press key for R",&KEY_R_2,2)==TRUE)
    if(AskKey("Press key for SELECT",&KEY_SELECT_2,2)==TRUE)
    if(AskKey("Press key for START",&KEY_START_2,2)==TRUE)
    {
     PlotStringShadow(ControlSetup_window,default_font,"Press ESC to exit",84,96,15,8);
     if(sneese)
      stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
     draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
     CopyGUIScreen();

     while(!key[KEY_ESC]);
    }
    while (key[KEY_ESC]);

    UpdateGUI(3);
    UpdateConfigWindow(CursorAt);
   }
   if(CursorAt==7){
    if(SPC_ENABLED){
     strcpy(Config_Options[7],"Skip SPC               ");
     SPC_ENABLED=0;
    } else {
     strcpy(Config_Options[7],"Emulate SPC            ");
     SPC_ENABLED=1;
    }
   }
   if(CursorAt==8){
    switch(SOUND_ENABLED){
     case 0:
      if(Install_Sound(0)){
       strcpy(Config_Options[8],"Sound Enabled (mono)    ");
       SOUND_ENABLED=1;
      }
      break;
     case 1:
      Remove_Sound();
      if(Install_Sound(1)){
       strcpy(Config_Options[8],"Sound Enabled (stereo)  ");
       SOUND_ENABLED=2;
      } else {
       strcpy(Config_Options[8],"Sound Disabled          ");
       SOUND_ENABLED=0;
      }
      break;
     case 2:
      Remove_Sound();
      strcpy(Config_Options[8],"Sound Disabled          ");
      SOUND_ENABLED=0;
    }
   }
   if(CursorAt==9){
    strcpy(FPS_ctr_str + 13, on_off[(FPS_ENABLED ^= -1) ? 1 : 0]);
   }
   UpdateConfigWindow(CursorAt);
  }
 }
 return 1;          // Signify normal exit
}

WINDOW *ROMInfo_window=0;

void RomInfo(void){
 ROMInfo_window->refresh();
 PlotString(ROMInfo_window,default_font,rom_romfile,0,0);
 PlotString(ROMInfo_window,default_font,rom_romname,0,default_font->get_heightspace());
 PlotString(ROMInfo_window,default_font,rom_romtype,0,default_font->get_heightspace()*2);
 PlotString(ROMInfo_window,default_font,rom_romhilo,0,default_font->get_heightspace()*3);
 PlotString(ROMInfo_window,default_font,rom_romsize,0,default_font->get_heightspace()*4);
 PlotString(ROMInfo_window,default_font,rom_sram,0,default_font->get_heightspace()*5);
 PlotString(ROMInfo_window,default_font,rom_country,0,default_font->get_heightspace()*6);
 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();

 while(!key[KEY_ESC]);   // :)
 while(key[KEY_ESC]);
}

extern unsigned char BGMODE,VMAIN;
extern "C" unsigned char W12SEL,W34SEL,WOBJSEL,WH0,WH1,WH2,WH3;
extern "C" unsigned char TM,TS,TMW,TSW,WBGLOG,WOBJLOG,CGWSEL,CGADSUB;

WINDOW *HWStatus_window=0;

void HWStatus(void){
 char Number[5];

 HWStatus_window->refresh();
 PlotString(HWStatus_window,default_font,"65c816 A:#### X:#### Y:#### DB:##",0,0);
 sprintf(Number,"%04X",(unsigned short) _65816_A);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*9,0);
 sprintf(Number,"%04X",(unsigned short) _65816_X);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*16,0);
 sprintf(Number,"%04X",(unsigned short) _65816_Y);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*23,0);
 sprintf(Number,"%02X",(unsigned short) _65816_DB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*31,0);
 PlotString(HWStatus_window,default_font,"PB:## PC:#### S:#### D:#### P:###",0,default_font->get_heightspace());
 sprintf(Number,"%02X",(unsigned short) _65816_PB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*3,default_font->get_heightspace());
 sprintf(Number,"%04X",(unsigned short) _65816_PC);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*9,default_font->get_heightspace());
 sprintf(Number,"%04X",(unsigned short) _65816_S);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*16,default_font->get_heightspace());
 sprintf(Number,"%04X",(unsigned short) _65816_D);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*23,default_font->get_heightspace());
// sprintf(Number,"%03X",_65816_P&0x1FF
// PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*30,default_font->get_heightspace());
 PlotString(HWStatus_window,default_font,"OLD_PB:## OLD_PC:####",0,default_font->get_heightspace()*2);
 sprintf(Number,"%02X",(unsigned short) OLD_PB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*7,default_font->get_heightspace()*2);
 sprintf(Number,"%04X",(unsigned short) OLD_PC);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*17,default_font->get_heightspace()*2);

 PlotString(HWStatus_window,default_font,"BGMODE:## VMAIN:## TM:## TS:##",0,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",BGMODE);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*7,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",VMAIN);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*16,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",TM);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*22,default_font->get_heightspace()*3);
 sprintf(Number,"%02X",TS);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*28,default_font->get_heightspace()*3);
 PlotString(HWStatus_window,default_font,"WH0: ## WH1: ## WH2: ## WH3: ##",0,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH0);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*5,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH1);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*13,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH2);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*21,default_font->get_heightspace()*4);
 sprintf(Number,"%02X",WH3);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*29,default_font->get_heightspace()*4);
 PlotString(HWStatus_window,default_font,"TMW: ## TSW: ## WSEL: ## ## ##",0,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",TMW);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*5,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",TSW);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*13,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",W12SEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*22,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",W34SEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*25,default_font->get_heightspace()*5);
 sprintf(Number,"%02X",WOBJSEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*28,default_font->get_heightspace()*5);

 PlotString(HWStatus_window,default_font,"WLOG:## ## CGWSEL:## CGADSUB:##",0,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",WBGLOG);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*5,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",WOBJLOG);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*8,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",CGWSEL);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*18,default_font->get_heightspace()*6);
 sprintf(Number,"%02X",CGADSUB);
 PlotString(HWStatus_window,default_font,Number,default_font->get_widthspace()*29,default_font->get_heightspace()*6);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();

 while(!key[KEY_ESC]);
 while(key[KEY_ESC]);
}

extern unsigned char MDMAEN asm("MDMAEN");
extern unsigned char HDMAEN asm("HDMAEN");
extern unsigned char
 DMAP0 asm("DMAP_0"),DMAP1 asm("DMAP_1"),
 DMAP2 asm("DMAP_2"),DMAP3 asm("DMAP_3"),
 DMAP4 asm("DMAP_4"),DMAP5 asm("DMAP_5"),
 DMAP6 asm("DMAP_6"),DMAP7 asm("DMAP_7");
extern unsigned char
 BBAD0 asm("BBAD_0"),BBAD1 asm("BBAD_1"),
 BBAD2 asm("BBAD_2"),BBAD3 asm("BBAD_3"),
 BBAD4 asm("BBAD_4"),BBAD5 asm("BBAD_5"),
 BBAD6 asm("BBAD_6"),BBAD7 asm("BBAD_7");
extern unsigned char
 A1TL0 asm("A1TL_0"),A1TL1 asm("A1TL_1"),
 A1TL2 asm("A1TL_2"),A1TL3 asm("A1TL_3"),
 A1TL4 asm("A1TL_4"),A1TL5 asm("A1TL_5"),
 A1TL6 asm("A1TL_6"),A1TL7 asm("A1TL_7");
extern unsigned char
 A1TH0 asm("A1TH_0"),A1TH1 asm("A1TH_1"),
 A1TH2 asm("A1TH_2"),A1TH3 asm("A1TH_3"),
 A1TH4 asm("A1TH_4"),A1TH5 asm("A1TH_5"),
 A1TH6 asm("A1TH_6"),A1TH7 asm("A1TH_7");
extern unsigned char
 A1B0 asm("A1B_0"),A1B1 asm("A1B_1"),
 A1B2 asm("A1B_2"),A1B3 asm("A1B_3"),
 A1B4 asm("A1B_4"),A1B5 asm("A1B_5"),
 A1B6 asm("A1B_6"),A1B7 asm("A1B_7");
extern unsigned char
 DASL0 asm("DASL_0"),DASL1 asm("DASL_1"),
 DASL2 asm("DASL_2"),DASL3 asm("DASL_3"),
 DASL4 asm("DASL_4"),DASL5 asm("DASL_5"),
 DASL6 asm("DASL_6"),DASL7 asm("DASL_7");
extern unsigned char
 DASH0 asm("DASH_0"),DASH1 asm("DASH_1"),
 DASH2 asm("DASH_2"),DASH3 asm("DASH_3"),
 DASH4 asm("DASH_4"),DASH5 asm("DASH_5"),
 DASH6 asm("DASH_6"),DASH7 asm("DASH_7");
extern unsigned char
 DASB0 asm("DASB_0"),DASB1 asm("DASB_1"),
 DASB2 asm("DASB_2"),DASB3 asm("DASB_3"),
 DASB4 asm("DASB_4"),DASB5 asm("DASB_5"),
 DASB6 asm("DASB_6"),DASB7 asm("DASB_7");

WINDOW *DMAStatus_window=0;

void DMAStatus(void){
 char Number[5];
 char Line[34];

 DMAStatus_window->refresh();
 sprintf(Line,"MDMAEN:%02X HDMAEN:%02X",(unsigned char) MDMAEN,
  (unsigned char) HDMAEN);
 PlotString(DMAStatus_window,default_font,Line,0,0);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP0,(unsigned char) BBAD0,
 (unsigned char) A1B0,(unsigned char) A1TH0,(unsigned char) A1TL0,
 (unsigned char) DASB0,(unsigned char) DASH0,(unsigned char) DASL0);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace());

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP1,(unsigned char) BBAD1,
 (unsigned char) A1B1,(unsigned char) A1TH1,(unsigned char) A1TL1,
 (unsigned char) DASB1,(unsigned char) DASH1,(unsigned char) DASL1);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*2);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP2,(unsigned char) BBAD2,
 (unsigned char) A1B2,(unsigned char) A1TH2,(unsigned char) A1TL2,
 (unsigned char) DASB2,(unsigned char) DASH2,(unsigned char) DASL2);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*3);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP3,(unsigned char) BBAD3,
 (unsigned char) A1B3,(unsigned char) A1TH3,(unsigned char) A1TL3,
 (unsigned char) DASB3,(unsigned char) DASH3,(unsigned char) DASL3);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*4);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP4,(unsigned char) BBAD4,
 (unsigned char) A1B4,(unsigned char) A1TH4,(unsigned char) A1TL4,
 (unsigned char) DASB4,(unsigned char) DASH4,(unsigned char) DASL4);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*5);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP5,(unsigned char) BBAD5,
 (unsigned char) A1B5,(unsigned char) A1TH5,(unsigned char) A1TL5,
 (unsigned char) DASB5,(unsigned char) DASH5,(unsigned char) DASL5);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*6);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP6,(unsigned char) BBAD6,
 (unsigned char) A1B6,(unsigned char) A1TH6,(unsigned char) A1TL6,
 (unsigned char) DASB6,(unsigned char) DASH6,(unsigned char) DASL6);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*7);

 sprintf(Line,"%02X %02X %02X%02X%02X %02X%02X%02X",
 (unsigned char) DMAP7,(unsigned char) BBAD7,
 (unsigned char) A1B7,(unsigned char) A1TH7,(unsigned char) A1TL7,
 (unsigned char) DASB7,(unsigned char) DASH7,(unsigned char) DASL7);
 PlotString(DMAStatus_window,default_font,Line,0,default_font->get_heightspace()*8);

 if(sneese)
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 draw_sprite(Allegro_Bitmap,GUI_Bitmap,0,0);
 vsync();
 CopyGUIScreen();

 while(!key[KEY_ESC]);
 while(key[KEY_ESC]);
}

CTL *joypad_bitmap=0;

CTL_CLEAR GUI_clear;
CTL_CLEAR Main_clear;
CTL_CLEAR Credits_clear;
CTL_CLEAR Screen_clear;
CTL_CLEAR Config_clear;
CTL_CLEAR ControlSetup_clear;
CTL_CLEAR ROMInfo_clear;
CTL_CLEAR HWStatus_clear;
CTL_CLEAR DMAStatus_clear;
CTL_CLEAR File_clear;

int allocate_windows(){
 static int file_cleared=0;
 if(!file_cleared) File_window+=File_clear;

 GUI_window=new BORDER_WINDOW(-1,-1,320,240);
 Main_window=new BORDER_WINDOW(0,0,16*default_font->get_widthspace(),NUM_OPTIONS*default_font->get_heightspace());
 Credits_window=new BORDER_WINDOW(10,160,32*default_font->get_widthspace(),3*default_font->get_heightspace());
 Screen_window=new BORDER_WINDOW(100,30,20*default_font->get_widthspace(),NUM_SCREEN_OPTIONS*default_font->get_heightspace());
 Config_window=new BORDER_WINDOW(70,15,24*default_font->get_widthspace(),NUM_CONFIG_OPTIONS*default_font->get_heightspace());
 ControlSetup_window=new BORDER_WINDOW(29,34,168,104);
 ROMInfo_window=new BORDER_WINDOW(10,70,33*default_font->get_widthspace(),7*default_font->get_heightspace());
 HWStatus_window=new BORDER_WINDOW(10,70,33*default_font->get_widthspace(),7*default_font->get_heightspace());
 DMAStatus_window=new BORDER_WINDOW(10,70,33*default_font->get_widthspace(),9*default_font->get_heightspace());
 if(!(GUI_window && Main_window && Credits_window &&
    Screen_window && Config_window && ControlSetup_window &&
    ROMInfo_window && HWStatus_window && DMAStatus_window)) return 1;
 *GUI_window+=GUI_clear;
 *Main_window+=Main_clear;
 *Credits_window+=Credits_clear;
 *Screen_window+=Screen_clear;
 *Config_window+=Config_clear;
 *ControlSetup_window+=ControlSetup_clear;
 if(joypad)
  joypad_bitmap=new CTL_BITMAP(ControlSetup_window,joypad,7,12);
 *ROMInfo_window+=ROMInfo_clear;
 *HWStatus_window+=HWStatus_clear;
 *DMAStatus_window+=DMAStatus_clear;
 return 0;
}

void free_windows(){
 delete GUI_window;
 delete Main_window; delete Credits_window;
 delete joypad_bitmap;
 delete Screen_window; delete Config_window; delete ControlSetup_window;
 delete ROMInfo_window; delete HWStatus_window; delete DMAStatus_window;

 GUI_window=0;
 Main_window=0; Credits_window=0;
 joypad_bitmap=0;
 Screen_window=0; Config_window=0; ControlSetup_window=0;
 ROMInfo_window=0; HWStatus_window=0; DMAStatus_window=0;
}

extern volatile unsigned Timer_Counter_Throttle asm("Throttle");
extern volatile unsigned Timer_Counter_Profile,Timer_Counter_FPS;

int GUI(){

#ifdef DEBUG
 unsigned long Timer_Value,Timer_Value2,Timer_Value3;
#endif
 unsigned char Old_SPC=SPC_ENABLED;

#ifdef DEBUG
 Timer_Value=Timer_Counter_Profile;
 Timer_Value2=Timer_Counter_FPS;
 Timer_Value3=Timer_Counter_Throttle;
#endif

 allocate_windows();

 if(sneese){ // Prevent a crash if file not found!
  set_palette(sneesepal);
  stretch_blit(sneese,Allegro_Bitmap,0,0,GUI_ScreenWidth,GUI_ScreenHeight,0,0,SCREEN_W,SCREEN_H);
 }
 while (key[KEY_ESC]);

 char TempBufferP[MAXPATH]; // For saving the current directory to use!

 set_palette_range(&GUIPal[-240],240,255,1);    // Set the GUI palette up.

 UpdateGUI(0);
 CopyGUIScreen();

 int CursorAt=0;

 for(;;){
  if(key[KEY_UP]){
   CursorAt--;
   if(CursorAt == -1) // so it wraps
    CursorAt = NUM_OPTIONS-1;
   UpdateGUI(CursorAt);
   CopyGUIScreen();
   while(key[KEY_UP]);
  }
  if(key[KEY_DOWN]){
   CursorAt++;
   if(CursorAt == NUM_OPTIONS) // so it wraps
    CursorAt=0;
   UpdateGUI(CursorAt);
   CopyGUIScreen();
   while (key[KEY_DOWN]);
  }
  if(key[KEY_ESC]){
   if(ROM_Loaded){
    while(key[KEY_ESC]);
    break;
   }
  }
  if(key[KEY_ENTER]){
   while (key[KEY_ENTER]);
   if(ROM_Loaded){
    if(CursorAt==0) break;
    if(CursorAt==1){
     Reset_Memory();
     install_int_ex(Timer_Handler_Throttle,Frametime);
     Reset_CPU();
     remove_int(Timer_Handler_Throttle);
     Reset_SPC();
     break;
    }
   }
   if(CursorAt==2){
    getcwd(TempBufferP,MAXPATH);
    if(FileWindow()==0){
     chdir(TempBufferP);
     break;
    }
    chdir(TempBufferP);
   }
   if(CursorAt==3)
    ConfigWindow();
   if(ROM_Loaded){
    if(CursorAt==4)
     RomInfo();
    if(CursorAt==5)
     HWStatus();
    if(CursorAt==6)
     DMAStatus();
   }
   if(CursorAt==7){
#ifdef DEBUG
    Timer_Counter_Profile=Timer_Value;
    Timer_Counter_FPS=Timer_Value2;
    Timer_Counter_Throttle=Timer_Value3;
#endif
    free_windows();
    return 0;
   }
   UpdateGUI(CursorAt);
   CopyGUIScreen();
  }
 }

 if(ROM_Loaded && SPC_ENABLED && (SPC_ENABLED!=Old_SPC)){
  Reset_Memory();
  install_int_ex(Timer_Handler_Throttle,Frametime);
  Reset_CPU();          // Must be done to ensure ROM works ok!
  remove_int(Timer_Handler_Throttle);
  Reset_SPC();
 }

#ifdef DEBUG
 Timer_Counter_Profile=Timer_Value;
 Timer_Counter_FPS=Timer_Value2;
 Timer_Counter_Throttle=Timer_Value3;
#endif

 clear(screen);
 clear(Allegro_Bitmap);

 free_windows();
 return 1;          // Signify normal exit
}
