// IRQ/NMI implementation below!

#include "misc.i"
#include "scankeys.h"
#include "key.i"
#include "cpu.i"
#include "screen.i"
#include "DMA.i"
#include "SPC.i"

#define FPSTicks _FPSTicks
#define FPSMaxTicks _FPSMaxTicks
#define FPSLast _FPSLast

#define FRAME_LIMIT 0
// Set this to force emulation loop to break after x frames have been
//  emulated (not displayed!) This does not save CPU regs, etc.!
//  Primary use is for profiling!
//#define VAR_FRAME_LIMIT
// Set this to force emulation loop to break after FrameLimit frames

.data
.balign 4
Event_Table_224_NTSC_NoRender:
DUPLICATE long,1,IRQHDMAFirst
DUPLICATE long,222,IRQHDMA  //222
.long IRQHDMALast
.long NMI_NoRender
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,34,VBL
.long IRQNewFrame

Event_Table_239_NTSC_NoRender:
DUPLICATE long,1,IRQHDMAFirst
DUPLICATE long,237,IRQHDMA
.long IRQHDMALast
.long NMI_NoRender
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,19,VBL
.long IRQNewFrame

Event_Table_224_NTSC:
DUPLICATE long,1,IRQHDMARenderFirst
DUPLICATE long,222,IRQHDMARender    //222
.long IRQHDMARenderLast
.long NMI
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,34,VBL
.long IRQNewFrame

Event_Table_239_NTSC:
DUPLICATE long,1,IRQHDMARenderFirst
DUPLICATE long,237,IRQHDMARender
.long IRQHDMARenderLast
.long NMI
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,19,VBL
.long IRQNewFrame

Event_Table_224_PAL_NoRender:
DUPLICATE long,1,IRQHDMAFirst
DUPLICATE long,222,IRQHDMA
.long IRQHDMALast
.long NMI_NoRender
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,84,VBL
.long IRQNewFrame

Event_Table_239_PAL_NoRender:
DUPLICATE long,1,IRQHDMAFirst
DUPLICATE long,237,IRQHDMA
.long IRQHDMALast
.long NMI_NoRender
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,69,VBL
.long IRQNewFrame

Event_Table_224_PAL:
DUPLICATE long,1,IRQHDMARenderFirst
DUPLICATE long,222,IRQHDMARender
.long IRQHDMARenderLast
.long NMI
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,84,VBL
.long IRQNewFrame

Event_Table_239_PAL:
DUPLICATE long,1,IRQHDMARenderFirst
DUPLICATE long,237,IRQHDMARender
.long IRQHDMARenderLast
.long NMI
DUPLICATE long,1,VBL_First_Line
DUPLICATE long,1,VBL_Update_Controllers
DUPLICATE long,69,VBL
.long IRQNewFrame

EXPORT_C Event_Handler_Table_NTSC
.long Event_Table_224_NTSC_NoRender
.long Event_Table_239_NTSC_NoRender
.long Event_Table_224_NTSC
.long Event_Table_239_NTSC
EXPORT_C Event_Handler_Table_PAL
.long Event_Table_224_PAL_NoRender
.long Event_Table_239_PAL_NoRender
.long Event_Table_224_PAL
.long Event_Table_239_PAL

.bss
ALIGN
.globl ScreenLineHoles
ScreenLineHoles:skipb (256+128)/8
.globl ScreenDestPtr
ScreenDestPtr:  skipl
.globl FrameCount
FrameCount:     skipl   // Number of frames waited, for speeding up emu
.globl Event_Handler
Event_Handler:  skipl   // Controls screen height and render
.globl Old_Event_Handler
Old_Event_Handler:  skipl
.globl FPSTicks,FPSMaxTicks,FPSLast
FPSTicks:       skipl   // Count of timer ticks for this second
FPSCount:       skipl   // Count of frames executed this second
FPSMaxTicks:    skipl   // Number of timer ticks after which FPS is calc'd
FPSLast:        skipl   // Calculated FPS for last second

HIRQOldTable:   skipl

.text
.globl E1_IRQ,E0_IRQ
E1_IRQ:
    // Emulation mode IRQ
    movl _PC,%eax
    E1_PUSH_W
    CLR_FLAG SNES_FLAG_B    // Clear break bit on stack
    E1_SETUPFLAGS           // put flags into SNES packed flag format
    SET_FLAG SNES_FLAG_B
    E1_PUSH_B

    movl _IRQ_Evector,%eax  // Get Emulation mode IRQ vector
    movl _IRQ_Eoffset,%edx
    jmp 2f

ALIGN
E0_IRQ:
    // Native mode IRQ
    movb _PB,%al
    E0_PUSH_B
    movl _PC,%eax
    E0_PUSH_W
    E0_SETUPFLAGS           // put flags into SNES packed flag format
    E0_PUSH_B

    movl _IRQ_Nvector,%eax  // Get Native mode IRQ vector
    movl _IRQ_Noffset,%edx
2:  movl %eax,_PC           // Setup PC vector
    movl %edx,_PBOffset
    movb $0,_PB             // Setup bank
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    movl %eax,_PB_Read

    SET_FLAG SNES_FLAG_I    // Disable interrupts
    CLR_FLAG SNES_FLAG_D
    ret

.macro Save_Cycles Execute=0
    movl SNES_Cycles,%eax   // Update CPU and SPC cycles
    movl %ebp,SNES_Cycles
    subl SPC_last_cycles,%eax
    movl %ebp,SPC_last_cycles
.if \Execute-0
    cmpb $0,_SPC_ENABLED
    jz  8f
    addl SPC_CPU_cycles,%eax
    Execute_SPC SaveCycles=0
8:
.else
    addl %eax,SPC_CPU_cycles
.endif
.endm

.globl IRQNewFrameReset
ALIGN
IRQNewFrameReset:
    // Reset frameskip counter
    movl $1,%edx
    cmpl _FRAME_SKIP_MIN,%edx
    ja  2f
    movl _FRAME_SKIP_MIN,%edx
2:  movl %edx,FrameCount

    xorl %eax,%eax
    movl %eax,HDMAYCount    // Reset scanline counter
    incl FPSCount           // For FPS counter - 0.25b14
    movl %eax,Throttle

    movl _Event_Handler_Table,%edi
    movl 8(%edi),%eax
    movl %eax,Event_Handler
    RESTORE_HDMA_ADDRESS    // Relatch HDMA registers
    movl _BKG,%edi
    addl $64,%edi
    movl %edi,ScreenDestPtr // Setup destination screen pointer

    movl FPSMaxTicks,%eax
    cmpl %eax,FPSTicks
    jb 0f
    movl FPSCount,%eax
    je 3f
    cmpl $1,%eax
    ja 3f
    movl $0,FPSLast
    jmp 4f
3:  movl %eax,FPSLast
4:  movl $0,FPSTicks
    movl $0,FPSCount

ALIGN
0:  xorl %eax,%eax
    ret

.globl IRQNewFrame
ALIGN
2:  hlt
    jmp 1f
ALIGN
IRQNewFrame:                // Check IRQ, Frame Skip
    Save_Cycles Execute=0

    movb _INIDISP,%al
    andb $0x0F,%al
    cmpb _BrightnessLevel,%al
    je  0f
    movb $1,PaletteChanged
    movb %al,_BrightnessLevel
0:
    movl _OAMAddress_VBL,%eax
    movl %eax,_OAMAddress   // Restore Sprite address every VBLANK!
    xorl %eax,%eax
    movb %al,OAMOdd         // Reset OAM read/write odd/even selects
    movb %al,OAMReadOdd

    decl FrameCount         // Should we redraw the screen this frame?
    movl %eax,HDMAYCount    // Reset scanline counter
    jnz 0f
    incl FPSCount           // For FPS counter - 0.25b14
// If minimum frameskip set, don't wait for timer to tell us to draw one
    cmpl %eax,_FRAME_SKIP_MIN
    jnz 2f
// Wait for timer to tell us to draw a frame
1:  movl Throttle,%edx
    testl %edx,%edx
// HLT should be more multitasking-friendly but appears
// to cause problems with pure DOS in some cases?
    jz  1b  // 2b
2:  movl %eax,Throttle
    cmpl _FRAME_SKIP_MAX,%edx
    jb  2f
    movl _FRAME_SKIP_MAX,%edx
2:  cmpl _FRAME_SKIP_MIN,%edx
    ja  2f
    movl _FRAME_SKIP_MIN,%edx
2:  orb $8,%al
    movl %edx,FrameCount    // Reset frame counter
#ifdef DEBUG
    incl _Frames
.if FRAME_LIMIT-0
    cmpl $FRAME_LIMIT,_Frames
    jb 0f
    ret
.endif
#endif
0:  movb _SETINI,%dl
    andb $4,%dl             // 224 or 239 lines tall?
    orb %dl,%al
0:  movl 0x80000000(%eax),%eax
.globl _Event_Handler_Table
_Event_Handler_Table=.-4
    movl %eax,Event_Handler
    xorb $0x80,STAT78       // Toggle current field
    RESTORE_HDMA_ADDRESS    // Relatch HDMA registers
    movl _BKG,%edi
    addl $64,%edi
    movl %edi,ScreenDestPtr // Setup destination screen pointer

    movl FPSMaxTicks,%eax
    cmpl %eax,FPSTicks
    jb 0f
    movl FPSCount,%eax
    je 3f
    cmpl $1,%eax
    ja 3f
    movl $0,FPSLast
    jmp 4f
3:  movl %eax,FPSLast
4:  movl $0,FPSTicks
    movl $0,FPSCount

ALIGN
0:  xorl %eax,%eax
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
.globl _IRQHDMARenderFirst
_IRQHDMARenderFirst:
IRQHDMARenderFirst: // Check HDMA, IRQ, Render
    Save_Cycles Execute=0
    movb HVBJOY,%al
    andb $0x7F,%al  // Vblank off
    orb  $1,%al     // Controllers ready?
    movb %al,HVBJOY
    movb $2,RDNMI

    incl HDMAYCount
    call do_HDMA
    RenderScanline
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
.globl _IRQHDMARender
_IRQHDMARender:
IRQHDMARender:      // Check HDMA, IRQ, Render
    Save_Cycles Execute=0

    incl HDMAYCount
    call do_HDMA
    RenderScanline
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
.globl _IRQHDMARenderLast
_IRQHDMARenderLast:
IRQHDMARenderLast:  // Check HDMA, IRQ
    Save_Cycles Execute=0

    incl HDMAYCount
    call do_HDMA
    RenderScanline
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
VBL_First_Line:
    Save_Cycles Execute=0

    incl HDMAYCount
    orb  $0x80,HVBJOY       // VBlank on
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
VBL_Update_Controllers:
    Save_Cycles Execute=0

    call UPDATE_CONTROLLERS
    incl HDMAYCount
    andb $0xFE,HVBJOY   // Controllers not ready?
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
VBL:Save_Cycles Execute=0

    incl HDMAYCount
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
.globl _IRQHDMAFirst
_IRQHDMAFirst:
IRQHDMAFirst:       // Check HDMA, IRQ, Render
    Save_Cycles Execute=0
    movb HVBJOY,%al
    andb $0x7F,%al  // Vblank off
    orb  $1,%al     // Controllers ready?
    movb %al,HVBJOY
    movb $2,RDNMI

    incl HDMAYCount
    call do_HDMA
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
IRQHDMA:            // Check HDMA, IRQ
    Save_Cycles Execute=0

    incl HDMAYCount
    call do_HDMA
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
IRQHDMALast:        // Check HDMA, IRQ
    Save_Cycles Execute=0

    incl HDMAYCount
    call do_HDMA
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGN
EXPORT_C NMI
EXPORT NMI
    pushl %ebp
    movw _Real_SNES_Palette,%ax
    pushl %eax

    movb _fixedpalettecheck,%al
    testb %al,%al
    jz  Palette_Check

Fixed_Back_Color_Add_Hack:
    movw _COLDATA,%ax
    movw %ax,_Real_SNES_Palette
    jmp Do_Palette

Palette_Check:
    cmpb $0,PaletteChanged  // If 1, palette changed!
    jz Do_Copy
Do_Palette:
    movb $0,PaletteChanged

    pusha
    call _SetPalette        // VESA friendly method of setting palette
    popa

Do_Copy:
    cmpb $0,_FPS_ENABLED
    jz  0f
    call _ShowFPS
0:
    call Copy_Screen
//  call _Display_Debug

    popl %eax
    movw %ax,_Real_SNES_Palette
    popl %ebp
.globl NMI_NoRender,_NMI_NoRender
_NMI_NoRender:
NMI_NoRender:
    // Force SPC to catch up to eliminate lags and improve sound
    Save_Cycles Execute=1

//  call UPDATE_JOYSTICKS
//  call UPDATE_CONTROLLERS
#define BUTCHA_SOUND
#ifdef BUTCHA_SOUND
    call _SNDUpdate         // Added by Butcha
#endif

    movb $0x81,%al
    movb %al,HVBJOY         // VBlank on
    movb $0x82,%al
    movb %al,RDNMI          // Set NMI enabled bit in 0x4210
    incl HDMAYCount

CheckNMI:
#ifdef DEBUG
    movb _PB,%al
    movb %al,_OLD_PB        // Save program bank
    movl _PC,%eax
    movl %eax,_OLD_PC       // Save program counter
#endif
    testb $0x80,NMITIMEN
    jz  CHECK_KEYS          // NMI off? (via hardware register)

ALIGN
    movl In_WAI,%eax
    testl %eax,%eax
    jz  0f
    incw _PC
    movl $0,In_WAI
0:  LOAD_P
    JUMP_NOT_FLAG SNES_FLAG_E,E0_NMI    // Are we in emulation mode?

E1_NMI:
    // Emulation mode NMI
    movl _PC,%eax
    E1_PUSH_W
    CLR_FLAG SNES_FLAG_B    // Clear break bit on stack
    E1_SETUPFLAGS           // put flags into SNES packed flag format
    SET_FLAG SNES_FLAG_B
    E1_PUSH_B

    movl _NMI_Evector,%eax  // Get Emulation mode NMI vector
    movl _NMI_Eoffset,%edx
    jmp 2f

ALIGN
E0_NMI:
    // Native mode NMI
    movb _PB,%al
    E0_PUSH_B
    movl _PC,%eax
    E0_PUSH_W
    E0_SETUPFLAGS           // put flags into SNES packed flag format
    E0_PUSH_B

    movl _NMI_Nvector,%eax  // Get Native mode NMI vector
    movl _NMI_Noffset,%edx
2:  movl %eax,_PC           // Setup PC vector
    movl %edx,_PBOffset
    movb $0,_PB             // Setup bank
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    movl %eax,_PB_Read  // Bank 0 read through memory mapper

    SET_FLAG SNES_FLAG_I    // Disable interrupts
    CLR_FLAG SNES_FLAG_D
    SAVE_P

CHECK_KEYS:
    xorb %ah,%ah
    JUMP_KEY_UP KEY_F1,0f,%ah   // Disable/enable channel 0
    movb $1,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F1,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F2,0f,%ah   // Disable/enable channel 1
    movb $2,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F2,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F3,0f,%ah   // Disable/enable channel 2
    movb $4,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F3,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F4,0f,%ah   // Disable/enable channel 3
    movb $8,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F4,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F5,0f,%ah   // Disable/enable channel 4
    movb $0x10,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F5,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F6,0f,%ah   // Disable/enable channel 5
    movb $0x20,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F6,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F7,0f,%ah   // Disable/enable channel 6
    movb $0x40,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F7,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F8,0f,%ah   // Disable/enable channel 7
    movb $0x80,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F8,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_1,0f,%ah    // Toggle BG1 disable
    movb $1,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_1,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_2,0f,%ah    // Toggle BG2 disable
    movb $2,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_2,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_3,0f,%ah    // Toggle BG3 disable
    movb $4,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_3,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_4,0f,%ah    // Toggle BG4 disable
    movb $8,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_4,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_5,0f,%ah    // Disable/enable sprites
    movb $0x10,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_5,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_7,0f,%ah    // Toggle Offset Change
    xorb $0xFF,Offset_Change_Disable
1:  JUMP_KEY_DOWN KEY_7,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_6,0f,%ah    // Reset layers, sprites, channels
    movb $0xFF,%al
    movb $0,Offset_Change_Disable
    movb %al,TM_MASK
    movb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_6,1b,%ah  // Wait for key release

0:  // Update layer mask
    movb _TM,%al
    andb TM_MASK,%al
    movb %al,SCR_TM
    notb %al
    andb _TS,%al
    andb TM_MASK,%al
    movb %al,SCR_TS

    JUMP_KEY_UP KEY_0,0f,%ah    // Save PCX screenshot

    pusha
    call _OutputScreen
    popa

1:  JUMP_KEY_DOWN KEY_0,1b,%ah  // Wait for key release

0:  JUMP_KEY_DOWN KEY_ESC,Emulation_Break,%ah   // Is ESC pressed?

// Return to emulation
    movl HDMAYCount,%eax    // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

Emulation_Break:
0:  JUMP_KEY_DOWN KEY_ESC,0b,%ah    // Wait for key release
    pusha
    movl _PB_Shifted,%ebx
    movw _PC,%bx

    call *_PB_Read
    movb %al,_Map_Byte

    popa

    movl HDMAYCount,%eax    // Get current scanline
    CheckIRQ
    ret
