/* Macro's included in this ASM header

All macros assume ebx contains 24-bit address, al or ax (depending on size)
the location to return data (for read) or accept data (for write) at.
edi is always corrupted.

GET_BYTE    Get byte from any bank
GET_BYTE_00 Get byte from bank 00
GET_BYTE_PB Get byte from program bank
GET_WORD    Get word from any bank
GET_WORD_00 Get word from bank 00
GET_WORD_PB Get word from program bank
SET_BYTE    Set byte in any bank
SET_BYTE_00 Set byte in bank 00
SET_WORD    Set word in any bank
SET_WORD_00 Set word in bank 00

GET_WORD, GET_WORD_00, and GET_WORD_PB may be passed an optional
'Reverse=1' parameter to skip the last xchg. If this is done, the low
byte will be returned in ah, high byte in al.

SET_WORD and SET_WORD_00 may be passed an optional 'NoSave=1' parameter
to prevent the push/pop used to preserve the data word.

SET_WORD and SET_WORD_00 may be passed an optional 'Double=1' parameter
to use al for both high and low bytes of write data. Using this forces
'NoSave=1', because eax is not modified.

SET_WORD_HL and SET_WORD_HL_00 are alternate macros for when the address
being passed in is of the high byte, and the high byte will be written first.
These alternate macros do not support NoSave or Double.

*/

.macro GET_BYTE
    movl %ebx,%edi
    shrl $16,%edi   // Get bank for indexing
    call _Read_BankMapping(,%edi,4)
.endm

.macro GET_BYTE_00
    call *_Read_BankMapping
.endm

.macro GET_BYTE_PB
    call *_PB_Read
.endm

.macro GET_WORD Reverse=0
	GET_BYTE
    movb %al,%ah
    incl %ebx
	GET_BYTE
.if \Reverse-1
    xchgb %al,%ah
.endif
.endm

.macro GET_WORD_00 Reverse=0
    GET_BYTE_00
    movb %al,%ah
    incl %ebx
	GET_BYTE
.if \Reverse-1
    xchgb %al,%ah
.endif
.endm

.macro GET_WORD_PB Reverse=0
    GET_BYTE_PB
    movb %al,%ah
    incl %ebx
    GET_BYTE_PB
.if \Reverse-1
    xchgb %al,%ah
.endif
.endm

.macro SET_BYTE_00
    call *_Write_BankMapping
.endm

.macro SET_BYTE
    movl %ebx,%edi
    shrl $16,%edi
    call _Write_BankMapping(,%edi,4)
.endm

.macro SET_WORD_00 NoSave=0 Double=0
    SET_BYTE_00
.if \Double-1
.if \NoSave-1
    pushl %eax
.endif
    movb %ah,%al
.endif
    incl %ebx
	SET_BYTE
.if \Double-1
.if \NoSave-1
    popl %eax
.endif
.endif
.endm

.macro SET_WORD NoSave=0 Double=0
	SET_BYTE
.if \Double-1
.if \NoSave-1
    pushl %eax
.endif
    movb %ah,%al
.endif
    incl %ebx
	SET_BYTE
.if \Double-1
.if \NoSave-1
    popl %eax
.endif
.endif
.endm

.macro SET_WORD_HL
    pushl %eax
    movb %ah,%al
	SET_BYTE
    popl %eax
    decl %ebx
    SET_BYTE
.endm

.macro SET_WORD_00_HL
    pushl %eax
    movb %ah,%al
    SET_BYTE_00
    popl %eax
    decl %ebx
    SET_BYTE
.endm
