/*

 SNEeSe SNES memory mapper v1.6

 There are numerous jump tables in this code, but a lot less than before...
  back down from 4k to 256... easier on the cache this way, and with
  dynamic memory map init both LoROM and HiROM are very simply mapped...

  Files are split up as follows:

    mem.S      - LoROM memory map code/tables.
    memhi.S    - HiROM memory map code/tables.
    memmap.i   - Memory mapping macros.
    memblank.S - 'Blank' ROM bank handlers for dynamic memory map
                 and code common to both LoROM and HiROM.
    memhard.S  - Hardware port (21xx/40xx/42xx/43xx) handlers.
    memDMA.S   - DMA emulation code.

*/

#include "misc.i"
#include "memmap.i"

EXPORT_C memblank_start

.data
.globl _LoROM_Blank
ALIGN
_LoROM_Blank:
// Banks 00-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,64,_BlankL_00_3F

// Banks 40-5F are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,32,_BlankL_40_5F

// Banks 60-6F are not mapped!
    DUPLICATE long,16,UNSUPPORTED_READ

// Banks 70-73 is SRAM (mirrored)
    DUPLICATE long,4,_SRAM_READ

// Banks 74-7D are not mapped!
    DUPLICATE long,10,UNSUPPORTED_READ

// Bank  7E-7F is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE long,2,_RAM_READ

// Banks 80-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,64,_BlankL_80_BF

// Banks C0-DF are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,64,_BlankL_C0_FF

.globl _HiROM_Blank
ALIGN
_HiROM_Blank:
// Banks 00-2F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
// Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE long,64,_BlankH_00_3F_No_SRAM

// Banks 40-5F are 0000-FFFF ROM
    DUPLICATE long,32,_BlankH_40_5F

// Banks 60-7D are not mapped!
    DUPLICATE long,30,UNSUPPORTED_READ

// Bank  7E-7F is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE long,2,_RAM_READ

// Banks 80-AF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
// Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE long,64,_BlankH_80_BF_No_SRAM

// Banks C0-FF are 0000-FFFF ROM
    DUPLICATE long,64,_BlankH_C0_FF

.bss
ALIGN
.globl _Read_BankMapping
_Read_BankMapping:  skipl 256

ALIGN
.globl _Write_BankMapping
_Write_BankMapping: skipl 256

ALIGN
.globl _Read_BankOffset
_Read_BankOffset:   skipl 256*2

.text
.globl UNSUPPORTED_READ,_UNSUPPORTED_READ
ALIGN
_UNSUPPORTED_READ:
UNSUPPORTED_READ:
    movb $0xFF,%al
#ifdef DEBUG
#ifdef TRAP_BAD_READS
    movl %ebx,_Map_Address  // Set up Map Address so message works!
    movb %al,_Map_Byte      // Set up Map Byte so message works
    pusha
    call _InvalidHWRead     // Unmapped hardware address!
    popa
#endif
#endif
    ret

.globl UNSUPPORTED_WRITE,_UNSUPPORTED_WRITE
ALIGN
_UNSUPPORTED_WRITE:
UNSUPPORTED_WRITE:
#ifdef DEBUG
#ifdef TRAP_BAD_WRITES
    movl %ebx,_Map_Address  // Set up Map Address so message works!
    movb %al,_Map_Byte      // Set up Map Byte so message works
    pusha
    call _InvalidHWWrite    // Unmapped hardware address!
    popa
#endif
#endif
    ret

.globl _RAM_READ
ALIGN
_RAM_READ:
    movb _WRAM-0x7E0000(%ebx),%al
    ret

.globl _RAM_WRITE
ALIGN
_RAM_WRITE:
    movb %al,_WRAM-0x7E0000(%ebx)
    ret

.globl _READ_00_3F_No_SRAM
ALIGN
_READ_00_3F_No_SRAM:
    testb %bh,%bh
    jns 0f
    ROM00_READ_32k
ALIGN
0:  cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  NON_MAPPED_READ

.globl _READ_80_BF_No_SRAM
ALIGN
_READ_80_BF_No_SRAM:
    testb %bh,%bh
    jns 0f
    ROM80_READ_32k
ALIGN
0:  cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  NON_MAPPED_READ

.globl _BlankL_00_3F,_BlankL_80_BF,_BlankH_00_3F_No_SRAM,_BlankH_80_BF_No_SRAM
ALIGN
_BlankL_00_3F:
_BlankL_80_BF:
_BlankH_00_3F_No_SRAM:
_BlankH_80_BF_No_SRAM:
    cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  testb %bh,%bh
    jns 0f
    NO_ROM_READ
ALIGN
0:  NON_MAPPED_READ

ALIGN
_BlankL_40_5F:
_BlankL_C0_FF:
    testb %bh,%bh
    jns 0f
    NO_ROM_READ
ALIGN
0:  NON_MAPPED_READ

.macro _BlankH_30_3F bank
ALIGN
.globl _BlankH_\bank
_BlankH_\bank:
    cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  testb %bh,%bh
    js  0f
    SRAM30_READ \bank
ALIGN
0:  NO_ROM_READ
.endm

.macro _BlankH_B0_BF bank
ALIGN
.globl _BlankH_\bank
_BlankH_\bank:
    cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  testb %bh,%bh
    js  0f
    SRAMB0_READ \bank
ALIGN
0:  NO_ROM_READ
.endm

ALIGN
.globl _BlankH_40_5F,_BlankH_C0_FF
_BlankH_40_5F:
_BlankH_C0_FF:
    NO_ROM_READ

    _BlankH_30_3F 30
    _BlankH_30_3F 31
    _BlankH_30_3F 32
    _BlankH_30_3F 33
    _BlankH_30_3F 34
    _BlankH_30_3F 35
    _BlankH_30_3F 36
    _BlankH_30_3F 37
    _BlankH_B0_BF B0
    _BlankH_B0_BF B1
    _BlankH_B0_BF B2
    _BlankH_B0_BF B3
    _BlankH_B0_BF B4
    _BlankH_B0_BF B5
    _BlankH_B0_BF B6
    _BlankH_B0_BF B7

.globl _WRITE_00_3F,_WRITE_80_BF,_HWRITE_00_3F_No_SRAM,_HWRITE_80_BF_No_SRAM
ALIGN
_WRITE_00_3F:           // Used in LoROM!
_WRITE_80_BF:
_HWRITE_00_3F_No_SRAM:  // Used in HiROM!
_HWRITE_80_BF_No_SRAM:
    cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_WRITE
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_WRITE
ALIGN
0:  ROM_WRITE

.text; ALIGN
.data; ALIGN
.bss; ALIGN
