/*

 SNEeSe SNES memory mapper v1.6

 Notes: HiROM has max ROM address space of 4MBytes.

 HiROM version
  00-3F:0000-1FFF = Low RAM 8k (7E:0000-1FFF mirrored)
  00-3F:2000-5FFF = Hardware ports (PPU, FX, DSP)
  00-2F:6000-7FFF = Unmapped
  30-3F:6000-7FFF = SRAM (up to 8k/bank) (continuously mirrored: see below)
  00-3F:8000-FFFF = ROM 32k (C0-FF:8000-FFFF mirrored)
  40-5F:0000-FFFF = ROM 64k (0-16Mbit)
  60-7D:0000-FFFF = Unmapped
  7E-7F:0000-FFFF = WRAM 128k
  80-DF:0000-FFFF = 00-5F mirrored
  E0-FF:0000-FFFF = ROM 64k (16-32Mbit)

  SRAM has 128k (kbytes, not kbits) of address space. This is mapped
   with 'continuous mirroring', as such:
    2k (16kbit): 30:6000-67FF == 30:6800-6FFF == 3F:7800-7FFF
    8k (64kbit): 30:6000-7FFF == 31:6000-7FFF == 3F:6000-7FFF
   etc.
  Since I know of no ROMs that require >64k of SRAM, only 64k is
   supported (however, 128k mirroring is performed).
*/

#include "misc.i"
#include "memmap.i"

EXPORT_C memhi_start

.data
.globl _HiROM_Read
ALIGN
_HiROM_Read:

// Banks 00-2F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
// Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE long,64,_READ_00_3F_No_SRAM

// Banks 40-5F are 0000-FFFF ROM
    DUPLICATE long,32,_HREAD_40_5F

// Banks 60-7D are not mapped!
    DUPLICATE long,30,UNSUPPORTED_READ

// Bank  7e-7f is  0000-ffff RAM Nb 7e is used to store that 8KB block thats repeated in mem
    DUPLICATE long,2,_RAM_READ

// Banks 80-AF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
// Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE long,64,_READ_80_BF_No_SRAM

// Banks C0-FF are 0000-FFFF ROM
    DUPLICATE long,64,_HREAD_C0_FF

.globl _HiROM_Write
ALIGN
_HiROM_Write:

// Banks 00-2F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
// Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE long,64,_HWRITE_00_3F_No_SRAM

// Banks 40-5F are 0000-FFFF ROM
    DUPLICATE long,32,UNSUPPORTED_WRITE

// Banks 60-7D are not mapped!
    DUPLICATE long,30,UNSUPPORTED_WRITE

// Bank  7e-7f is  0000-ffff RAM Nb 7e is used to store that 8KB block thats repeated in mem
    DUPLICATE long,2,_RAM_WRITE

// Banks 80-AF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
// Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    DUPLICATE long,64,_HWRITE_80_BF_No_SRAM

// Banks C0-FF are 0000-FFFF ROM
    DUPLICATE long,64,UNSUPPORTED_WRITE

.globl _HiROM_Read_SRAM_ROM_30,_HiROM_Read_No_SRAM_ROM_30
.globl _HiROM_Read_SRAM_ROM_B0,_HiROM_Read_No_SRAM_ROM_B0
.globl _HiROM_Read_SRAM_Blank_30,_HiROM_Read_No_SRAM_Blank_30
.globl _HiROM_Read_SRAM_Blank_B0,_HiROM_Read_No_SRAM_Blank_B0
.globl _HiROM_Write_SRAM_30,_HiROM_Write_No_SRAM_30
.globl _HiROM_Write_SRAM_B0,_HiROM_Write_No_SRAM_B0
ALIGN
_HiROM_Read_SRAM_ROM_30:
    .long _HREAD_30
    .long _HREAD_31
    .long _HREAD_32
    .long _HREAD_33
    .long _HREAD_34
    .long _HREAD_35
    .long _HREAD_36
    .long _HREAD_37
_HiROM_Read_SRAM_ROM_B0:
    .long _HREAD_B0
    .long _HREAD_B1
    .long _HREAD_B2
    .long _HREAD_B3
    .long _HREAD_B4
    .long _HREAD_B5
    .long _HREAD_B6
    .long _HREAD_B7
_HiROM_Read_SRAM_Blank_30:
    .long _BlankH_30
    .long _BlankH_31
    .long _BlankH_32
    .long _BlankH_33
    .long _BlankH_34
    .long _BlankH_35
    .long _BlankH_36
    .long _BlankH_37
_HiROM_Read_SRAM_Blank_B0:
    .long _BlankH_B0
    .long _BlankH_B1
    .long _BlankH_B2
    .long _BlankH_B3
    .long _BlankH_B4
    .long _BlankH_B5
    .long _BlankH_B6
    .long _BlankH_B7
_HiROM_Write_SRAM_30:
    .long _HWRITE_30
    .long _HWRITE_31
    .long _HWRITE_32
    .long _HWRITE_33
    .long _HWRITE_34
    .long _HWRITE_35
    .long _HWRITE_36
    .long _HWRITE_37
_HiROM_Write_SRAM_B0:
    .long _HWRITE_B0
    .long _HWRITE_B1
    .long _HWRITE_B2
    .long _HWRITE_B3
    .long _HWRITE_B4
    .long _HWRITE_B5
    .long _HWRITE_B6
    .long _HWRITE_B7
_HiROM_Read_No_SRAM_ROM_30:
    .long _READ_00_3F_No_SRAM
_HiROM_Read_No_SRAM_ROM_B0:
    .long _READ_80_BF_No_SRAM
_HiROM_Read_No_SRAM_Blank_30:
    .long _BlankH_00_3F_No_SRAM
_HiROM_Read_No_SRAM_Blank_B0:
    .long _BlankH_80_BF_No_SRAM
_HiROM_Write_No_SRAM_30:
    .long _HWRITE_00_3F_No_SRAM
_HiROM_Write_No_SRAM_B0:
    .long _HWRITE_80_BF_No_SRAM

.text
/* ebx - contains the actual address, al is where the info should be stored, edi is free */
/* NB ebx is not corrupted! edi is corrupted! */
/* NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc! */

.macro _HREAD_30_3F bank
.globl _HREAD_\bank
ALIGN
_HREAD_\bank:
    testb %bh,%bh
    jns 0f
    ROM00_READ_32k
ALIGN
0:  cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  SRAM30_READ \bank
.endm

.globl _HREAD_40_5F
ALIGN
_HREAD_40_5F:
    HIROM40_LIN_READ

.macro _HREAD_B0_BF bank
.globl _HREAD_\bank
ALIGN
_HREAD_\bank:
    testb %bh,%bh
    jns 0f
    ROM80_READ_32k
ALIGN
0:  cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  SRAMB0_READ \bank
.endm

.globl _HREAD_C0_FF
ALIGN
_HREAD_C0_FF:
    HIROMC0_LIN_READ

// Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HREAD_30_3F 30
    _HREAD_30_3F 31
    _HREAD_30_3F 32
    _HREAD_30_3F 33
    _HREAD_30_3F 34
    _HREAD_30_3F 35
    _HREAD_30_3F 36
    _HREAD_30_3F 37
// Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HREAD_B0_BF B0
    _HREAD_B0_BF B1
    _HREAD_B0_BF B2
    _HREAD_B0_BF B3
    _HREAD_B0_BF B4
    _HREAD_B0_BF B5
    _HREAD_B0_BF B6
    _HREAD_B0_BF B7

.macro _HWRITE_30_3F bank
.globl _HWRITE_\bank
ALIGN
_HWRITE_\bank:
    cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_WRITE
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_WRITE
ALIGN
0:  testb %bh,%bh
    js  1f
    SRAM30_WRITE \bank
ALIGN
1:  ROM_WRITE
.endm

.macro _HWRITE_B0_BF bank
.globl _HWRITE_\bank
ALIGN
_HWRITE_\bank:
    cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_WRITE
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_WRITE
ALIGN
0:  testb %bh,%bh
    js  1f
    SRAMB0_WRITE \bank
ALIGN
1:  ROM_WRITE
.endm

// Banks 30-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HWRITE_30_3F 30
    _HWRITE_30_3F 31
    _HWRITE_30_3F 32
    _HWRITE_30_3F 33
    _HWRITE_30_3F 34
    _HWRITE_30_3F 35
    _HWRITE_30_3F 36
    _HWRITE_30_3F 37
// Banks B0-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF SRAM 8000-FFFF ROM
    _HWRITE_B0_BF B0
    _HWRITE_B0_BF B1
    _HWRITE_B0_BF B2
    _HWRITE_B0_BF B3
    _HWRITE_B0_BF B4
    _HWRITE_B0_BF B5
    _HWRITE_B0_BF B6
    _HWRITE_B0_BF B7

.text; ALIGN
.data; ALIGN
.bss; ALIGN
