/*

 SNEeSe SNES memory mapper v1.6

 Notes: LoROM has max ROM address space of 4MBytes.

 LoROM version
  00-3F:0000-1FFF = Low RAM 8k (7E:0000-1FFF mirrored)
  00-3F:2000-5FFF = Hardware ports (PPU, FX, DSP)
  00-3F:6000-7FFF = Unmapped
  00-3F:8000-FFFF = ROM 32k (0-16Mbit)
  40-5F:8000-FFFF = ROM 32k (16-24Mbit)
  60-6F:0000-FFFF = Unmapped
  70-73:0000-FFFF = SRAM (up to 64k) (continuously mirrored: see below)
  74-7D:0000-FFFF = Unmapped
  7E-7F:0000-FFFF = WRAM 128k
  80-DF:0000-FFFF = 00-5F mirrored
  E0-FF:8000-FFFF = ROM 32k (24-32Mbit)

  SRAM has 256k (kbytes, not kbits) of address space. This is mapped
   with 'continuous mirroring', as such:
    2k (16kbit): 70:0000-07FF == 70:0800-0FFF == 73:F800-FFFF
    8k (64kbit): 70:0000-1FFF == 70:2000-3FFF == 73:E000-FFFF
   etc.
  Since I know of no ROMs that require >64k of SRAM, only 64k is
   supported (however, 256k mirroring is performed).
  Note: I HAVE discovered one ROM requiring 128k of SRAM (Dezaemon),
   will implement this later...
*/

#include "misc.i"
#include "memmap.i"

EXPORT_C memlo_start

.data
// _LoROM_Read is _LoROM_Blank - see MemBlank.S

.globl _LoROM_Write
ALIGN
_LoROM_Write:

// Banks 00-3F are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,64,_WRITE_00_3F

// Banks 40-5F are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,32,UNSUPPORTED_WRITE

// Banks 60-6F are not mapped!
    DUPLICATE long,16,UNSUPPORTED_WRITE

// Banks 70-73 is SRAM (mirrored)
    DUPLICATE long,4,_SRAM_WRITE

// Banks 74-7D are not mapped!
    DUPLICATE long,10,UNSUPPORTED_WRITE

// Bank  7E-7f is  0000-FFFF RAM Nb 7E is used to store that 8KB block thats repeated in mem
    DUPLICATE long,2,_RAM_WRITE

// Banks 80-BF are 0000-1FFF LOW RAM 2000-2FFF PPU 3000-5FFF FX/DSP 6000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,64,_WRITE_80_BF

// Banks C0-DF are 0000-7FFF RESERVED 8000-FFFF ROM
    DUPLICATE long,32,UNSUPPORTED_WRITE

// Rest are unmapped
    DUPLICATE long,32,UNSUPPORTED_WRITE

.text
/* ebx - contains the actual address, al is where the info should be stored, edi is free */
/* NB ebx is not corrupted! edi is corrupted! */
/* NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc! */

.globl _SRAM_READ
ALIGN
_SRAM_READ:
    movl _SRAM_Mask,%edi
    andl %ebx,%edi
    movb _SRAM(%edi),%al
    ret

.globl _SRAM_WRITE
ALIGN
_SRAM_WRITE:
    movl _SRAM_Mask,%edi
    andl %ebx,%edi
    movb %al,_SRAM(%edi)
    ret

/* ebx - contains the actual address, al is where the info should be stored, edi is free */
/* NB ebx is not corrupted! edi is corrupted! */
/* NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc! */

.globl _READ_00_3F_2
ALIGN
_READ_00_3F_2:
    testb %bh,%bh
    jns 0f
    ROM00_READ_32k_2
ALIGN
0:  cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  NON_MAPPED_READ

.globl _READ_40_5F_2    // v0.17: corrected 16-24Mbit
ALIGN
_READ_40_5F_2:
    testb %bh,%bh
    jns 0f
    ROM00_READ_32k_2
ALIGN
0:  NON_MAPPED_READ

.globl _READ_80_BF_2
ALIGN
_READ_80_BF_2:
    testb %bh,%bh
    jns 0f
    ROM80_READ_32k_2
ALIGN
0:  cmpb $0x20,%bh
    jnb 0f
    LOW_RAM_READ
ALIGN
0:  cmpb $0x60,%bh
    jnb 0f
    PPU_READ
ALIGN
0:  NON_MAPPED_READ

.globl _READ_C0_FF_2    // v0.17: corrected 16-32Mbit
ALIGN
_READ_C0_FF_2:
    testb %bh,%bh
    jns 0f
    ROM80_READ_32k_2
ALIGN
0:  NON_MAPPED_READ

.text; ALIGN
.data; ALIGN
.bss; ALIGN
