#include "misc.i"

// Read Low RAM - 0000-1FFF in 00-3F/80-BF
.macro LOW_RAM_READ
    movl $0x1FFF,%edi
    andl %ebx,%edi
    movb _WRAM(%edi),%al
    ret
.endm

// Write Low RAM - 0000-1FFF in 00-3F/80-BF
.macro LOW_RAM_WRITE
    movl $0x1FFF,%edi
    andl %ebx,%edi
    movb %al,_WRAM(%edi)
    ret
.endm

// Read hardware - 2000-5FFF in 00-3F/80-BF
.macro PPU_READ
    movl $0xFFFF,%edi
    andl %ebx,%edi
    jmp Read_Map_20_5F-0x2000*4(,%edi,4)
.endm

// Write hardware - 2000-5FFF in 00-3F/80-BF
.macro PPU_WRITE
    movl $0xFFFF,%edi
    andl %ebx,%edi
    jmp Write_Map_20_5F-0x2000*4(,%edi,4)
.endm

/*
// Read hardware - 2100-21FF in 00-3F/80-BF
.macro PPU_21_READ
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Read_Map_21(,%edi,4)
.endm

// Write hardware - 2100-21FF in 00-3F/80-BF
.macro PPU_21_WRITE
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Write_Map_21(,%edi,4)
.endm

// Read hardware - 4000-40FF in 00-3F/80-BF
.macro PPU_40_READ
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Read_Map_40(,%edi,4)
.endm

// Write hardware - 4000-40FF in 00-3F/80-BF
.macro PPU_40_WRITE
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Write_Map_40(,%edi,4)
.endm

// Read hardware - 4200-42FF in 00-3F/80-BF
.macro PPU_42_READ
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Read_Map_42(,%edi,4)
.endm

// Write hardware - 4200-42FF in 00-3F/80-BF
.macro PPU_42_WRITE
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Write_Map_42(,%edi,4)
.endm

// Read hardware - 4300-43FF in 00-3F/80-BF
.macro PPU_43_READ
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Read_Map_43(,%edi,4)
.endm

// Write hardware - 4300-43FF in 00-3F/80-BF
.macro PPU_43_WRITE
    movl $0xFF,%edi
    andl %ebx,%edi
    jmp Write_Map_43(,%edi,4)
.endm
*/

// Read LoROM ROM 00-5F (first half of banks)
// Read HiROM ROM 00-3F
.macro ROM00_READ_32k
    movl _RomReadAdjust,%edi
    movb (%ebx,%edi),%al
    ret
.endm

// Read LoROM ROM 00-5F (second half of banks)
.macro ROM00_READ_32k_2
    movl _RomReadAdjust2,%edi
    movb (%ebx,%edi),%al
    ret
.endm

// Read LoROM ROM 80-FF (first half of banks)
// Read HiROM ROM 80-BF
.macro ROM80_READ_32k
    movl _RomHiReadAdjust,%edi
    movb (%ebx,%edi),%al
    ret
.endm

// Read LoROM ROM 80-FF (second half of banks)
.macro ROM80_READ_32k_2
    movl _RomHiReadAdjust2,%edi
    movb (%ebx,%edi),%al
    ret
.endm

// Read HiROM ROM 40-5F
.macro HIROM40_LIN_READ
    movl _HRomLinReadAdjust,%edi
    movb (%ebx,%edi),%al
    ret
.endm

// Read HiROM ROM C0-FF
.macro HIROMC0_LIN_READ
    movl _HRomHiLinReadAdjust,%edi
    movb (%ebx,%edi),%al
    ret
.endm

// Read from ROM address space not containing ROM
.macro NO_ROM_READ
    movb $0xFF,%al
    ret
.endm

// Write to ROM address space
.macro ROM_WRITE
    ret
.endm

// Read HiROM SRAM 30-3F
.macro SRAM30_READ num
    movl _SRAM_Mask,%edi
    andl %ebx,%edi
    movb _SRAM+((0x\num-0x30)*0x2000)(%edi),%al
    ret
.endm

// Write HiROM SRAM 30-3F
.macro SRAM30_WRITE num
    movl _SRAM_Mask,%edi
    andl %ebx,%edi
    movb %al,_SRAM+((0x\num-0x30)*0x2000)(%edi)
    ret
.endm

// Read HiROM SRAM B0-BF
.macro SRAMB0_READ num
    movl _SRAM_Mask,%edi
    andl %ebx,%edi
    movb _SRAM+((0x\num-0xB0)*0x2000)(%edi),%al
    ret
.endm

// Write HiROM SRAM B0-BF
.macro SRAMB0_WRITE num
    movl _SRAM_Mask,%edi
    andl %ebx,%edi
    movb %al,_SRAM+((0x\num-0xB0)*0x2000)(%edi)
    ret
.endm

// Non-mapped/unsupported read
.macro NON_MAPPED_READ
    movb $0xFF,%al
    ret
.endm

// Non-mapped/unsupported write
.macro NON_MAPPED_WRITE
    ret
.endm
