/*

 Screen mode setup/blitting code
 -Handlers to set up special display modes;
 -Virtual SNES screen -> PC screen copiers.

*/

#include "misc.i"

EXPORT_C scrmode_start

ALIGN
.globl Clear_Screen
Clear_Screen:
    pushl %eax
    pushl %ecx
    pushl %edi

	movl _SNES_Screen,%edi
    xorl %eax,%eax

    movl _ScreenX,%ecx
    shrl $2,%ecx
    imull _ScreenY,%ecx
//  rep; stosl
ALIGN
0:  subl $4,%ecx
    movl %eax,(%edi,%ecx)
    jz  0f
    subl $4,%ecx
    movl %eax,(%edi,%ecx)
    jz  0f
    subl $4,%ecx
    movl %eax,(%edi,%ecx)
    jz  0f
    subl $4,%ecx
    movl %eax,(%edi,%ecx)
    jnz 0b
0:  

    popl %edi
    popl %ecx
    popl %eax
	ret

ALIGN
.globl Copy_Screen
Copy_Screen:
	pusha
    movb _PC_SCREEN_MODE,%al
    andl $7,%eax
    jmpl Copy_Screen_Table(,%eax,4)

ALIGN
VGA_COPY:   // Directly blit to VGA memory
    pushl %es
    pushl %gs
    popl %es

	movl _ScreenX,%edx
    decb %dh

    movl %edx,%eax
	movl _BKG,%esi
    shrl $1,%eax
    addl $0xA0000,%eax
    movl %eax,%edi              // Offset to center of actual screen
    addl $64,%esi               // Offset to center of virtual screen

    movl _ScreenY,%eax
0:  movl $(256/4),%ecx
    rep; movsl

    addl %edx,%edi  // Adjust to next line
    addl $128,%esi  // Sixteen tile positions added for next line
    decl %eax
    jnz 0b

    popl %es
	popa
	ret

ALIGN
SQUASH:     // This routine is specialized! 320x200 squash copy
    pushl %es
    pushl %gs
    popl %es
	movl _BKG,%esi
    movl $0xA0000+(320-256)/2,%edi  // Offset to center of actual screen
    addl $64,%esi                   // Offset to center of virtual screen
    movb $40,%dh
0:  // Outer Loop
    movb $5,%dl
2:  // Inner Loop
    movl $(256/4),%ecx
    rep; movsl

    addl $(320-256),%edi    // Adjust to next line
    addl $128,%esi          // Sixteen tile positions added for next line
    decb %dl
	jnz 2b

	addl $(256+2*64),%esi		/* Skip a scan line */

    decb %dh
	jnz 0b

    popl %es
	popa
	ret

ALIGN
MODEX_COPY:
	movl _ScreenX,%edx
    decb %dh

	movl _SNES_Screen,%edi
    movl %edx,%eax
	movl _BKG,%esi
    shrl $1,%eax
    addl %eax,%edi              // Offset to center of actual screen
    addl $64,%esi               // Offset to center of virtual screen

	movl _ScreenY,%eax
0:  movl $(256/4),%ecx
    rep; movsl

    addl %edx,%edi  // Adjust to next line
    addl $128,%esi  // Sixteen tile positions added for next line
	decl %eax
	jnz 0b

	popa
    jmp _CopySNESScreen // Have Allegro blit the screen and ret for us

ALIGN
HICOLOUR0_COPY:     // Temp hack, converts 256 cols to 65536 col screen

/* 320x200 screen... */

	movl _SNES_Screen,%edi
	movl _BKG,%esi
    addl $(320-256),%edi        // Offset to center of actual screen
    addl $64,%esi               // Offset to center of virtual screen
	xorl %eax,%eax

	movb $200,%dl
0:  movb $0,%dh         // do 256 times
1:  movb (%esi),%al
	incl %esi
    movw _HICOLOUR_Palette(,%eax,4),%bx
    movw %bx,(%edi)
	addl $2,%edi
   	decb %dh
	jnz 1b

    addl $(320-256)*2,%edi  // Adjust to next line
    addl $128,%esi          // Sixteen tile positions added for next line
	decb %dl
	jnz 0b	

	popa
    jmp _CopySNESScreen // Have Allegro blit the screen and ret for us

ALIGN
HICOLOUR1_COPY:     // Temp hack, converts 256 cols to 65536 col screen

/* 320x240 screen... */

	movl _SNES_Screen,%edi
	movl _BKG,%esi
    addl $(320-256),%edi        // Offset to center of actual screen
    addl $64,%esi               // Offset to center of virtual screen
	xorl %eax,%eax

    movb $240,%dl
0:  movb $0,%dh         // do 256 times
1: 	movb (%esi),%al
	incl %esi
    movw _HICOLOUR_Palette(,%eax,4),%bx
    movw %bx,(%edi)
	addl $2,%edi
   	decb %dh
	jnz 1b

    addl $(320-256)*2,%edi  // Adjust to next line
    addl $128,%esi          // Sixteen tile positions added for next line
	decb %dl
	jnz 0b	

	popa
    jmp _CopySNESScreen // Have Allegro blit the screen and ret for us

ALIGN
HICOLOUR2_COPY:     // Temp hack, converts 256 cols to 65536 col screen

/* 640x480 screen... 256 used 384 remains */

	movl _SNES_Screen,%edi
	movl _BKG,%esi
    addl $(640-256)+(640*240),%edi  // Offset to center of actual screen
    addl $64,%esi                   // Offset to center of virtual screen
	xorl %eax,%eax

	movb $240,%dl
0:  movb $0,%dh         // do 256 times
1:	movb (%esi),%al
	incl %esi
    movw _HICOLOUR_Palette(,%eax,4),%bx
    movw %bx,(%edi)
	addl $2,%edi
	decb %dh
	jnz 1b

    addl $(640-256)*2,%edi  // Adjust to next line
    addl $128,%esi          // Sixteen tile positions added for next line
	decb %dl
	jnz 0b	

	popa
    jmp _CopySNESScreen // Have Allegro blit the screen and ret for us

ALIGN
STRETCH_COPY:       // Temp hack, converts 256 cols to 65536 col screen

/* 640x480 screen... 256*2 used 128 remains */

	movl _SNES_Screen,%edi
	movl _BKG,%esi
    addl $(640-512),%edi        // Offset to center of actual screen
    addl $64,%esi               // Offset to center of virtual screen
	xorl %eax,%eax

	movb $240,%dl
0:  movb $0,%dh         // do 256 times
1:	movb (%esi),%al
	incl %esi
    movl _HICOLOUR_Palette(,%eax,4),%ebx
	movl %ebx,(%edi)
    movl %ebx,640*2(%edi)
	addl $4,%edi
	decb %dh
	jnz 1b

    addl $128*2+(640*2),%edi    // Adjust to next line
    addl $128,%esi              // Sixteen tile positions added for next line
	decb %dl
    jnz 0b

	popa
    jmp _CopySNESScreen // Have Allegro blit the screen and ret for us


Copy_Screen_Table:
.long VGA_COPY,SQUASH,MODEX_COPY,VGA_COPY
.long HICOLOUR0_COPY,HICOLOUR1_COPY,HICOLOUR2_COPY,STRETCH_COPY

ALIGN
.globl _Set256x239
_Set256x239:    // This sets up a 256x239 linear VGA mode
 pusha

 // 0 -> Display -> Overscan -> Blank -> Retrace -> End Blank -> Overscan

 // Disable CRTC write protect
 movl $0x3D4,%edx
 movb $0x11,%al
 outb %al,%dx
 incl %edx
 inb  %dx,%al
 andb $0x7F,%al
 outb %al,%dx

 // Start dumping regs to VGA
 // Horizontal Total: 95 (+5) clocks
 //  Display Enable: 63 (+1) clocks
 //  Start Blanking: Clock 64   End Blanking: 34
 //  Start Retrace:  Clock 78   End Retrace:  22
 //  Display Enable Skew: 0     Retrace Skew: 0
 // Vertical Total: 525 (+2) clocks
 //  Display Enable: 477 (+1) rows
 //  Start Blanking: Row 478    End Blanking: x0D
 //  Start Retrace:  Row 478    End Retrace:  0010b

 movl $0x3C2,%edx
 // - V sync, - H sync, 25.175MHz clock, enable memory, CGA CRTC addressing
 // 0: CRTC addressing mode (0:MDA 3B4/1:CGA 3D4): CGA 3D4
 // 1: Display memory access enable: enabled
 // 2-3: Dot clock frequency select:
 //  00: 25.175MHz (640 pixel wide modes)
 //  01: 28.322MHz (720 pixel wide modes)
 //  10, 11: External clock selects (max 65MHz)
 // 4: Reserved = 0
 // 5: Address extension bit: 0
 // 6,7: Horizontal, Vertical sync polarity selects
 //  0 = +, 1 = -
 //  00 = ???   01 = 400   10 = 350    11 = 480
 movb $0xE3,%al     // Misc. Output
 outb %al,%dx
 movl $0x3D4,%edx   // CRTC
 movb $0,%al        // Horizontal Total
 movb $0x5F,%ah     // 95 (+5) clocks
 outw %ax,%dx
 movb $1,%al        // Horizontal Display Enable
 movb $0x3F,%ah     // 63 (+1) clocks
 outw %ax,%dx
 movb $2,%al        // Start Horizontal Blanking
 // Primary centering control - increase number to move left
 movb $0x40,%ah     // Clock 64
 outw %ax,%dx
 movb $3,%al        // End Horizontal Blanking
 // 0-4: End Horizontal Blanking = 34 (bits 0-4 of 6)
 // 5-6: Display Enable Skew = 0 clocks
 // 7: Reserved/Lightpen (must be set for normal operation)
 movb $0x82,%ah
 outw %ax,%dx
 movb $4,%al        // Start Horizonal Retrace Pulse
 movb $0x4E,%ah     // Clock 78
 outw %ax,%dx
 movb $5,%al        // End Horizontal Retrace
 // 0-4: End Horizontal Retrace = 22 (clocks?)
 // 5-6: Horizontal Retrace Skew = 0 clocks
 // 7: End Horizontal Blanking = 34 (bit 5 of 6)
 movb $0x96,%ah
 outw %ax,%dx
 movb $6,%al        // Vertical Total (bits 0-7 of 10)
 movb $0x0D,%ah     // 525 (+2)
 outw %ax,%dx
 movb $7,%al        // Overflow
 // 0,5: Vertical Total = 525 rows (+2) (bits 8-9 of 10)
 // 1,6: Vertical Display Enable End = Row 479 (bits 8-9 of 10)
 // 2,7: Vertical Retrace Start = Row 480 (bits 8-9 of 10)
 // 3: Start Vertical Blanking = Row 480 (bit 8 of 10)
 // 4: Line Compare = (bit 8 of 10)
 movb $0x3E,%ah
 outw %ax,%dx
 movb $8,%al        // Preset Row Scan
 // 0-4: Preset Row Scan = 0
 // 5-6: Byte Panning Control = 0
 // 7: Reserved = 0
 movb $0,%ah
 outw %ax,%dx
 movb $9,%al        // Maximum Scan Line
 // 0-4: Maximum Scan Lines = 1 (+1)
 // 5: Start Vertical Blanking = Row 480 (bit 9 of 10)
 // 6: Line Compare = () (bit 9 of 10)
 // 7: Double Scan Line Display = 0
 movb $0x41,%ah
 outw %ax,%dx
 movb $0x10,%al     // Vertical Retrace Start (bits 0-7 of 10)
 // Primary centering control - increase number to move up
 movb $0xEA,%ah     // Row 490
 outw %ax,%dx
 movb $0x11,%al     // Vertical Retrace End
 // 0-3: Vertical Retrace End = 0010
 // 4: Clear Vertical Interrupt: state cleared
 // 5: Vertical Interrupt: disabled
 // 6: Refresh Bandwidth Select: 3 memory refresh cycles/scan line
 // 7: CRTC Register Protect: enabled
 movb $0xA2,%ah
 outw %ax,%dx
 movb $0x12,%al     // Vertical Display Enable End (bits 0-7 of 10)
 movb $0xDD,%ah     // Row 477
 outw %ax,%dx
 movb $0x13,%al     // Offset
 movb $0x20,%ah     // 32 * 8 = 256
 outw %ax,%dx
 movb $0x14,%al     // Underline Location
 // 0-4: Underline Location = 0
 // 5: Display Address Dwell 4 (clocks per address inc): 0 (1)
 // 6: Display Address Unit 4 (unit size of address inc): 1 (4)
 // 7: Reserved = 0
 movb $0x40,%ah
 outw %ax,%dx
 movb $0x15,%al     // Start Vertical Blanking (bits 0-7 of 10)
 movb $0xDD,%ah     // Row 477
 outw %ax,%dx
 movb $0x16,%al     // End Vertical Blanking
 movb $0x0D,%ah     // Row x0D
 outw %ax,%dx
 movb $0x17,%al     // CRTC Mode Control
 // 0: Display address bit 13 = Row bit 0: disabled
 // 1: Display address bit 14 = Row bit 1: disabled
 // 2: Row counter increment frequency (by line or by 2): by line
 // 3: Display Address Dwell 2 (clocks per address inc): 0 (1)
 // 4: Reserved = 0
 // 5: Display Address Unit Bank Selection (8k/32k): 32k
 // 6: Display Address Unit 2 (unit size of address inc): 0 (4)
 // 7: Reset/generate retrace signals: Generate
 movb $0xA3,%ah
 outw %ax,%dx
 movl $0x3C4,%edx    // Sequencer
 movb $1,%al
 movb $1,%ah
 outw %ax,%dx
 movb $4,%al
 movb $0x0E,%ah
 outw %ax,%dx
 movl $0x3CE,%edx    // Graphics controller
 movb $5,%al
 movb $0x40,%ah
 outw %ax,%dx
 movb $6,%al
 movb $5,%ah
 outw %ax,%dx
 movl $0x3DA,%edx
 inb %dx,%al
 movl $0x3C0,%edx
 movb $0x30,%al
 outb %al,%dx
 movb $0x41,%al
 outb %al,%dx
 movl $0x3DA,%edx
 inb %dx,%al
 movl $0x3C0,%edx
 movb $0x33,%al
 outb %al,%dx
 movb $0,%al
 outb %al,%dx
 popa
 ret

ALIGN
.globl _Set256x224
_Set256x224:    // This sets up a 256x224 linear VGA mode
 pusha

 // 0 -> Display -> Overscan -> Blank -> Retrace -> End Blank -> Overscan

 // Disable CRTC write protect
 movl $0x3D4,%edx
 movb $0x11,%al
 outb %al,%dx
 incl %edx
 inb  %dx,%al
 andb $0x7F,%al
 outb %al,%dx

 // Start dumping regs to VGA
 movl $0x3C2,%edx
 // - V sync, - H sync, 25.175MHz clock, enable memory, CGA CRTC addressing
 movb $0xE3,%al     // Misc. Output
 outb %al,%dx
 movl $0x3D4,%edx   // CRTC
 movb $0,%al        // Horizontal Total
 movb $0x5F,%ah     // 95 (+5) clocks
 outw %ax,%dx
 movb $1,%al        // Horizontal Display Enable
 movb $0x3F,%ah     // 63 (+1) clocks
 outw %ax,%dx
 movb $2,%al        // Start Horizontal Blanking
 // Primary centering control - increase number to move left
 movb $0x40,%ah     // Clock 64
 outw %ax,%dx
 movb $3,%al        // End Horizontal Blanking
 // 0-4: End Horizontal Blanking = 34 (bits 0-4 of 6)
 // 5-6: Display Enable Skew = 0 clocks
 // 7: Reserved/Lightpen (must be set for normal operation)
 movb $0x82,%ah
 outw %ax,%dx
 movb $4,%al        // Start Horizonal Retrace Pulse
 movb $0x4E,%ah     // Clock 78
 outw %ax,%dx
 movb $5,%al        // End Horizontal Retrace
 // 0-4: End Horizontal Retrace = 22 (clocks?)
 // 5-6: Horizontal Retrace Skew = 0 clocks
 // 7: End Horizontal Blanking = 34 (bit 5 of 6)
 movb $0x96,%ah
 outw %ax,%dx
 movb $6,%al        // Vertical Total (bits 0-7 of 10)
 movb $0x0D,%ah     // 525 (+2)
 outw %ax,%dx
 movb $7,%al        // Overflow
 // 0,5: Vertical Total = 525 rows (+2) (bits 8-9 of 10)
 // 1,6: Vertical Display Enable End = Row 479 (bits 8-9 of 10)
 // 2,7: Vertical Retrace Start = Row 480 (bits 8-9 of 10)
 // 3: Start Vertical Blanking = Row 480 (bit 8 of 10)
 // 4: Line Compare = (bit 8 of 10)
 movb $0x3E,%ah
 outw %ax,%dx
 movb $8,%al        // Preset Row Scan
 // 0-4: Preset Row Scan = 0
 // 5-6: Byte Panning Control = 0
 // 7: Reserved = 0
 movb $0,%ah
 outw %ax,%dx
 movb $9,%al        // Maximum Scan Line
 // 0-4: Maximum Scan Lines = 1 (+1)
 // 5: Start Vertical Blanking = Row 480 (bit 9 of 10)
 // 6: Line Compare = () (bit 9 of 10)
 // 7: Double Scan Line Display = 0
 movb $0x41,%ah
 outw %ax,%dx
 movb $0x10,%al     // Vertical Retrace Start (bits 0-7 of 10)
 // Primary centering control - increase number to move up
 movb $0xE0,%ah     // Row 480
 outw %ax,%dx
 movb $0x11,%al     // Vertical Retrace End
 // 0-3: Vertical Retrace End = 0010
 // 4: Clear Vertical Interrupt: state cleared
 // 5: Vertical Interrupt: disabled
 // 6: Refresh Bandwidth Select: 3 memory refresh cycles/scan line
 // 7: CRTC Register Protect: enabled
 movb $0xA2,%ah
 outw %ax,%dx
 movb $0x12,%al     // Vertical Display Enable End (bits 0-7 of 10)
 movb $0xBF,%ah     // Row 447
 outw %ax,%dx
 movb $0x13,%al     // Offset
 movb $0x20,%ah     // 32 * 8 = 256
 outw %ax,%dx
 movb $0x14,%al     // Underline Location
 // 0-4: Underline Location = 0
 // 5: Display Address Dwell 4 (clocks per address inc): 0 (1)
 // 6: Display Address Unit 4 (unit size of address inc): 1 (4)
 // 7: Reserved = 0
 movb $0x40,%ah
 outw %ax,%dx
 movb $0x15,%al     // Start Vertical Blanking (bits 0-7 of 10)
 movb $0xBD,%ah     // Row 445
 outw %ax,%dx
 movb $0x16,%al     // End Vertical Blanking
 movb $0x0D,%ah     // Row x0D
 outw %ax,%dx
 movb $0x17,%al     // CRTC Mode Control
 // 0: Display address bit 13 = Row bit 0: disabled
 // 1: Display address bit 14 = Row bit 1: disabled
 // 2: Row counter increment frequency (by line or by 2): by line
 // 3: Display Address Dwell 2 (clocks per address inc): 0 (1)
 // 4: Reserved = 0
 // 5: Display Address Unit Bank Selection (8k/32k): 32k
 // 6: Display Address Unit 2 (unit size of address inc): 0 (4)
 // 7: Reset/generate retrace signals: Generate
 movb $0xA3,%ah
 outw %ax,%dx
 movl $0x3C4,%edx    // Sequencer
 movb $1,%al
 movb $1,%ah
 outw %ax,%dx
 movb $4,%al
 movb $0x0E,%ah
 outw %ax,%dx
 movl $0x3CE,%edx    // Graphics controller
 movb $5,%al
 movb $0x40,%ah
 outw %ax,%dx
 movb $6,%al
 movb $5,%ah
 outw %ax,%dx
 movl $0x3DA,%edx
 inb %dx,%al
 movl $0x3C0,%edx
 movb $0x30,%al
 outb %al,%dx
 movb $0x41,%al
 outb %al,%dx
 movl $0x3DA,%edx
 inb %dx,%al
 movl $0x3C0,%edx
 movb $0x33,%al
 outb %al,%dx
 movb $0,%al
 outb %al,%dx
 popa
 ret

.text; ALIGN
.data; ALIGN
.bss; ALIGN
