#include "wrapaleg.h"

// TRAC 23/07/1999: Added some code to allow ADSR to be disabled at runtime

#define URATE 55
#define SAMP_LEN 32768
#define ADSR
//#define ADSR_OPTIONAL

#define ATTACK 0    //A of ADSR
#define DECAY 1     //D of ADSR
#define SUSTAIN 2   //S of ADSR
#define RELEASE 3   //R of ADSR
#define DECREASE 4  //GAIN linear decrease mode
#define EXP 5       //GAIN exponential decrease mode
#define INCREASE 6  //GAIN linear increase mode
#define BENT 7      //GAIN bent line increase mode
#define DIRECT 8    //Directly specify ENVX

extern unsigned char SPCRAM[65536];

unsigned long SPC_TIMER0,SPC_TIMER1,SPC_TIMER2;
unsigned long TotalCycles,CycleLatch0,CycleLatch1,CycleLatch2;
unsigned long SPC_TIMT0 ,SPC_TIMT1 ,SPC_TIMT2;
unsigned char SPC_CNT0  ,SPC_CNT1  ,SPC_CNT2  ,
          SPC_CTRL  ,SPC_MASK,
          SPC_DSP_ADDR, SPC_DSP_DATA;

#ifdef ADSR_OPTIONAL
unsigned char ADSR_ENABLED = -1;
#endif
unsigned char SOUND_ENABLED;

unsigned char SPC_DSP_Buffer[256];
unsigned char *SPC_DSP = &SPC_DSP_Buffer[0];

// INTERNAL (static) STUFF

static struct {
    SAMPLE *samp;
    long sampptr;
    int ave;
    unsigned int envx,envclk;
    int envstate;
    int ar,dr,sl,sr,gn;
} SNDvoices[8];

static int SNDkeys = 0;
static unsigned int C[0x20]={0x0,0x3FF00,0x54900,0x64D31,0x83126,0xAEC33,
 0xC9A63,0x10624D,0x154725,0x199999,0x202020,0x2B1DA4,0x333333,0x3F03F0,
 0x563B48,0x667000,0x7E07E0,0xAAAAAA,0xCCCCCC,0x1000000,0x1555555,0x1999999,
 0x2000000,0x2AAAAAA,0x3333333,0x4000000,0x5555555,0x6670000,0x8000000,
 0xAAAAAAA,0x10000000,0x20000000};      //How often to adjust ADSR/GAIN
static unsigned int C_chk=0x20000000;   //Check to see if its been processed

// OLD STUFF

void Sort_Count0(){
 SPC_TIMER0 += (TotalCycles - CycleLatch0) / 256;
 CycleLatch0 += (TotalCycles - CycleLatch0) & ~0xFF;

 SPC_CNT0 = (SPC_CNT0 + (SPC_TIMER0 / SPC_TIMT0)) & 0xF;
 SPC_TIMER0 %= SPC_TIMT0;
}

void Sort_Count1(){
 SPC_TIMER1 += (TotalCycles - CycleLatch1) / 256;
 CycleLatch1 += (TotalCycles - CycleLatch1) & ~0xFF;

 SPC_CNT1 = (SPC_CNT1 + (SPC_TIMER1 / SPC_TIMT1)) & 0xF;
 SPC_TIMER1 %= SPC_TIMT1;
}

void Sort_Count2(){
 SPC_TIMER2 += (TotalCycles - CycleLatch2) / 32;
 CycleLatch2 += (TotalCycles - CycleLatch2) & ~0x1F;

 SPC_CNT2 = (SPC_CNT2 + (SPC_TIMER2 / SPC_TIMT2)) & 0xF;
 SPC_TIMER2 %= SPC_TIMT2;
}

// NEW PRIVATE (static) STUFF

static void update_sample(int snum, int v)
{
    SAMPLE *s;
    int i;
    unsigned char *src;
    short *buf;
    unsigned short end;
    unsigned long brrptr, sampptr = 0;
    int output;
    unsigned char range, filter;
    struct { unsigned short vptr, lptr; } *SRCDIR;

    i = SPC_DSP[0x5D] << 8;
    SRCDIR = (void *) &SPCRAM[i];

    src = &SPCRAM[ SRCDIR[snum].vptr ];
    for (end = 0; !(src[ end ] & 1); end += 9);
    i = ((end + 9) / 9) << 4;
    s = SNDvoices[ v ].samp;
    s->len = s->loop_end = i;
    buf = s->data;

    if (src[ end ] & 2)
    {
        i = ((SRCDIR[ snum ].lptr - SRCDIR[ snum ].vptr) / 9) << 4;
        if ((i > s->len) || (i < 0))
        i = -1;
    }
    else i = -1;
    s->loop_start = i;
    for (brrptr = 0; brrptr <= end; )
    {
        range = src[ brrptr++ ];
        filter = (range >> 2) & 3;
        range >>= 4;
        for(i = 0; i < 8 ; i++, brrptr++)
        {
            output = (src[ brrptr ] >> 4) & 0x0F;
            output = (output ^ 8) - 8;
            output <<= range;
            switch (filter){
             case 1:
                output += (int)(/*(double)*/buf[ sampptr - 1 ] * 15 / 16);
                break;
             case 2:
                output += (int)((/*(double)*/buf[ sampptr - 1 ] * 61 / 32) -
                 (/*(double)*/buf[ sampptr - 2 ] * 15 / 16));
                break;
             case 3:
                output += (int)((/*(double)*/buf[ sampptr - 1 ] * 115 / 64) -
                 (/*(double)*/buf[ sampptr - 2 ] * 13 / 16));
                break;
            }
            if (output > 32767)
             output = 32767;
            else if (output < -32768)
             output = -32768;
            buf[ sampptr++ ] = output;
            output = src[brrptr] & 0x0F;
            output = (output ^ 8) - 8;
            output <<= range;
            switch (filter){
             case 1:
                output += (int)(/*(double)*/buf[ sampptr - 1 ] * 15 / 16);
                break;
             case 2:
                output += (int)((/*(double)*/buf[ sampptr - 1 ] * 61 / 32) -
                 (/*(double)*/buf[ sampptr - 2 ] * 15 / 16));
                break;
             case 3:
                output += (int)((/*(double)*/buf[ sampptr - 1 ] * 115 / 64) -
                 (/*(double)*/buf[ sampptr - 2 ] * 13 / 16));
                break;
            }
            if (output > 32767)
             output = 32767;
            else if (output < -32768)
             output = -32768;
            buf[ sampptr++ ] = output;
        }
    }
    // Signed -> Unsigned
    for (i = s->len - 1; i >= 0; i--)
     buf[i] ^= 0x8000;
}

static void SNDNoteOn(int v)
{
    int i,cursamp,lvol,rvol=0,pitch,adsr1,adsr2,gain;
    v&=SPC_MASK;
    for(i=0;i<8;i++)
    if(v&(1<<i))
    {
        cursamp=SPC_DSP[4+(i<<4)];
        pitch=(int)(*(unsigned short *)&SPC_DSP[(i<<4)+2])<<3;
#ifdef ADSR_OPTIONAL
        if(ADSR_ENABLED){
#endif
#if defined(ADSR) || defined(ADSR_OPTIONAL)
            adsr1=SPC_DSP[(i<<4)+5];
            if(adsr1&0x80)
            {
                //ADSR mode
                adsr2=SPC_DSP[(i<<4)+6];
                SNDvoices[i].envx=0;
                SNDvoices[i].envclk=0;
                SNDvoices[i].envstate=ATTACK;
                SNDvoices[i].ar=adsr1&0xF;
                SNDvoices[i].dr=adsr1>>4&7;
                SNDvoices[i].sr=adsr2&0x1f;
                SNDvoices[i].sl=adsr2>>5;
            }
            else
            {
                //GAIN mode
                gain=SPC_DSP[(i<<4)+7];
                if(gain&0x80)
                {
                    SNDvoices[i].envclk=0;
                    SNDvoices[i].envstate=gain>>5;
                    SNDvoices[i].gn=gain&0x1F;
                }
                else
                {
                    SNDvoices[i].envx=(gain&0x7F)<<24;
                    SNDvoices[i].envstate=DIRECT;
                }
            }
            if(SOUND_ENABLED==2)
            {
                lvol=((SNDvoices[i].envx>>24)*
                 (int)(SPC_DSP[i<<4]&0x7F)*
                 (int)(SPC_DSP[0x0C]&0x7F))>>13;
                rvol=((SNDvoices[i].envx>>24)*
                 (int)(SPC_DSP[(i<<4)+1]&0x7F)*
                 (int)(SPC_DSP[0x1C]&0x7F))>>13;
            }
            else
            {
                lvol=((SNDvoices[i].envx>>24)*
                 ((int)(SPC_DSP[i<<4]&0x7F)+
                 (int)(SPC_DSP[(i<<4)+1]&0x7F))*
                 ((int)(SPC_DSP[0x0C]&0x7F)+
                 (int)(SPC_DSP[0x1C]&0x7F)))>>15;
            }
#endif
#ifdef ADSR_OPTIONAL
        } else {
#endif
#if !defined(ADSR) || defined(ADSR_OPTIONAL)
            if(SOUND_ENABLED==2)
            {
                lvol=((int)(SPC_DSP[i<<4]&0x7F)*
                 (int)(SPC_DSP[0x0C]&0x7F))>>6;
                rvol=((int)(SPC_DSP[(i<<4)+1]&0x7F)*
                 (int)(SPC_DSP[0x1C]&0x7F))>>6;
            }
            else
            {
                lvol=(((int)(SPC_DSP[i<<4]&0x7F)+
                 (int)(SPC_DSP[(i<<4)+1]&0x7F))*
                 ((int)(SPC_DSP[0x0C]&0x7F)+
                 (int)(SPC_DSP[0x1C]&0x7F)))>>8;
            }
#endif
#ifdef ADSR_OPTIONAL
        }
#endif
        if(SNDkeys&(1<<i))
        {
            if(SOUND_ENABLED){
             voice_stop(i);
             if(SOUND_ENABLED==2)
              voice_stop(i+8);
            }
        }
        SNDkeys|=(1<<i);
    //  update_sample(cursamp,i);
        if(SOUND_ENABLED)
        {
         update_sample(cursamp,i);
         reallocate_voice(i,SNDvoices[i].samp);
         voice_set_volume(i,lvol);
         voice_set_frequency(i,pitch);
         if(SOUND_ENABLED==2)
         {
          reallocate_voice(i+8,SNDvoices[i].samp);
          voice_set_volume(i+8,rvol);
          voice_set_frequency(i+8,pitch);
          voice_set_pan(i,0);
          voice_set_pan(i+8,255);
         }
        }
        if(SNDvoices[i].samp->loop_start==-1)
        {
            if(SOUND_ENABLED){
             voice_set_playmode(i,PLAYMODE_PLAY);
             if(SOUND_ENABLED==2)
              voice_set_playmode(i+8,PLAYMODE_PLAY);
            }
        }
        else
        {
            if(SOUND_ENABLED){
             voice_set_playmode(i,PLAYMODE_LOOP);
             if(SOUND_ENABLED==2)
              voice_set_playmode(i+8,PLAYMODE_LOOP);
            }
        }
        if(SOUND_ENABLED){
         voice_start(i);
         if(SOUND_ENABLED==2)
          voice_start(i+8);
        }
    }
}

static void SNDNoteOff(int v)
{
    int i;
    for(i=0;i<8;i++)
        if(v&(1<<i))
        {
#ifdef ADSR_OPTIONAL
            if(ADSR_ENABLED){
#endif
#if defined(ADSR) || defined(ADSR_OPTIONAL)
                SNDvoices[i].envstate=RELEASE;
                SNDvoices[i].envclk=0;
#endif
#ifdef ADSR_OPTIONAL
            } else {
#endif
#if !defined(ADSR) || defined(ADSR_OPTIONAL)
                if(SOUND_ENABLED){
                 voice_stop(i);
                 if(SOUND_ENABLED==2)
                  voice_stop(i+8);
                }
                SNDkeys&=~(1<<i);
#endif
#ifdef ADSR_OPTIONAL
            }
#endif
        }
}

#if defined(ADSR) || defined(ADSR_OPTIONAL)
static int SNDDoEnv(int voice)
{
    unsigned int envx=SNDvoices[voice].envx,clk=SNDvoices[voice].envclk;
    switch (SNDvoices[voice].envstate)
    {
        case ATTACK:
            clk+=C[(SNDvoices[voice].ar<<1)+1];
            while(clk>=0x1000000)
            {
                clk-=0x1000000;
                envx+=0x2000000;        //add 1/64
                if(envx>=0x7F000000)
                {
                    envx=0x7F000000;
                    clk=0;
                    if(SNDvoices[voice].sl!=7)
                         SNDvoices[voice].envstate=DECAY;
                    else
                         SNDvoices[voice].envstate=SUSTAIN;
                }
            }
            break;
        case DECAY:
            clk+=C[(SNDvoices[voice].dr<<1)+0x10];
            while(clk>=0x1000000)       //May be slow, but what better way?
            {
                clk-=0x1000000;
                envx=(envx>>8)*255;     //mult by 1-1/256
                if(envx<=0x10000000*(SNDvoices[voice].sl+1))
                {
                    clk=0;
                    envx=0x10000000*(SNDvoices[voice].sl+1);
                    SNDvoices[voice].envstate=SUSTAIN;
                }
            }
            break;
        case SUSTAIN:
            clk+=C[SNDvoices[voice].sr];
            while(clk>=0x1000000)       //May be slow, but what better way?
            {
                clk-=0x1000000;
                envx=(envx>>8)*255;     //mult by 1-1/256
            }
            break;
        case RELEASE:
            //says add 1/256??  That won't release, must be subtract.
            //But how often?  Oh well, who cares, I'll just
            //pick a number. :)
            clk+=C[0x1A];
            while(clk>=0x1000000)
            {
                clk-=0x1000000;
                envx-=0x800000;         //sub 1/256
                if((envx==0)||(envx>0x10000000*(SNDvoices[voice].sl+1)))
                {
                    if(SOUND_ENABLED){
                     voice_stop(voice);
                     if(SOUND_ENABLED==2)
                      voice_stop(voice+8);
                    }
                    SNDkeys&=~(1<<voice);
                    envx=0;
                    clk=0;
                }
            }
            break;
        case INCREASE:
            clk+=C[SNDvoices[voice].gn];
            while(clk>=0x1000000)
            {
                clk-=0x1000000;
                envx+=0x2000000;        //add 1/64
                if(envx>0x7F000000)
                {
                    envx=0x7F000000;
                    clk=0;
                }
            }
            break;
        case DECREASE:
            clk+=C[SNDvoices[voice].gn];
            while(clk>=0x1000000)
            {
                clk-=0x1000000;
                envx-=0x2000000;        //sub 1/64
                if(envx>0x7F000000)     //underflow
                {
                    clk=0;
                    envx=0;
                }
            }
            break;
        case EXP:
            clk+=C[SNDvoices[voice].gn];
            while(clk>=0x1000000)       //May be slow, but what better way?
            {
                clk-=0x1000000;
                envx=(envx>>8)*255;     //mult by 1-1/256
            }
            break;
        case BENT:
            clk+=C[SNDvoices[voice].gn];
            while(clk>=0x1000000)
            {
                clk-=0x1000000;
                if(envx<0x60000000)
                    envx+=0x2000000;    //add 1/64
                else
                    envx+=0x800000;     //add 1/256
                if(envx>0x7F000000)
                {
                    envx=0x7F000000;
                    clk=0;
                }
            }
    }
    SNDvoices[voice].envclk=clk;
    return SNDvoices[voice].envx=envx;
}
#endif
// NEW SHARED STUFF

int Install_Sound(int stereo)
{
    SAMPLE *tmp;
    int i;
    unsigned int t;

    SNDkeys = 0;
    if(stereo) reserve_voices(16,0);
    else reserve_voices(8,0);

    if(install_sound(DIGI_AUTODETECT,MIDI_NONE,NULL))
        return 0;

    SPC_MASK=0xFF;
    if(C[0x1F]==C_chk)      // Clock table needs initialization
    {
        C[0]*=(1000/URATE);
        for(i=1;i<0x20;i++)             //ADSR clock table
        {
            t=C[i]*(1000/URATE);
            if((t+0x1000000)<C[i-1])    //overflow
                C[i]=0xFEFFFFFF;
            else C[i]=t;
        }
    }
    tmp=create_sample(16,FALSE,1000,1);
    for (i = 0; i < (stereo ? 16 : 8); i++)
        if(allocate_voice(tmp)==-1)
            return 0;
    for(i=0;i<8;i++)
    {
        reallocate_voice(i,SNDvoices[i].samp=create_sample(16,FALSE,
         1000,SAMP_LEN));
        if(stereo)
        {
            reallocate_voice(i+8,SNDvoices[i].samp);
            voice_set_pan(i,0);
            voice_set_pan(i+8,255);
        }
    }
    destroy_sample(tmp);
    return -1;
}

void Remove_Sound()
{
    int i;
    SNDkeys = 0;
    remove_sound();
    for(i=0;i<8;i++)
        destroy_sample(SNDvoices[i].samp);
}

void SPC_READ_DSP()
{
    if((SPC_DSP_ADDR&0xF)==8)
#ifdef NOENVX
        SPC_DSP[SPC_DSP_ADDR]=0;
#else
        SPC_DSP[SPC_DSP_ADDR]=SNDvoices[SPC_DSP_ADDR>>4].envx>>24;
#endif
}

void SPC_WRITE_DSP()
{
    int i;
    int addr_lo=SPC_DSP_ADDR&0xF, addr_hi=SPC_DSP_ADDR>>4;
    switch(addr_lo)
    {
        //break just means nothing needs to be done right now
        //Commented cases are unsupported registers
        // These go for individual voices
        case 0:    //Volume left
            break;
        case 1:    //Volume right
            break;
        case 3: //Pitch hi
            SPC_DSP_DATA&=0x3F;
            break;
        case 2: //Pitch lo
            break;
        case 4:    //Source number
            break;
#if defined(ADSR) || defined(ADSR_OPTIONAL)
        case 5: //ADSR1
#ifdef ADSR_OPTIONAL
            if(!ADSR_ENABLED) break;
#endif
            if((SNDkeys&(1<<addr_hi))&&
             ((SPC_DSP_DATA&0x80)!=(SPC_DSP[SPC_DSP_ADDR]&0x80)))
            {
                //First of all, in case anything was already
                //going on, finish it up (FF2/4)
                SNDDoEnv(addr_hi);
                if(SPC_DSP_DATA&0x80)
                {
                    //switch to ADSR--not sure what to do
                    i=SPC_DSP[(addr_hi<<4)+6];
                    SNDvoices[addr_hi].envclk=0;
                    SNDvoices[addr_hi].envstate=ATTACK;
                    SNDvoices[addr_hi].ar=SPC_DSP_DATA&0xF;
                    SNDvoices[addr_hi].dr=SPC_DSP_DATA>>4&7;
                    SNDvoices[addr_hi].sr=i&0x1f;
                    SNDvoices[addr_hi].sl=i>>5;
                }
                else
                {
                    //switch to a GAIN mode
                    i=SPC_DSP[(addr_hi<<4)+7];
                    if(i&0x80)
                    {
                        SNDvoices[addr_hi].envclk=0;
                        SNDvoices[addr_hi].envstate=i>>5;
                        SNDvoices[addr_hi].gn=i&0x1F;
                    }
                    else
                    {
                        SNDvoices[addr_hi].envx=(i&0x7F)<<24;
                        SNDvoices[addr_hi].envstate=DIRECT;
                    }
                }
            }
            break;
        case 6: //ADSR2
#ifdef ADSR_OPTIONAL
            if(!ADSR_ENABLED) break;
#endif
            //not sure if it makes a difference, but might as well
            //update the stuff anyway...
            SNDvoices[addr_hi].sr=SPC_DSP_DATA&0x1f;
            SNDvoices[addr_hi].sl=SPC_DSP_DATA>>5;
            break;
        case 7: //GAIN
#ifdef ADSR_OPTIONAL
            if(!ADSR_ENABLED) break;
#endif
            if((SNDkeys&(1<<addr_hi))&&
             (SPC_DSP_DATA!=SPC_DSP[SPC_DSP_ADDR])&&
             !(SPC_DSP[(addr_hi<<4)+5]&0x80))
            {
                if(SPC_DSP_DATA&0x80)
                {
                    SNDvoices[addr_hi].envclk=0;
                    SNDvoices[addr_hi].envstate=SPC_DSP_DATA>>5;
                    SNDvoices[addr_hi].gn=SPC_DSP_DATA&0x1F;
                }
                else
                {
                    SNDvoices[addr_hi].envx=(SPC_DSP_DATA&0x7F)<<24;
                    SNDvoices[addr_hi].envstate=DIRECT;
                }
            }
            break;
#endif
        // These are general registers
        case 0xc:
            switch(addr_hi)
            {
                case 0: //Left main volume
                    break;
                case 1: //Right main volume
                    break;
//              case 2: //Left echo volume
//                  break;
//              case 3: //Right echo volume
//                  break;
                case 4: //Key on
                    if(SOUND_ENABLED)
                     SNDNoteOn(SPC_DSP_DATA);
                    SPC_DSP_DATA=SNDkeys;
                    break;
                case 5: //Key off
                    if(SOUND_ENABLED)
                     SNDNoteOff(SPC_DSP_DATA);
                    SPC_DSP_DATA=0;
                    break;
//              case 6: //Clocks
//                  break;
//              case 7: //Can't write to this!
//                  break;
            }
            break;
/*      case 0xd:
            switch(addr_hi)
            {
                case 0: //Echo feedback
                    break;
                case 1: //Not a valid register
                    break;
                case 2: //Modulation
                    break;
                case 3:    //Noise toggle
                    break;
                case 4: //Echo toggle
                    break;
                case 5:    //Source directory address
                    break;
                case 6:    //Echo start address
                    break;
                case 7: //Echo delay
                    break;
            }
        case 0xf:       //FIR Echo filter
            break;
*/
    }
    SPC_DSP[SPC_DSP_ADDR]=SPC_DSP_DATA;
}

void SNDUpdate(void)
{
    int i,t,envx;
    if(!SOUND_ENABLED)
        return;
    for(i=0;i<8;i++)
        if(SNDkeys&(1<<i))
        {
#ifdef ADSR_OPTIONAL
            if(ADSR_ENABLED){
#endif
#if defined(ADSR) || defined(ADSR_OPTIONAL)
                envx=SNDDoEnv(i);
#endif
#ifdef ADSR_OPTIONAL
            } else {
#endif
#if !defined(ADSR) || defined(ADSR_OPTIONAL)
                envx=0x7F000000;
#endif
#ifdef ADSR_OPTIONAL
            }
#endif
            if(SOUND_ENABLED==2)
            {
                t=((envx>>24)*
                 (int)(SPC_DSP[i<<4]&0x7F)*
                 (int)(SPC_DSP[0x0C]&0x7F))>>13;
                voice_set_volume(i,t);
                t=((envx>>24)*
                 (int)(SPC_DSP[(i<<4)+1]&0x7F)*
                 (int)(SPC_DSP[0x1C]&0x7F))>>13;
                voice_set_volume(i+8,t);
                t=(int)(*(unsigned short *)&SPC_DSP[(i<<4)+2])<<3;
                voice_set_frequency(i,t);
                voice_set_frequency(i+8,t);
            }
            else
            {
                t=((envx>>24)*
                 ((int)(SPC_DSP[i<<4]&0x7F)+
                 (int)(SPC_DSP[(i<<4)+1]&0x7F))*
                 ((int)(SPC_DSP[0x0C]&0x7F)+
                 (int)(SPC_DSP[0x1C]&0x7F)))>>15;
                voice_set_volume(i,t);
                t=(int)(*(unsigned short *)&SPC_DSP[(i<<4)+2])<<3;
                voice_set_frequency(i,t);
            }
        }
}

void sound_pause(void){
 int i;

 for (i = 0; i < 8; i++){
  if (SNDkeys & (1 << i)){
   if(SOUND_ENABLED){
    voice_stop(i);
    if(SOUND_ENABLED==2)
     voice_stop(i+8);
   }
  }
 }
}

void sound_resume(void){
 int i;

 for (i = 0; i < 8; i++){
  if (SNDkeys & (1 << i)){
   if(SOUND_ENABLED){
    voice_start(i);
    if(SOUND_ENABLED==2)
     voice_start(i+8);
   }
  }
 }
}
