/*

 Tile Functions - In Assembler cos everything else is!
  Line-from-tile plotters with tile caching

 VRAM tile address in %esi + LineAddress(Y) + VRAMAddress...
 Cache tile address varies by depth...

 Tile recache
  esi = cache address   (input)
      = tile address    (internal)
  edi = cache address   (internal)
  eax,ebx,ecx,edx,ebp destroyed
 In all the plotters...
  esi = cache address   (input)
  edi = cache address   (internal)
      = screen address  (input)
 In check-first plotters...
  dl  = palette bitmask (internal)
        (except 8-bpl)
  eax = pixels          (internal)
  eax,dl destroyed
 In blank tile plotters
  Nothing
 In no-check plotters...
  ah  = palette bitmask (internal)
        (except 8-bpl)
  al  = pixels          (internal)
  dl  = pixels          (internal)
  ax,dl destroyed
*/

#include "misc.i"

EXPORT_C tilesL_start

.data
ALIGN
.globl Tile_Line_8_2,Tile_Line_16_2
.globl Tile_Line_16_2_Even
/*
    .long   CacheMiss2Bpl,CacheMiss2Bpl,0,0
.set Tile_Line_8_2,.-16
    .long   PLOT8_2BplTile
    .long   PLOT8_2BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_8
    .long   Plot_Masked_X_8

.set Tile_Line_16_2_Even,.-16
    .long   PLOT16_2BplTile_Even
    .long   PLOT16_2BplTile_Even_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_Even_16
    .long   Plot_Masked_Even_X_16

.set Tile_Line_16_2,.-16
    .long   PLOT16_2BplTile
    .long   PLOT16_2BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_16
    .long   Plot_Masked_X_16
*/
.globl Tile_Line_8_4,Sprite_Tile_Plot_8_4
.globl Tile_Line_16_4,Sprite_Tile_Plot_16_4
.globl Sprite_Tile_Plot_32_4,Sprite_Tile_Plot_64_4
.globl Tile_Line_16_4_Even

    .long   CacheMiss2Bpl,CacheMiss2Bpl,0,0
.set Tile_Line_8_2,.-16
.set Tile_Line_8_4,.-16
.set Sprite_Tile_Plot_8_4,.-16
    .long   PLOT8_4BplTile
    .long   PLOT8_4BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_8
    .long   Plot_Masked_X_8

.set Tile_Line_16_2_Even,.-16
.set Tile_Line_16_4_Even,.-16
    .long   PLOT16_4BplTile_Even
    .long   PLOT16_4BplTile_Even_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_Even_16
    .long   Plot_Masked_Even_X_16

.set Tile_Line_16_2,.-16
.set Tile_Line_16_4,.-16
.set Sprite_Tile_Plot_16_4,.-16
    .long   PLOT16_4BplTile
    .long   PLOT16_4BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_Masked_16
    .long   Plot_Masked_X_16

Sprite_Tile_Plot_32_4:
    .long   PLOT32_4BplSprite
    .long   PLOT32_4BplSprite_X

Sprite_Tile_Plot_64_4:
    .long   PLOT64_4BplSprite
    .long   PLOT64_4BplSprite_X

    .long   CacheMiss8Bpl,CacheMiss8Bpl,0,0
.globl Tile_Line_8_8,Tile_Line_16_8
.set Tile_Line_8_8,.-16
    .long   PLOT8_8BplTile
    .long   PLOT8_8BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_8
    .long   Plot_X_8

.set Tile_Line_16_8,.-16
    .long   PLOT16_8BplTile
    .long   PLOT16_8BplTile_X
    .long   No_Plot
    .long   No_Plot
    .long   Plot_16
    .long   Plot_X_16

ALIGN
BPL0_2:
    .long   0x00000000,0xFD000000,0x00FD0000,0xFDFD0000
    .long   0x0000FD00,0xFD00FD00,0x00FDFD00,0xFDFDFD00
    .long   0x000000FD,0xFD0000FD,0x00FD00FD,0xFDFD00FD
    .long   0x0000FDFD,0xFD00FDFD,0x00FDFDFD,0xFDFDFDFD

BPL1_2:
    .long   0x00000000,0xFE000000,0x00FE0000,0xFEFE0000
    .long   0x0000FE00,0xFE00FE00,0x00FEFE00,0xFEFEFE00
    .long   0x000000FE,0xFE0000FE,0x00FE00FE,0xFEFE00FE
    .long   0x0000FEFE,0xFE00FEFE,0x00FEFEFE,0xFEFEFEFE

BPL0_4:
    .long   0x00000000,0xF1000000,0x00F10000,0xF1F10000
    .long   0x0000F100,0xF100F100,0x00F1F100,0xF1F1F100
    .long   0x000000F1,0xF10000F1,0x00F100F1,0xF1F100F1
    .long   0x0000F1F1,0xF100F1F1,0x00F1F1F1,0xF1F1F1F1

BPL1_4:
    .long   0x00000000,0xF2000000,0x00F20000,0xF2F20000
    .long   0x0000F200,0xF200F200,0x00F2F200,0xF2F2F200
    .long   0x000000F2,0xF20000F2,0x00F200F2,0xF2F200F2
    .long   0x0000F2F2,0xF200F2F2,0x00F2F2F2,0xF2F2F2F2

BPL2_4:
    .long   0x00000000,0xF4000000,0x00F40000,0xF4F40000
    .long   0x0000F400,0xF400F400,0x00F4F400,0xF4F4F400
    .long   0x000000F4,0xF40000F4,0x00F400F4,0xF4F400F4
    .long   0x0000F4F4,0xF400F4F4,0x00F4F4F4,0xF4F4F4F4

BPL3_4:
    .long   0x00000000,0xF8000000,0x00F80000,0xF8F80000
    .long   0x0000F800,0xF800F800,0x00F8F800,0xF8F8F800
    .long   0x000000F8,0xF80000F8,0x00F800F8,0xF8F800F8
    .long   0x0000F8F8,0xF800F8F8,0x00F8F8F8,0xF8F8F8F8

BPL0_8:
    .long   0x00000000,0x01000000,0x00010000,0x01010000
    .long   0x00000100,0x01000100,0x00010100,0x01010100
    .long   0x00000001,0x01000001,0x00010001,0x01010001
    .long   0x00000101,0x01000101,0x00010101,0x01010101

BPL1_8:
    .long   0x00000000,0x02000000,0x00020000,0x02020000
    .long   0x00000200,0x02000200,0x00020200,0x02020200
    .long   0x00000002,0x02000002,0x00020002,0x02020002
    .long   0x00000202,0x02000202,0x00020202,0x02020202

BPL2_8:
    .long   0x00000000,0x04000000,0x00040000,0x04040000
    .long   0x00000400,0x04000400,0x00040400,0x04040400
    .long   0x00000004,0x04000004,0x00040004,0x04040004
    .long   0x00000404,0x04000404,0x00040404,0x04040404

BPL3_8:
    .long   0x00000000,0x08000000,0x00080000,0x08080000
    .long   0x00000800,0x08000800,0x00080800,0x08080800
    .long   0x00000008,0x08000008,0x00080008,0x08080008
    .long   0x00000808,0x08000808,0x00080808,0x08080808

BPL4_8:
    .long   0x00000000,0x10000000,0x00100000,0x10100000
    .long   0x00001000,0x10001000,0x00101000,0x10101000
    .long   0x00000010,0x10000010,0x00100010,0x10100010
    .long   0x00001010,0x10001010,0x00101010,0x10101010

BPL5_8:
    .long   0x00000000,0x20000000,0x00200000,0x20200000
    .long   0x00002000,0x20002000,0x00202000,0x20202000
    .long   0x00000020,0x20000020,0x00200020,0x20200020
    .long   0x00002020,0x20002020,0x00202020,0x20202020

BPL6_8:
    .long   0x00000000,0x40000000,0x00400000,0x40400000
    .long   0x00004000,0x40004000,0x00404000,0x40404000
    .long   0x00000040,0x40000040,0x00400040,0x40400040
    .long   0x00004040,0x40004040,0x00404040,0x40404040

BPL7_8:
    .long   0x00000000,0x80000000,0x00800000,0x80800000
    .long   0x00008000,0x80008000,0x00808000,0x80808000
    .long   0x00000080,0x80000080,0x00800080,0x80800080
    .long   0x00008080,0x80008080,0x00808080,0x80808080

.bss
ALIGN
.globl ColourBase,PixMask
ColourBase: skipb
PixMask:    skipb
.balign 4
TileOR: skipl
TileAND:skipl

.text
ALIGN
.globl No_Plot
No_Plot:
    ret

.macro Plot_Masked_2 Offset,Width,Xflip,Mask=%ah
.if \Xflip-1
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,\Offset+1(%edi)
    movb %al,\Offset(%edi)
.else
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,(\Width-2-\Offset)(%edi)
    movb %al,(\Width-1-\Offset)(%edi)
.endif
.endm

.macro Plot_Masked_Even_2 Offset,Width,Xflip,Mask=%ah
.if \Xflip-1
    movb \Offset(%esi),%al
    movb \Offset+2(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,\Offset+2(%edi)
    movb %al,\Offset(%edi)
.else
    movb \Offset(%esi),%al
    movb \Offset+2(%esi),%dl
    andb \Mask,%dl
    andb \Mask,%al
    movb %dl,(\Width-3-\Offset)(%edi)
    movb %al,(\Width-1-\Offset)(%edi)
.endif
.endm

.macro Plot_2 Offset,Width,Xflip
.if \Xflip-1
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    movb %al,\Offset(%edi)
    movb %dl,\Offset+1(%edi)
.else
    movb \Offset(%esi),%al
    movb \Offset+1(%esi),%dl
    movb %al,(\Width-1-\Offset)(%edi)
    movb %dl,(\Width-2-\Offset)(%edi)
.endif
.endm

ALIGN
.globl Plot_Masked_8
Plot_Masked_8:
    movb ColourBase,%ah
    Plot_Masked_2 Offset=0 Width=8 Xflip=0
    Plot_Masked_2 Offset=2 Width=8 Xflip=0
    Plot_Masked_2 Offset=4 Width=8 Xflip=0
    Plot_Masked_2 Offset=6 Width=8 Xflip=0
    ret

ALIGN
.globl Plot_Masked_X_8
Plot_Masked_X_8:
    movb ColourBase,%ah
    Plot_Masked_2 Offset=0 Width=8 Xflip=1
    Plot_Masked_2 Offset=2 Width=8 Xflip=1
    Plot_Masked_2 Offset=4 Width=8 Xflip=1
    Plot_Masked_2 Offset=6 Width=8 Xflip=1
    ret

ALIGN
.globl Plot_Masked_Even_16
Plot_Masked_Even_16:
    movb ColourBase,%ah
    Plot_Masked_Even_2 Offset=0  Width=16 Xflip=0
    Plot_Masked_Even_2 Offset=4  Width=16 Xflip=0
    Plot_Masked_Even_2 Offset=8  Width=16 Xflip=0
    Plot_Masked_Even_2 Offset=12 Width=16 Xflip=0
    ret

ALIGN
.globl Plot_Masked_Even_X_16
Plot_Masked_Even_X_16:
    movb ColourBase,%ah
    Plot_Masked_Even_2 Offset=0  Width=16 Xflip=1
    Plot_Masked_Even_2 Offset=4  Width=16 Xflip=1
    Plot_Masked_Even_2 Offset=8  Width=16 Xflip=1
    Plot_Masked_Even_2 Offset=12 Width=16 Xflip=1
    ret

ALIGN
.globl Plot_Masked_16
Plot_Masked_16:
    movb ColourBase,%ah
    Plot_Masked_2 Offset=0  Width=16 Xflip=0
    Plot_Masked_2 Offset=2  Width=16 Xflip=0
    Plot_Masked_2 Offset=4  Width=16 Xflip=0
    Plot_Masked_2 Offset=6  Width=16 Xflip=0
    Plot_Masked_2 Offset=8  Width=16 Xflip=0
    Plot_Masked_2 Offset=10 Width=16 Xflip=0
    Plot_Masked_2 Offset=12 Width=16 Xflip=0
    Plot_Masked_2 Offset=14 Width=16 Xflip=0
    ret

ALIGN
.globl Plot_Masked_X_16
Plot_Masked_X_16:
    movb ColourBase,%ah
    Plot_Masked_2 Offset=0  Width=16 Xflip=1
    Plot_Masked_2 Offset=2  Width=16 Xflip=1
    Plot_Masked_2 Offset=4  Width=16 Xflip=1
    Plot_Masked_2 Offset=6  Width=16 Xflip=1
    Plot_Masked_2 Offset=8  Width=16 Xflip=1
    Plot_Masked_2 Offset=10 Width=16 Xflip=1
    Plot_Masked_2 Offset=12 Width=16 Xflip=1
    Plot_Masked_2 Offset=14 Width=16 Xflip=1
    ret

ALIGN
.globl Plot_8
Plot_8:
    Plot_2 Offset=0 Width=8 Xflip=0
    Plot_2 Offset=2 Width=8 Xflip=0
    Plot_2 Offset=4 Width=8 Xflip=0
    Plot_2 Offset=6 Width=8 Xflip=0
    ret

ALIGN
.globl Plot_X_8
Plot_X_8:
    Plot_2 Offset=0 Width=8 Xflip=1
    Plot_2 Offset=2 Width=8 Xflip=1
    Plot_2 Offset=4 Width=8 Xflip=1
    Plot_2 Offset=6 Width=8 Xflip=1
    ret

ALIGN
.globl Plot_16
Plot_16:
    Plot_2 Offset=0  Width=16 Xflip=0
    Plot_2 Offset=2  Width=16 Xflip=0
    Plot_2 Offset=4  Width=16 Xflip=0
    Plot_2 Offset=6  Width=16 Xflip=0
    Plot_2 Offset=8  Width=16 Xflip=0
    Plot_2 Offset=10 Width=16 Xflip=0
    Plot_2 Offset=12 Width=16 Xflip=0
    Plot_2 Offset=14 Width=16 Xflip=0
    ret

ALIGN
.globl Plot_X_16
Plot_X_16:
    Plot_2 Offset=0  Width=16 Xflip=1
    Plot_2 Offset=2  Width=16 Xflip=1
    Plot_2 Offset=4  Width=16 Xflip=1
    Plot_2 Offset=6  Width=16 Xflip=1
    Plot_2 Offset=8  Width=16 Xflip=1
    Plot_2 Offset=10 Width=16 Xflip=1
    Plot_2 Offset=12 Width=16 Xflip=1
    Plot_2 Offset=14 Width=16 Xflip=1
    ret

ALIGN
.globl CacheMiss8Bpl
CacheMiss8Bpl:
    pushl %eax
    pushl %ebx
    pushl %edi
    movl _TileCache8,%edi
    pushl %esi
    subl %edi,%esi
    andl $0xFFC0,%esi
    addl %esi,%edi
    addl $_VRAM,%esi
    movl $0,TileOR
    movl $-1,TileAND
    movl $8,%edx
0:  xorl %ebx,%ebx
    //  Bp0=*(LineAddress+0)
    movb (%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    movl BPL0_8(,%ebx,4),%eax
    movl BPL0_8(,%ecx,4),%ebp

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL1_8(,%ebx,4),%eax
    orl  BPL1_8(,%ecx,4),%ebp

    //  Bp2=*(LineAddress+16)
    movb 16(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL2_8(,%ebx,4),%eax
    orl  BPL2_8(,%ecx,4),%ebp

    //  Bp3=*(LineAddress+17)
    movb 17(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL3_8(,%ebx,4),%eax
    orl  BPL3_8(,%ecx,4),%ebp

    //  Bp4=*(LineAddress+32)
    movb 32(%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL4_8(,%ebx,4),%eax
    orl  BPL4_8(,%ecx,4),%ebp

    //  Bp5=*(LineAddress+33)
    movb 33(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL5_8(,%ebx,4),%eax
    orl  BPL5_8(,%ecx,4),%ebp

    //  Bp6=*(LineAddress+48)
    movb 48(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL6_8(,%ebx,4),%eax
    orl  BPL6_8(,%ecx,4),%ebp

    //  Bp7=*(LineAddress+49)
    movb 49(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL7_8(,%ebx,4),%eax
    orl  BPL7_8(,%ecx,4),%ebp

    movl %eax,(%edi)
    orl %eax,TileOR
    andl %eax,TileAND
    movl %ebp,4(%edi)
    orl %ebp,TileOR
    andl %ebp,TileAND
    addl $2,%esi
    addl $8,%edi
    decl %edx
    jnz 0b
    popl %esi
    xorl %eax,%eax
    popl %edi
    cmpl %eax,TileOR
    popl %ebx
    jnz 0f
    popl %eax
    movb $0x18,_CacheDirty8(%eax)
    jmpl Tile_Line_8_8+0x18(%ebx)
0:  movl TileAND,%eax
    movl %eax,%edx
    shrl $16,%eax
    andl %edx,%eax
    andb %ah,%al
    popl %eax
    jz  0f
//  movb $0x20,_CacheDirty8(%eax)
//  jmpl Tile_Line_8_8+0x20(%ebx)
0:  movb $0x10,_CacheDirty8(%eax)
    jmpl Tile_Line_8_8+0x10(%ebx)

ALIGN
.globl CacheMiss4Bpl
CacheMiss4Bpl:
    pushl %eax
    pushl %esi
    pushl %edi
    andl $0xFFE0*2,%esi
    movl %esi,%edi
    shrl $1,%esi
    addl _TileCache4,%edi
    addl $_VRAM,%esi
    movb $0x10,_CacheDirty4(%eax)
    movl $0,TileOR
    movl $-1,TileAND
    movl $8,%edx
0:  xorl %ebx,%ebx
    //  Bp0=*(LineAddress+0)
    movb (%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    movl BPL0_4(,%ebx,4),%eax
    movl BPL0_4(,%ecx,4),%ebp

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL1_4(,%ebx,4),%eax
    orl  BPL1_4(,%ecx,4),%ebp

    //  Bp2=*(LineAddress+16)
    movb 16(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL2_4(,%ebx,4),%eax
    orl  BPL2_4(,%ecx,4),%ebp

    //  Bp3=*(LineAddress+17)
    movb 17(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL3_4(,%ebx,4),%eax
    orl  BPL3_4(,%ecx,4),%ebp

    movl %eax,(%edi)
    orl %eax,TileOR
    andl %eax,TileAND
    movl %ebp,4(%edi)
    orl %ebp,TileOR
    andl %ebp,TileAND
    addl $2,%esi
    addl $8,%edi
    decl %edx
    jnz 0b
    popl %edi
    xorl %eax,%eax
    popl %esi
    cmpl %eax,TileOR
    jnz 0f
    popl %eax
    movb $0x18,_CacheDirty4(%eax)
    ret
0:  movl TileAND,%eax
    movl %eax,%edx
    shrl $16,%eax
    andl %edx,%eax
    andb %ah,%al
    popl %eax
    jz  0f
//  movb $0x20,_CacheDirty4(%eax)
0:  ret

ALIGN
.globl CacheMiss2Bpl
CacheMiss2Bpl:
    pushl %eax
    pushl %ebx
    pushl %edi
    pushl %esi
    movl _TileCache2,%edi
    subl %edi,%esi
    andl $0xFFF0*4,%esi
    addl %esi,%edi
    shrl $2,%esi
    addl $_VRAM,%esi
    movl $0,TileOR
    movl $-1,TileAND
    movl $8,%edx
0:  xorl %ebx,%ebx
    //  Bp0=*(LineAddress+0)
    movb (%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    movl BPL0_2(,%ebx,4),%eax
    movl BPL0_2(,%ecx,4),%ebp

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL1_2(,%ebx,4),%eax
    orl  BPL1_2(,%ecx,4),%ebp

    movl %eax,(%edi)
    orl %eax,TileOR
    andl %eax,TileAND
    movl %ebp,4(%edi)
    orl %ebp,TileOR
    andl %ebp,TileAND
    addl $2,%esi
    addl $8,%edi
    decl %edx
    jnz 0b
    popl %esi
    xorl %eax,%eax
    popl %edi
    cmpl %eax,TileOR
    popl %ebx
    jnz 0f
    popl %eax
    movb $0x18,_CacheDirty2(%eax)
    jmpl Tile_Line_8_2+0x18(%ebx)
0:  movl TileAND,%eax
    movl %eax,%edx
    shrl $16,%eax
    andl %edx,%eax
    andb %ah,%al
    popl %eax
    jz  0f
//  movb $0x20,_CacheDirty2(%eax)
//  jmpl Tile_Line_8_2+0x20(%ebx)
0:  movb $0x10,_CacheDirty2(%eax)
    jmpl Tile_Line_8_2+0x10(%ebx)

ALIGN
.globl CacheMiss8Bpl_16
CacheMiss8Bpl_16:
    pushl %esi
    pushl %edi
    xorl %edx,%edx
    movw $0x1010,_CacheDirty8(%eax)
    movl _TileCache8,%edi
    andl $0xFFC0,%esi
    pushl %edx
    addl %esi,%edi
    movb $8,%dl
    addl $_VRAM,%esi
    pushl %edx
0:  xorl %ebx,%ebx
    //  Bp0=*(LineAddress+0)
    movb (%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    movl BPL0_8(,%ebx,4),%eax
    movl BPL0_8(,%ecx,4),%ebp

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL1_8(,%ebx,4),%eax
    orl  BPL1_8(,%ecx,4),%ebp

    //  Bp2=*(LineAddress+16)
    movb 16(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL2_8(,%ebx,4),%eax
    orl  BPL2_8(,%ecx,4),%ebp

    //  Bp3=*(LineAddress+17)
    movb 17(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL3_8(,%ebx,4),%eax
    orl  BPL3_8(,%ecx,4),%ebp

    //  Bp4=*(LineAddress+32)
    movb 32(%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL4_8(,%ebx,4),%eax
    orl  BPL4_8(,%ecx,4),%ebp

    //  Bp5=*(LineAddress+33)
    movb 33(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL5_8(,%ebx,4),%eax
    orl  BPL5_8(,%ecx,4),%ebp

    //  Bp6=*(LineAddress+48)
    movb 48(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL6_8(,%ebx,4),%eax
    orl  BPL6_8(,%ecx,4),%ebp

    //  Bp7=*(LineAddress+49)
    movb 49(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL7_8(,%ebx,4),%eax
    orl  BPL7_8(,%ecx,4),%ebp

    movl %eax,(%edi)
    movl %ebp,4(%edi)
    addl $2,%esi
    addl $8,%edi
    decl %edx
    jnz 0b
    popl %edx
    addl $48,%esi
    testl %edx,%edx
    jnz 0b
    popl %edi
    popl %esi
    ret

ALIGN
.globl CacheMiss4Bpl_16
CacheMiss4Bpl_16:
    pushl %esi
    pushl %edi
    andl $0xFFE0*2,%esi
    movl %esi,%edi
    shrl $1,%esi
    addl _TileCache4,%edi
    addl $_VRAM,%esi
    movw $0x1010,_CacheDirty4(%eax)
    xorl %edx,%edx
    pushl %edx
    movb $8,%dl
    pushl %edx
0:  xorl %ebx,%ebx
    //  Bp0=*(LineAddress+0)
    movb (%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    movl BPL0_4(,%ebx,4),%eax
    movl BPL0_4(,%ecx,4),%ebp

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL1_4(,%ebx,4),%eax
    orl  BPL1_4(,%ecx,4),%ebp

    //  Bp2=*(LineAddress+16)
    movb 16(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL2_4(,%ebx,4),%eax
    orl  BPL2_4(,%ecx,4),%ebp

    //  Bp3=*(LineAddress+17)
    movb 17(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL3_4(,%ebx,4),%eax
    orl  BPL3_4(,%ecx,4),%ebp

    movl %eax,(%edi)
    movl %ebp,4(%edi)
    addl $2,%esi
    addl $8,%edi
    decl %edx
    jnz 0b
    popl %edx
    addl $16,%esi
    testl %edx,%edx
    jnz 0b
    popl %edi
    popl %esi
    ret

ALIGN
.globl CacheMiss2Bpl_16
CacheMiss2Bpl_16:
    pushl %esi
    pushl %edi
    andl $0xFFF0*4,%esi
    movl %esi,%edi
    shrl $2,%esi
    addl _TileCache2,%edi
    addl $_VRAM,%esi
    movw $0x1010,_CacheDirty2(%eax)
    xorl %edx,%edx
    pushl %edx
    movb $8,%dl
    pushl %edx
0:  xorl %ebx,%ebx
    //  Bp0=*(LineAddress+0)
    movb (%esi),%bl
    movl %ebx,%ecx
    shrb $4,%bl
    andb $0x0F,%cl
    movl BPL0_2(,%ebx,4),%eax
    movl BPL0_2(,%ecx,4),%ebp

    //  Bp1=*(LineAddress+1)
    movb 1(%esi),%bl
    movb %bl,%cl
    shrb $4,%bl
    andb $0x0F,%cl
    orl  BPL1_2(,%ebx,4),%eax
    orl  BPL1_2(,%ecx,4),%ebp

    movl %eax,(%edi)
    movl %ebp,4(%edi)
    addl $2,%esi
    addl $8,%edi
    decl %edx
    jnz 0b
    popl %edx
    testl %edx,%edx
    jnz 0b
    popl %edi
    popl %esi
    ret

.macro P8_Plot_8 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFC0,%eax
    shrl $6,%eax
    movb _CacheDirty8(%eax),%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss8Bpl
1:  addl _TileCache8,%esi
.endif

    movb $0,%dh
.if \Xflip-1
    cmpb \Add(%edi),%dh
    jnz 1f
    movb (%esi),%al
    movb %al,\Add(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 1(%esi),%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 3(%esi),%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+4(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    movb %al,\Add+4(%edi)
1:  cmpb \Add+5(%edi),%dh
    jnz 1f
    movb 5(%esi),%al
    movb %al,\Add+5(%edi)
1:  cmpb \Add+6(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    movb %al,\Add+6(%edi)
1:  cmpb \Add+7(%edi),%dh
    jnz 1f
    movb 7(%esi),%al
    movb %al,\Add+7(%edi)
1:
.else
    cmpb \Add+7(%edi),%dh
    jnz 1f
    movb (%esi),%al
    movb %al,\Add+7(%edi)
1:  cmpb \Add+6(%edi),%dh
    jnz 1f
    movb 1(%esi),%al
    movb %al,\Add+6(%edi)
1:  cmpb \Add+5(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    movb %al,\Add+5(%edi)
1:  cmpb \Add+4(%edi),%dh
    jnz 1f
    movb 3(%esi),%al
    movb %al,\Add+4(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 5(%esi),%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add(%edi),%dh
    jnz 1f
    movb 7(%esi),%al
    movb %al,\Add(%edi)
1:
.endif
.endm

.macro P16_Plot_8 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFC0,%eax
    shrl $6,%eax
    movw _CacheDirty8(%eax),%cx
    andb %ch,%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss8Bpl_16
1:  addl _TileCache8,%esi
.endif

.if \Xflip-1
    pushl %esi
    call PLOT8_8BplTile
    popl %esi
    addl $64,%esi
    addl $8,%edi
    call PLOT8_8BplTile
    subl $8,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_8BplTile_X
    popl %esi
    addl $8,%edi
    call PLOT8_8BplTile_X
    subl $8,%edi
.endif
.endm

.macro P8_Plot_4 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFE0*2,%eax
    shrl $6,%eax
    movb _CacheDirty4(%eax),%cl
    testb %cl,%cl
    jnz 1f
    call CacheMiss4Bpl
1:  addl _TileCache4,%esi
.endif

    movb ColourBase,%dl
    movb $0,%dh
.if \Xflip-1
    cmpb \Add(%edi),%dh
    jnz 1f
    movb (%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 1(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 3(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+4(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+4(%edi)
1:  cmpb \Add+5(%edi),%dh
    jnz 1f
    movb 5(%esi),%al
    andb %dl,%al
    movb %al,\Add+5(%edi)
1:  cmpb \Add+6(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add+6(%edi)
1:  cmpb \Add+7(%edi),%dh
    jnz 1f
    movb 7(%esi),%al
    andb %dl,%al
    movb %al,\Add+7(%edi)
1:
.else
    cmpb \Add+7(%edi),%dh
    jnz 1f
    movb (%esi),%al
    andb %dl,%al
    movb %al,\Add+7(%edi)
1:  cmpb \Add+6(%edi),%dh
    jnz 1f
    movb 1(%esi),%al
    andb %dl,%al
    movb %al,\Add+6(%edi)
1:  cmpb \Add+5(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+5(%edi)
1:  cmpb \Add+4(%edi),%dh
    jnz 1f
    movb 3(%esi),%al
    andb %dl,%al
    movb %al,\Add+4(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 5(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add(%edi),%dh
    jnz 1f
    movb 7(%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:
.endif
.endm

.macro P8_Plot_Half_4 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFE0*2,%eax
    shrl $6,%eax
    movb _CacheDirty4(%eax),%cl
    testb %cl,%cl
    jnz 1f
    call CacheMiss4Bpl
1:  addl _TileCache4,%esi
.endif

    movb ColourBase,%dl
    movb $0,%dh
.if \Xflip-1
    cmpb \Add(%edi),%dh
    jnz 1f
    movb (%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:
.else
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 0(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:
.endif
.endm

.macro P16_Plot_Half_4 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFE0*2,%eax
    shrl $6,%eax
    movw _CacheDirty4(%eax),%cx
    andb %ch,%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss4Bpl_16
1:  addl _TileCache4,%esi
.endif

.if \Xflip-1
    pushl %esi
    call PLOT8_4BplTile_Even
    popl %esi
    addl $64,%esi
    addl $4,%edi
    call PLOT8_4BplTile_Even
    subl $4,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_4BplTile_Even_X
    popl %esi
    addl $4,%edi
    call PLOT8_4BplTile_Even_X
    subl $4,%edi
.endif
.endm

.macro P16_Plot_4 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFE0*2,%eax
    shrl $6,%eax
    movw _CacheDirty4(%eax),%cx
    andb %ch,%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss4Bpl_16
1:  addl _TileCache4,%esi
.endif

.if \Xflip-1
    pushl %esi
    call PLOT8_4BplTile
    popl %esi
    addl $64,%esi
    addl $8,%edi
    call PLOT8_4BplTile
    subl $8,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_4BplTile_X
    popl %esi
    addl $8,%edi
    call PLOT8_4BplTile_X
    subl $8,%edi
.endif
.endm

.macro P8_Plot_2 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFF0*4,%eax
    shrl $6,%eax
    movb _CacheDirty2(%eax),%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss2Bpl
1:  addl _TileCache2,%esi
.endif

    movb ColourBase,%dl
    movb $0,%dh
.if \Xflip-1
    cmpb \Add(%edi),%dh
    jnz 1f
    movb (%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 1(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 3(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+4(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+4(%edi)
1:  cmpb \Add+5(%edi),%dh
    jnz 1f
    movb 5(%esi),%al
    andb %dl,%al
    movb %al,\Add+5(%edi)
1:  cmpb \Add+6(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add+6(%edi)
1:  cmpb \Add+7(%edi),%dh
    jnz 1f
    movb 7(%esi),%al
    andb %dl,%al
    movb %al,\Add+7(%edi)
1:
.else
    cmpb \Add+7(%edi),%dh
    jnz 1f
    movb (%esi),%al
    andb %dl,%al
    movb %al,\Add+7(%edi)
1:  cmpb \Add+6(%edi),%dh
    jnz 1f
    movb 1(%esi),%al
    andb %dl,%al
    movb %al,\Add+6(%edi)
1:  cmpb \Add+5(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+5(%edi)
1:  cmpb \Add+4(%edi),%dh
    jnz 1f
    movb 3(%esi),%al
    andb %dl,%al
    movb %al,\Add+4(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 5(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add(%edi),%dh
    jnz 1f
    movb 7(%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:
.endif
.endm

.macro P8_Plot_All_2 Xflip,Add=0
    movb ColourBase,%dl
.if \Xflip-1
    movl (%esi),%eax
    movl 4(%esi),%ebx
    andb %dl,%al
    andb %dl,%bl
    movb %al,\Add(%edi)
    movb %bl,\Add+4(%edi)
    andb %dl,%ah
    andb %dl,%bh
    movb %ah,\Add+1(%edi)
    shrl $16,%eax
    movb %bh,\Add+5(%edi)
    shrl $16,%ebx
    andb %dl,%al
    andb %dl,%bl
    movb %al,\Add+2(%edi)
    movb %bl,\Add+6(%edi)
    andb %dl,%ah
    andb %dl,%bh
    movb %ah,\Add+3(%edi)
    movb %bh,\Add+7(%edi)
.else
    movl (%esi),%eax
    movl 4(%esi),%ebx
    andb %dl,%al
    andb %dl,%bl
    movb %al,\Add+7(%edi)
    movb %bl,\Add+3(%edi)
    andb %dl,%ah
    andb %dl,%bh
    movb %ah,\Add+6(%edi)
    shrl $16,%eax
    movb %bh,\Add+2(%edi)
    shrl $16,%ebx
    andb %dl,%al
    andb %dl,%bl
    movb %al,\Add+5(%edi)
    movb %bl,\Add+1(%edi)
    andb %dl,%ah
    andb %dl,%bh
    movb %ah,\Add+4(%edi)
    movb %bh,\Add(%edi)
.endif
.endm


.macro P8_Plot_Half_2 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFF0*4,%eax
    shrl $6,%eax
    movb _CacheDirty2(%eax),%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss2Bpl
1:  addl _TileCache2,%esi
.endif

    movb ColourBase,%dl
    movb $0,%dh
.if \Xflip-1
    cmpb \Add(%edi),%dh
    jnz 1f
    movb (%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:
.else
1:  cmpb \Add+3(%edi),%dh
    jnz 1f
    movb 0(%esi),%al
    andb %dl,%al
    movb %al,\Add+3(%edi)
1:  cmpb \Add+2(%edi),%dh
    jnz 1f
    movb 2(%esi),%al
    andb %dl,%al
    movb %al,\Add+2(%edi)
1:  cmpb \Add+1(%edi),%dh
    jnz 1f
    movb 4(%esi),%al
    andb %dl,%al
    movb %al,\Add+1(%edi)
1:  cmpb \Add(%edi),%dh
    jnz 1f
    movb 6(%esi),%al
    andb %dl,%al
    movb %al,\Add(%edi)
1:
.endif
.endm

.macro P16_Plot_Half_2 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFF0*4,%eax
    shrl $6,%eax
    movw _CacheDirty2(%eax),%cx
    andb %ch,%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss2Bpl_16
1:  addl _TileCache2,%esi
.endif

.if \Xflip-1
    pushl %esi
    call PLOT8_2BplTile_Even
    popl %esi
    addl $64,%esi
    addl $4,%edi
    call PLOT8_2BplTile_Even
    subl $4,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_2BplTile_Even_X
    popl %esi
    addl $4,%edi
    call PLOT8_2BplTile_Even_X
    subl $4,%edi
.endif
.endm

.macro P16_Plot_2 Xflip,Add=0,NoCacheCheck=0
.if \NoCacheCheck-1
    movl %esi,%eax
    andl $0xFFF0*4,%eax
    shrl $6,%eax
    movw _CacheDirty2(%eax),%cx
    andb %ch,%cl
    andl $0x7F,%ecx
    jnz 1f
    call CacheMiss2Bpl_16
1:  addl _TileCache2,%esi
.endif

.if \Xflip-1
    pushl %esi
    call PLOT8_2BplTile
    popl %esi
    addl $64,%esi
    addl $8,%edi
    call PLOT8_2BplTile
    subl $8,%edi
.else
    pushl %esi
    addl $64,%esi
    call PLOT8_2BplTile_X
    popl %esi
    addl $8,%edi
    call PLOT8_2BplTile_X
    subl $8,%edi
.endif
.endm

ALIGN
PLOT8_8BplTile:
    P8_Plot_8 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT8_8BplTile_X:
    P8_Plot_8 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT8_4BplTile:
    P8_Plot_4 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT8_4BplTile_X:
    P8_Plot_4 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT8_2BplTile:
    P8_Plot_2 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT8_2BplTile_X:
    P8_Plot_2 Xflip=1 NoCacheCheck=1
    ret

PLOT8_4BplTile_Even:
    P8_Plot_Half_4 Xflip=0 NoCacheCheck=1
    ret

PLOT8_4BplTile_Even_X:
    P8_Plot_Half_4 Xflip=1 NoCacheCheck=1
    ret

PLOT8_2BplTile_Even:
    P8_Plot_Half_2 Xflip=0 NoCacheCheck=1
    ret

PLOT8_2BplTile_Even_X:
    P8_Plot_Half_2 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT16_4BplTile_Even:
    P16_Plot_Half_4 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT16_4BplTile_Even_X:
    P16_Plot_Half_4 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT16_2BplTile_Even:
    P16_Plot_Half_2 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT16_2BplTile_Even_X:
    P16_Plot_Half_2 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT16_8BplTile:
    P16_Plot_8 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT16_8BplTile_X:
    P16_Plot_8 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT16_4BplTile:
    P16_Plot_4 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT16_4BplTile_X:
    P16_Plot_4 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT16_2BplTile:
    P16_Plot_2 Xflip=0 NoCacheCheck=1
    ret

ALIGN
PLOT16_2BplTile_X:
    P16_Plot_2 Xflip=1 NoCacheCheck=1
    ret

ALIGN
PLOT8_4BplTile_Check:
    P8_Plot_4 Xflip=0
    ret

ALIGN
PLOT8_4BplTile_Check_X:
    P8_Plot_4 Xflip=1
    ret

ALIGN
PLOT32_4BplSprite:
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    jmp PLOT8_4BplTile_Check
ALIGN
PLOT32_4BplSprite_X:
    pushl %esi
    addl $64*3,%esi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    call PLOT8_4BplTile_Check_X
    popl %esi
    addl $8,%edi
    jmp PLOT8_4BplTile_Check_X

ALIGN
PLOT64_4BplSprite:
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check
    popl %esi
    addl $64,%esi
    addl $8,%edi
    jmp PLOT8_4BplTile_Check
ALIGN
PLOT64_4BplSprite_X:
    pushl %esi
    addl $64*7,%esi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    pushl %esi
    call PLOT8_4BplTile_Check_X
    popl %esi
    subl $64,%esi
    addl $8,%edi
    call PLOT8_4BplTile_Check_X
    popl %esi
    addl $8,%edi
    jmp PLOT8_4BplTile_Check_X

ALIGN
.globl Invalidate_Tile_Caches
Invalidate_Tile_Caches:
    cld
    movl $0,%eax
    movl $_CacheDirty2,%edi
    movl $0x400,%ecx
    rep; stosl
    movl $_CacheDirty4,%edi
    movl $0x200,%ecx
    rep; stosl
    movl $_CacheDirty8,%edi
    movl $0x100,%ecx
    rep; stosl
    ret

.text; ALIGN
.data; ALIGN
.bss; ALIGN
